/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

/*
 * A cidev (cluster information device) is a device (/dev/whatever) which
 * is used to store and retrieve basic information (ostensibly cluster info
 * which is shared by making the cidev shared). The data on a cidev
 * is easily accessed from both user space and the kernel.
 *
 * This is an alternative to a couple approaches:
 * 1. config file on each system in the cluster.  Main difficulties are
 *    managing all the config files, accessing the config file data from
 *    the kernel, and accessing it from the kernel early in system startup
 *    when the local fs may not be mounted.
 * 2. info embedded in the lilo boot loader.  This is a plainly ugly, 
 *    unmanageable and inextensible approach.
 *
 * This library is simply concerned with reading and writing generic cidev
 * entries.  Code above this should manage consistency of the data, deciding
 * when to reread data which may have changed.  The code above this should
 * also do any synchronization which may be desired for handling changes.
 *
 * The point of this library is to be very generic and deal only with
 * keyword/string pairs each of which goes in a separate disk block.
 * Higher level routines should use these functions intelligently probably
 * based on specific keywords and string contents.
 *
 * At the start of any function which reads from the cidev, there is a
 * "ioctl(BLKFLSBUF)" which is supposed to invalidate any locally cached
 * data for the device.  This means we should always read the latest data
 * on disk, not anything from our cache.  At the end of any functin which
 * writes to the cidev, there is an "fsync()" which should write all cached
 * buffers to disk.
 *
 * This library and the cidev concept assume that the configuration information
 * placed on the cidev will be mostly static and read at only key times like
 * system boot.  It does allow cidev contents to be modified for cases where
 * cluster configuration info changes.  Cluster management including things
 * like cluster status information should not be kept here - the api's are
 * not designed for such use.
 *
 * System crashes during a write operation should not be a problem because 
 * each update involves only a single block and a single block write is atomic.
 * (I know this is true for sectors wrt to storage devices, can we stretch
 * this to include 1k or 4k blocks?)
 *
 * These routines don't do checks on the cidev magic number or between the 
 * ondisk and program versions.  Other code should use the cidev_read_header() 
 * to check the magic number compare versions before doing anything.  Some of 
 * these routines initially check that the cidev is inuse - seems kind of 
 * arbitrary, i think there should be a clear policy for when checks are done.
 */


#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/mount.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <search.h>
#include <errno.h>

#include "cidev.h"



/**
 * dev_flush
 * @fd: file descriptor for opened device
 *
 * Flushes buffers for a device, forcing next read to uncached
 */

#if defined(__FreeBSD__)
void dev_flush(int fd)
{
}
#endif

#if defined(__linux__)
void dev_flush(int fd)
{
    ioctl(fd, BLKFLSBUF);
}
#endif


/**
 * cidev_inuse
 *
 * Sets the boolean value accordingly.
 */

int cidev_inuse(int cfd, int *bool)
{
    cidev_header_t header;
    char buf[CIDEV_BLOCKSIZE];
    int len;

    dev_flush(cfd);

    if (lseek(cfd, 0, SEEK_SET) != 0)
    {
        perror("Error seeking on cidev");
        return -EIO;
    }

    len = CIDEV_BLOCKSIZE;
    if (read(cfd, buf, len) != len)
    {
        perror("Error reading cidev");
        return -EIO;
    }

    memset(&header, 0, sizeof(cidev_header_t));
    cidev_header_in(&header, buf);

    if (header.inuse == TRUE)
        *bool = TRUE;
    else
        *bool = FALSE;

    return 0;
}


/**
 * cidev_read_header
 *
 * Read the header which contains the version number and inuse flag.
 */

int cidev_read_header(int cfd, cidev_header_t *head)
{
    char buf[CIDEV_BLOCKSIZE];
    int len;

    dev_flush(cfd);
    
    if (lseek(cfd, 0, SEEK_SET) != 0)
    {
        perror("Error seeking on cidev");
        return -EIO;
    }
        
    len = CIDEV_BLOCKSIZE;
    if (read(cfd, buf, len) != len)
    {
        perror("Error reading cidev");
        return -EIO;
    }

    memset(head, 0, sizeof(cidev_header_t));
    cidev_header_in(head, buf);
    return 0;
}


/**
 * cidev_create
 *
 * Place all new entries on cidev.
 */

int cidev_create(int cfd, char *name, cidev_entry_t *list)
{
    cidev_header_t header;
    cidev_marker_t marker;
    cidev_entry_t *ent;
    char buf[CIDEV_BLOCKSIZE];
    int len, offset, start, block = 0;


    /* Write header block */

    memset(buf, 0, CIDEV_BLOCKSIZE);
    memset(&header, 0, sizeof(cidev_header_t));
    header.magic = CIDEV_MAGIC;
    header.version = CIDEV_VERSION;
    header.inuse = TRUE;
    strcpy(header.name, name);
    cidev_header_out(&header, buf);

    if (lseek(cfd, 0, SEEK_SET) != 0)
    {
        perror("Error seeking on cidev");
        return -EIO;
    }

    len = CIDEV_BLOCKSIZE;
    if (write(cfd, buf, len) != len)
    {
        perror("Error writing cidev");
        return -EIO;
    }


    /* Write each entry */

    for (ent = list; ent; ent = ent->next)
    {
        memset(buf, 0, CIDEV_BLOCKSIZE);

        marker.magic = CIDEV_ENTRY_MAGIC;
        marker.inuse = TRUE;
        cidev_marker_out(&marker, buf);
        offset = sizeof(cidev_marker_t);
        memcpy(buf + offset, ent->keyword, CIDEV_KEYWORD_SIZE);
        offset += CIDEV_KEYWORD_SIZE;
        memcpy(buf + offset, ent->string, CIDEV_STRING_SIZE);

        block++;
        start = block * CIDEV_BLOCKSIZE;
        len = CIDEV_BLOCKSIZE;

        if (lseek(cfd, start, SEEK_SET) != start)
        {
            perror("Error seeking on cidev");
            return -EIO;
        }

        if (write(cfd, buf, len) != len)
        {
            perror("Error writing on cidev");
            return -EIO;
        }
    }


    /* Write out a header to mark the end */

    memset(buf, 0, CIDEV_BLOCKSIZE);
    memset(&header, 0, sizeof(cidev_header_t));
    header.magic = CIDEV_MAGIC;
    header.version = CIDEV_VERSION;
    header.inuse = (uint32)(-1);
    cidev_header_out(&header, buf);


    start = (block + 1) * CIDEV_BLOCKSIZE;
    len = CIDEV_BLOCKSIZE;
     
    if (lseek(cfd, start, SEEK_SET) != start)
    {
        perror("Error seeking on cidev");
        return -EIO;
    }

    if (write(cfd, buf, len) != len)
    {
        perror("Error writing cidev");
        return -EIO;
    }

    if (fsync(cfd))
    {
        perror("Error syncing cidev");
        return -EIO;
    }

    return 0;
}


/**
 * cidev_erase
 *
 * Make cidev not "inuse".  Clears the inuse flag in the cidev header.
 */

int cidev_erase(int cfd)
{
    cidev_header_t header;
    char buf[CIDEV_BLOCKSIZE];
    int len;

    dev_flush(cfd);
    
    memset(buf, 0, CIDEV_BLOCKSIZE);
    
    if (lseek(cfd, 0, SEEK_SET) != 0)
    {
        perror("Error seeking on cidev");
        return -EIO;
    }
        
    len = CIDEV_BLOCKSIZE;
    if (read(cfd, buf, len) != len)
    {
        perror("Error reading cidev");
        return -EIO;
    }

    cidev_header_in(&header, buf);
    header.inuse = FALSE;
    cidev_header_out(&header, buf);

    if (lseek(cfd, 0, SEEK_SET) != 0)
    {
        perror("Error seeking on cidev");
        return -EIO;
    }

    if (write(cfd, buf, len) != len)
    {
        perror("Error writing cidev");
        return -EIO;
    }

    if (fsync(cfd))
    {
        perror("Error syncing cidev");
        return -EIO;
    }

    return 0;
}


/**
 * cidev_entry_count
 *
 * Returns the current number of inuse entries.
 */

int cidev_entry_count(int cfd, int *count)
{
    cidev_marker_t marker;
    char buf[CIDEV_BLOCKSIZE];
    int len, block, start, entries = 0, inuse = 0;

    if (cidev_inuse(cfd, &inuse) || inuse != TRUE)
        return -EINVAL;

    for (block = 1; ; block++)
    {
        memset(buf, 0, CIDEV_BLOCKSIZE);
        start = block * CIDEV_BLOCKSIZE;
        len = CIDEV_BLOCKSIZE;

        if (lseek(cfd, start, SEEK_SET) != start)
        {
            perror("Error seeking on cidev");
            return -EIO;
        }

        if (read(cfd, buf, len) != len)
        {
            perror("Error reading cidev");
            return -EIO;
        }

        memset(&marker, 0, sizeof(cidev_marker_t));
        cidev_marker_in(&marker, buf);

        if (marker.magic == CIDEV_MAGIC)
            break;
        
        if (marker.magic != CIDEV_ENTRY_MAGIC)
        {
            perror("Bad entry on cidev");
            return -EIO;
        }

        if (marker.inuse == TRUE)
            entries++;
    }

    *count = entries;
    return 0;
}


/**
 * cidev_read_entries
 *
 * Reads inuse entries.  Quits when there are no more list entries or all 
 * inuse entries have been read.  The number of entries found and placed
 * in list is returned in count.
 *
 * Returns: 0, -EXXX
 */

int cidev_read_entries(int cfd, cidev_entry_t *list, int *count)
{
    cidev_marker_t marker;
    cidev_entry_t *ent;
    char buf[CIDEV_BLOCKSIZE];
    int block, len, start, offset, inuse = 0, entries = 0;

    *count = 0;

    if (cidev_inuse(cfd, &inuse) || inuse != TRUE)
        return -EINVAL;

    ent = list;
    for (block = 1; ; block++)
    {
        if (!ent)
            break;

        memset(buf, 0, CIDEV_BLOCKSIZE);
        start = block * CIDEV_BLOCKSIZE;
        len = CIDEV_BLOCKSIZE;

        if (lseek(cfd, start, SEEK_SET) != start)
        {
            perror("Error seeking on cidev");
            return -EIO;
        }

        if (read(cfd, buf, len) != len)
        {
            perror("Error reading cidev");
            return -EIO;
        }

        memset(&marker, 0, sizeof(cidev_marker_t));
        cidev_marker_in(&marker, buf);

        if (marker.magic == CIDEV_MAGIC)
            break;

        if (marker.magic != CIDEV_ENTRY_MAGIC)
        {
            perror("Bad entry on cidev");
            return -EIO;
        }

        if (marker.inuse != TRUE)
            continue;
        
        ent->id = block;
        ent->inuse = TRUE;
        offset = sizeof(cidev_marker_t);
        memcpy(ent->keyword, buf + offset, CIDEV_KEYWORD_SIZE);
        offset += CIDEV_KEYWORD_SIZE;
        memcpy(ent->string, buf + offset, CIDEV_STRING_SIZE);

        entries++;
        ent = ent->next;
    }
    *count = entries;
    return 0; 
}


/** 
 * cidev_read_entry_num
 *
 * Read an entry by number.  Valid numbers are in the range of zero to
 * the value returned by cidev_num_entries().  A number in this range will
 * return an inuse entry.
 *
 * Warning: a specific entry could likely be a different number if entries 
 * are removed and/or added, so this is not a reliable way of retrieving 
 * a specific entry (use id numbers for that).
 *
 * Returns: 0, -EXXX
 */

int cidev_read_entry_num(int cfd, int entry_num, cidev_entry_t *entry)
{
    cidev_marker_t marker;
    char buf[CIDEV_BLOCKSIZE];
    int len, block, start, offset, inuse, count = 0;

    if (cidev_inuse(cfd, &inuse) || inuse != TRUE)
        return -EINVAL;

    for (block = 1; ; block++)
    {
        memset(buf, 0, CIDEV_BLOCKSIZE);
        start = block * CIDEV_BLOCKSIZE;
        len = CIDEV_BLOCKSIZE;

        if (lseek(cfd, start, SEEK_SET) != start)
        {
            perror("Error seeking on cidev");
            return -EIO;
        }

        if (read(cfd, buf, len) != len)
        {
            perror("Error reading cidev");
            return -EIO;
        }

        memset(&marker, 0, sizeof(cidev_marker_t));
        cidev_marker_in(&marker, buf);

        if (marker.magic == CIDEV_MAGIC)
            break;

        if (marker.magic != CIDEV_ENTRY_MAGIC)
        {
            perror("Bad entry on cidev");
            return -EINVAL;
        }

        if (marker.inuse != TRUE)
            continue;

        count++;

        if (count == entry_num)
        {
            memset(entry, 0, sizeof(cidev_entry_t));
            entry->id = block;
            entry->inuse = TRUE;
            offset = sizeof(cidev_marker_t);
            memcpy(entry->keyword, buf + offset, CIDEV_KEYWORD_SIZE);
            offset += CIDEV_KEYWORD_SIZE;
            memcpy(entry->string, buf + offset, CIDEV_STRING_SIZE);
            return 0;
        }
    }

    return -ENOENT;
}


/**
 * cidev_read_entry
 *
 * Read an entry by id.  This retrieves a specific entry (id's returned by
 * _create, _read_entry_num, _read_entries).  Does not require a "valid" 
 * cidev.  Warning: the entry returned may not be inuse.
 *
 * Returns: 0, -EXXX
 */

int cidev_read_entry(int cfd, int entry_id, cidev_entry_t *entry)
{
    cidev_marker_t marker;
    char buf[CIDEV_BLOCKSIZE];
    int len, start, offset;

    dev_flush(cfd);

    start = entry_id * CIDEV_BLOCKSIZE;
    len = CIDEV_BLOCKSIZE;

    if (lseek(cfd, start, SEEK_SET) != start)
    {
        perror("Error seeking on cidev");
        return -EIO;
    }

    if (read(cfd, buf, len) != len)
    {
        perror("Error reading cidev");
        return -EIO;
    }

    memset(&marker, 0, sizeof(cidev_marker_t));
    cidev_marker_in(&marker, buf);

    if (marker.magic != CIDEV_ENTRY_MAGIC)
        return -EINVAL;

    memset(entry, 0, sizeof(cidev_entry_t));
    entry->id = entry_id;
    entry->inuse = marker.inuse;
    offset = sizeof(cidev_marker_t);
    memcpy(entry->keyword, buf + offset, CIDEV_KEYWORD_SIZE);
    offset += CIDEV_KEYWORD_SIZE;
    memcpy(entry->string, buf + offset, CIDEV_STRING_SIZE);
    return 0;
}


/**
 * cidev_write_entry
 *
 * Write an entry by id.  Does not require a "valid" cidev.
 * Warning: No checks at all, let's you clobber whatever.
 * 
 * Returns: 0, -EXXX
 */

int cidev_write_entry(int cfd, int entry_id, cidev_entry_t *entry)
{
    cidev_marker_t marker;
    char buf[CIDEV_BLOCKSIZE];
    int len, start, offset;

    memset(&marker, 0, sizeof(cidev_marker_t));
    marker.magic = CIDEV_ENTRY_MAGIC;
    marker.inuse = TRUE;
    cidev_marker_out(&marker, buf);
    
    offset = sizeof(cidev_marker_t);
    memcpy(buf + offset, entry->keyword, CIDEV_KEYWORD_SIZE);
    offset += CIDEV_KEYWORD_SIZE;
    memcpy(buf + offset, entry->string, CIDEV_STRING_SIZE);
    
    start = entry_id * CIDEV_BLOCKSIZE;
    len = CIDEV_BLOCKSIZE;

    if (lseek(cfd, start, SEEK_SET) != start)
    {
        perror("Error seeking on cidev");
        return -EIO;
    }

    if (write(cfd, buf, len) != len)
    {
        perror("Error writing to cidev");
        return -EIO;
    }

    if (fsync(cfd))
    {
        perror("Error syncing cidev");
        return -EIO;
    }

    return 0;
}


/**
 * cidev_remove_entry
 *
 * Sets the entry to not inuse.
 */

int cidev_remove_entry(int cfd, int entry_id)
{
    cidev_marker_t marker;
    char buf[CIDEV_BLOCKSIZE];
    int len, start, inuse;

    if (cidev_inuse(cfd, &inuse) || inuse != TRUE)
        return -EINVAL;

    start = entry_id * CIDEV_BLOCKSIZE;
    len = CIDEV_BLOCKSIZE;

    if (lseek(cfd, start, SEEK_SET) != start)
    {
        perror("Error seeking on cidev");
        return -EIO;
    }

    if (read(cfd, buf, len) != len)
    {
        perror("Error reading cidev");
        return -EIO;
    }

    cidev_marker_in(&marker, buf);

    if (marker.magic != CIDEV_ENTRY_MAGIC)
    {
        perror("Bad cidev entry");
        return -EINVAL;
    }
    marker.inuse = FALSE;
    cidev_marker_out(&marker, buf);

    if (lseek(cfd, start, SEEK_SET) != start)
    {
        perror("Error seeking on cidev");
        return -EIO;
    }

    if (write(cfd, buf, len) != len)
    {
        perror("Error writing to cidev");
        return -EIO;
    }

    if (fsync(cfd))
    {
        perror("Error syncing cidev");
        return -EIO;
    }

    return 0;
}


/**
 * cidev_add_entry
 *
 * Find an unused block (of a removed entry) for the new entry or add it
 * to the end.
 *
 * Returns: 0, -EXXX
 */

int cidev_add_entry(int cfd, cidev_entry_t *entry)
{
    cidev_marker_t marker;
    char buf[CIDEV_BLOCKSIZE];
    int len, start, block, inuse;

    if (cidev_inuse(cfd, &inuse) || inuse != TRUE)
        return -EINVAL;


    for (block = 1; ; block++)
    {
        memset(buf, 0, CIDEV_BLOCKSIZE);
        start = block * CIDEV_BLOCKSIZE;
        len = CIDEV_BLOCKSIZE;

        if (lseek(cfd, start, SEEK_SET) != start)
        {
            perror("Error seeking on cidev");
            return -EIO;
        }

        if (read(cfd, buf, len) != len)
        {
            perror("Error reading cidev");
            return -EIO;
        }

        cidev_marker_in(&marker, buf);

        if (marker.magic == CIDEV_MAGIC)
            break;

        if (marker.magic != CIDEV_ENTRY_MAGIC)
        {
            perror("Bad entry on cidev");
            return -EINVAL;
        }

        if (marker.inuse != TRUE)
            return cidev_write_entry(cfd, block, entry);
    }


    /* move tail block out one and then write new entry (tail block just read) */

    start = (block + 1) * CIDEV_BLOCKSIZE;
    if (lseek(cfd, start, SEEK_SET) != start)
    {
        perror("Error seeking on cidev");
        return -EIO;
    }

    len = CIDEV_BLOCKSIZE;
    if (write(cfd, buf, len) != len)
    {
        perror("Error reading cidev");
        return -EIO;
    }

    return cidev_write_entry(cfd, block, entry);
}


/**
 * cidev_read_entry_keyword
 *
 * Reads n'th entry with keyword.  First entry is n = 1
 */

int cidev_read_entry_keyword(int cfd, char *kw, int n, cidev_entry_t *entry)
{
    cidev_marker_t marker;
    char buf[CIDEV_BLOCKSIZE];
    int block, len, start, inuse = 0, found = 0, slen1, slen2, offset;

    if (cidev_inuse(cfd, &inuse) || inuse != TRUE)
        return -EINVAL;

    for (block = 1; ; block++)
    {
        memset(buf, 0, CIDEV_BLOCKSIZE);
        start = block * CIDEV_BLOCKSIZE;
        len = CIDEV_BLOCKSIZE;

        if (lseek(cfd, start, SEEK_SET) != start)
        {
            perror("Error seeking on cidev");
            return -EIO;
        }

        if (read(cfd, buf, len) != len)
        {
            perror("Error reading cidev");
            return -EIO;
        }

        memset(&marker, 0, sizeof(cidev_marker_t));
        cidev_marker_in(&marker, buf);

        if (marker.magic == CIDEV_MAGIC)
            break;

        if (marker.magic != CIDEV_ENTRY_MAGIC)
        {
            perror("Bad entry on cidev");
            return -EINVAL;
        }

        if (marker.inuse != TRUE)
            continue;

        slen1 = strlen(kw);
        slen2 = strlen(buf + sizeof(cidev_marker_t));

        if ((slen1 != slen2) || strncmp(kw, buf + sizeof(cidev_marker_t), slen1))
            continue;

        found++;

        if (found < n)
            continue;

        memset(entry, 0, sizeof(cidev_entry_t));
        entry->id = block;
        entry->inuse = TRUE;
        offset = sizeof(cidev_marker_t);
        memcpy(entry->keyword, buf + offset, CIDEV_KEYWORD_SIZE);
        offset += CIDEV_KEYWORD_SIZE;
        memcpy(entry->string, buf + offset, CIDEV_STRING_SIZE);
        return 0;
    }

    return -ENOENT;
}


/**
 * cidev_keyword_count
 *
 * Returns the current number of inuse entries with matching keyword.
 */

int cidev_keyword_count(int cfd, char *kw, int *count)
{
    cidev_marker_t marker;
    char buf[CIDEV_BLOCKSIZE];
    int block, len, start, slen1, slen2, found = 0, inuse = 0;

    if (cidev_inuse(cfd, &inuse) || inuse != TRUE)
        return -EINVAL;

    for (block = 1; ; block++)
    {
        memset(buf, 0, CIDEV_BLOCKSIZE);
        start = block * CIDEV_BLOCKSIZE;
        len = CIDEV_BLOCKSIZE;

        if (lseek(cfd, start, SEEK_SET) != start)
        {
            perror("Error seeking on cidev");
            return -EIO;
        }

        if (read(cfd, buf, len) != len)
        {
            perror("Error reading cidev");
            return -EIO;
        }

        cidev_marker_in(&marker, buf);

        if (marker.magic == CIDEV_MAGIC)
            break;
        
        if (marker.magic != CIDEV_ENTRY_MAGIC)
        {
            perror("Bad entry on cidev");
            return -EIO;
        }

        if (marker.inuse != TRUE)
            continue;

        slen1 = strlen(kw);
        slen2 = strlen(buf + sizeof(cidev_marker_t));

        if ((slen1 != slen2) || strncmp(kw, buf + sizeof(cidev_marker_t), slen1))
            continue;

        found++;
    }

    *count = found;
    return 0;
}



/**
 * get_string_val
 *
 * Here and other places we assume that inside a string (src) we'll be looking
 * for something like: key = value where there may be white space around the
 * "=" and whitespace or newline ends the value.
 * The returned pointer points to the beginning of 'value'.
 * The returned len is the length of 'value'.
 */

char *get_string_val(char *src, char *key, int *len)
{
    char *k, *p, *ptr;
    int i = 0, max = CIDEV_STRING_SIZE;

    *len = 0;

    if (!src || !key)
        return NULL;

    ptr = src;

    while ( i < max )
    {
        int match = 0; /* 0 = no match; 1 = match */

        /* strip leading whitespace if '\0' is match, the entire src string 
         * has been searched without success*/
        p = ptr;
        while ( i < max )
        {
            if ( *p == '\0' )
                return NULL;
            else if ( *p != '\n' && *p != ' ' && *p != '\t')
                break;

            i++;
            p++;
        }
        ptr = p;

        /* FIXME -- is the cidev suppose to ignore comment characters? */
        /*if ( *p == '#' )
        {
            while ( i < max )
            {
                if ( *p == '\0' )
                    return NULL;
                else if ( *p != '\n')
                    break;
  
                i++;
                p++;
            }
            ptr = p;
            continue;
        }*/


        /* compare ptr with key */
        k = key;
        p = ptr;
        while ( i < max )
        {
            if ( *p == '\0' )
                return NULL;
             else if ( *k == '\0' )
            {
                match = 1;
                break;
            }
            else if ( *p == *k )
            {
                i++;
                k++;
                p++;
            }
            else
            {
                /* move ptr to next line */
                i++;
                p++;

                while ( i < max )
                {
                    if ( *p == '\0' )
                        return NULL;
                    else if ( *p == '\n' )
                        break;

                    i++;
                    p++;
                }
                break;
            }
        }
        ptr = p;

        if (!match)
            continue;

        /* verfiy that a key was found and not just a prefix*/
        while ( i < max )
        {
            if ( *p == '\0' )
                return NULL;
            else if ( *p == '=')
                break;
            else if ( *p == ' ' || *p == '\t')
            {
                i++;
                p++;
            }
            else
            {
                /* move ptr to next line */
                while ( i < max )
                {
                    if ( *p == '\n' )
                        break;
                    else if ( *p == '\0' )
                        return NULL;

                    ++i;
                    ++p;
                }
                match = 0;
                break;
            }
        }
        ptr = p;

        if (!match)
            continue;

        /* strip the trailing whitspace after '=' */
        i++;
        p++;
        while ( i < max )
        {
            if ( *p != ' ' && *p != '\t' )
                break;
            i++;
            p++;
        }
        ptr = p;

        /* scan the remainder of the line up to '\n' and '\0' and determine the length */
        while ( i < max )
        {
            if (*p == '\n' || *p == '\0' )
                break;
            i++;
            p++;
        }

        *len = p - ptr;
        break;
    }
    return ptr;
}

