/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "cidev.h"
#include "gfsconf.h"


#define print_entry(ent) printf("%s:\n%s\n", (ent)->keyword, (ent)->string)

extern char *prog_name;



/**
 * print_gfs_config
 *
 * Pretty much the same as print_gen_config for now.
 */

void print_gfs_config(gfs_config_t *gfs)
{
    cidev_entry_t *ent;

    if (gfs->gfsparam_count)
    {
        printf("# gfsparam_count: %d\n", gfs->gfsparam_count);
        print_entry(gfs->gfsparam);
    }

    if (gfs->lmparam_count)
    {
        printf("# lmparam_count: %d\n", gfs->lmparam_count);
        print_entry(gfs->lmparam);
    }

    if (gfs->cmparam_count)
    {
        printf("# cmparam_count: %d\n", gfs->cmparam_count);
        print_entry(gfs->cmparam);
    }

    if (gfs->stomith_count)
    {
        printf("# fence count: %d\n", gfs->stomith_count);
        for (ent = gfs->stomith; ent; ent = ent->next)
            print_entry(ent);
    }

    if (gfs->node_count)
    {
        printf("# node_count: %d\n", gfs->node_count);
        for (ent = gfs->node; ent; ent = ent->next)
            print_entry(ent);
    }

    if (gfs->other_count)
    {
        printf("# Warning - %d unrecognized gfs cidev entries\n", 
                gfs->other_count);

        for (ent = gfs->other; ent; ent = ent->next)
            print_entry(ent);
    }
}



/**
 * make_gfs_config
 *
 * Take a list of generic cidev entries and separate them based on keyword
 * into gfs config components.
 */

int make_gfs_config(cidev_entry_t *gen_config, gfs_config_t *gfs)
{
    cidev_entry_t *ent, **list, *p, *new;

    memset(gfs, 0, sizeof(gfs_config_t));

    for (ent = gen_config; ent; ent = ent->next)
    {
        new = (cidev_entry_t *)malloc(sizeof(cidev_entry_t));
        if (!new)
            return -ENOMEM;
        memset(new, 0, sizeof(cidev_entry_t));

        strncpy(new->keyword, ent->keyword, CIDEV_KEYWORD_SLEN);
        strncpy(new->string, ent->string, CIDEV_STRING_SLEN);
        new->id = ent->id;

        if (!strncmp(ent->keyword, "gfsparam", 8))
        {
            list = &gfs->gfsparam;
            gfs->gfsparam_count++;
        }

        else if (!strncmp(ent->keyword, "lmparam", 7))
        {
            list = &gfs->lmparam;
            gfs->lmparam_count++;
        }

        else if (!strncmp(ent->keyword, "cmparam", 7))
        {
            list = &gfs->cmparam;
            gfs->cmparam_count++;
        }

        else if (!strncmp(ent->keyword, "fence", 5))
        {
            list = &gfs->stomith;
            gfs->stomith_count++;
        }

        else if (!strncmp(ent->keyword, "node", 4))
        {
            list = &gfs->node;
            gfs->node_count++;
        }

        else
        {
            list = &gfs->other;
            gfs->other_count++;
        }


        if (!*list)
            *list = new;
        else
        {
            for (p = *list; p->next; p = p->next);
            p->next = new;
        }
    }

    return 0;
}


/**
 * update_config_add
 *
 * Any "stomith" or "node" entry which is new in the new config will be
 * added to the cidev.  Entries with any other keywords are ignored.
 * Stomith and node entries are compared using the "name" field.  Names
 * within these entries must be unique for all this to work.  (This assumes
 * that the "validate" routine has verified that "name" is in entries.)
 */

int update_config_add(int cfd, gfs_config_t *cur, gfs_config_t *new)
{
    cidev_entry_t *ent, *curent;
    int len1, len2;
    char *newname, *curname;
    int do_next;
    
    for (ent = new->node; ent; ent = ent->next)
    {
        newname = get_string_val(ent->string, "name", &len1);

	do_next = FALSE;

        for (curent = cur->node; curent; curent = curent->next)
        {
            curname = get_string_val(curent->string, "name", &len2);
            if ((len1 == len2) && !strncmp(newname, curname, len1))
	    {
	      do_next = TRUE;
	      break;
	    }
        }

        if (!do_next && cidev_add_entry(cfd, ent))
            return -1;
    }
    
    for (ent = new->stomith; ent; ent = ent->next)
    {
        newname = get_string_val(ent->string, "name", &len1);
        
	do_next = FALSE;

        for (curent = cur->stomith; curent; curent = curent->next)
        {
            curname = get_string_val(curent->string, "name", &len2);
            if ((len1 == len2) && !strncmp(newname, curname, len1))
	    {
	      do_next = TRUE;
	      break;
	    }
        }

        if (!do_next && cidev_add_entry(cfd, ent))
            return -1;
    }

    return 0;
}


int update_config_remove(int cfd, gfs_config_t *gfs, char *name)
{
    cidev_entry_t *ent;
    char *ename;
    int len;

    for (ent = gfs->node; ent; ent = ent->next)
    {
        ename = get_string_val(ent->string, "name", &len);

        if ((len == strlen(name)) && !strncmp(ename, name, len))
            break;
    }

    if (!ent)
        die("no entry found matching \"%s\"\n", name);

    return cidev_remove_entry(cfd, ent->id);
}





