/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include <sys/ioctl.h>
#include <string.h>
#include <errno.h>

#include "libgfs_ea.h"
#include "global.h"
#include <linux/gfs_ondisk.h>
#include <linux/gfs_ioctl.h>

/*
 * FIXME:
 * I can't find the requirements for extended attribute names documented 
 * anywhere, so I'm just going to allow any null terminated sequence of
 * characters. If the linux kernel people ever come up with some documentation
 * on the interface, I think I'll follow that, but until then, people can name
 * things however they want.
 */


/**
 * I allow mutliple ways of entering an attribute name
 * $<name> system.<name> or sys.<name> to make a system attribute named <name>
 * user.<name> usr.<name> or <name> make a user attribute named <name>
 */
static int get_type(char **namep){
  char *name = *namep;
  int len;
  if (name[0] == '$'){
    *namep = *namep + 1;
    return GFS_EATYPE_SYS;
  }
  name = strchr(*namep, '.');
  if (!name)
    return GFS_EATYPE_USR;
  len = name - *namep;
  name++;
  if(!strncmp(*namep, "user", len) || !strncmp(*namep, "usr", len)){
    *namep = name;
    return GFS_EATYPE_USR;
  }
  if(!strncmp(*namep, "system", len) || !strncmp(*namep, "sys", len)){
    *namep = name;
    return GFS_EATYPE_SYS;
  }
  errno = EINVAL;
  return -1;
}


int gfs_seteattr(int fd, char *name, void *data, unsigned int size, int cmd){
  int result;
  struct gfs_easet_io req;
  
  /* Insert mapping of user cmd to kernel cmd here, if necessary) */
  if (cmd < 0 || cmd > GFS_EA_REPLACE)
    goto fail;
  
  result = get_type(&name);
  if (result < 0)
    goto fail;
 
  if(strlen(name) > GFS_EA_MAX_NAME_LEN)
    goto fail; 

  if(size < 0 || size > GFS_EA_MAX_DATA_LEN)
    goto fail;

  req.es_type = (uint8)result;
  req.es_name = name;
  req.es_name_len = (uint8)strlen(name);
  req.es_data = data;
  req.es_data_len = (uint16)size;
  req.es_cmd = (uint8)cmd;
  return ioctl(fd, GFS_EATTR_SET, (unsigned long)&req);
 
 fail:
  errno = EINVAL;
  return -1;
}

int gfs_geteattrsize(int fd, char *name, unsigned int *sizep){ 
  int result;
  struct gfs_eaget_io req;
  
  result = get_type(&name);
  if (result < 0)
    return -1;

  if(strlen(name) > GFS_EA_MAX_NAME_LEN){
    errno = EINVAL;
    return -1;
  }

  req.eg_type = (uint8)result;
  req.eg_name = name;
  req.eg_name_len = (uint8)strlen(name);
  req.eg_data = NULL;
  req.eg_data_len = 0;
  req.eg_len = (char *)sizep;
  return ioctl(fd, GFS_EATTR_GET, (unsigned long)&req);
}
    

int gfs_geteattr(int fd, char *name, void *data, unsigned int *sizep){
  int result;
  struct gfs_eaget_io req;

  result = get_type(&name);
  if (result < 0)
    return -1;

  if(strlen(name) > GFS_EA_MAX_NAME_LEN){
    errno = EINVAL;
    return -1;
  }

  if(!data){
    errno = EINVAL;
    return -1;
  }
    
  req.eg_type = (uint8)result;
  req.eg_name = name;
  req.eg_name_len = (uint8)strlen(name);
  req.eg_data = data;
  req.eg_data_len = (uint32)*sizep;
  req.eg_len = (char *)sizep;
  return ioctl(fd, GFS_EATTR_GET, (unsigned long)&req);
}

int gfs_listeattrsize(int fd, unsigned int *sizep){
  struct gfs_eaget_io req;

  req.eg_name = NULL;
  req.eg_name_len = 0;
  req.eg_data = NULL;
  req.eg_data_len = 0;
  req.eg_len = (char *)sizep;
  return ioctl(fd, GFS_EATTR_GET, (unsigned long)&req);
}


int gfs_listeattr(int fd, char *list, unsigned int *sizep){
  struct gfs_eaget_io req;
 
  if (!list){
    errno = EINVAL;
    return -1;
  }

  req.eg_name = NULL;
  req.eg_name_len = 0;
  req.eg_data = list;
  req.eg_data_len = (uint32)*sizep;
  req.eg_len = (char *)sizep;
  return ioctl(fd, GFS_EATTR_GET, (unsigned long)&req);
}

int gfs_removeeattr(int fd, char *name){
  int result;
  struct gfs_easet_io req;
  
  result = get_type(&name);
  if (result < 0)
    return result;

  if(strlen(name) > GFS_EA_MAX_NAME_LEN)
    return -EINVAL;
  
  req.es_type = (uint8)result;
  req.es_name = name;
  req.es_name_len = (uint8)strlen(name);
  req.es_data = NULL;
  req.es_data_len = 0;
  req.es_cmd = GFS_EACMD_REMOVE;
  return ioctl(fd, GFS_EATTR_SET, (unsigned long)&req);
}

char *gfs_eaerror(int err){
  switch(err){
  case ENXIO:
    return "No such extended attribute";
  case EMSGSIZE:
    return "Extended attribute data to large for buffer";
  case ENOSPC:
    return "No space for extended attributes";
  case EEXIST:
    return "Extended attribute already exists";
  default:
    return strerror(err);
  }
}

