/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include <linux/xattr_acl.h>

#include "gfs.h"
#include "trans.h"
#include "dio.h"
#include "glock.h"
#include "eattr.h"
#include "acl.h"
#include "linux_inode.h"

/*
 * Check to make sure that the acl is actually valid
 */ 
int
gfs_validate_acl(struct gfs_inode *ip, const char *value, int size, int access)
{
	int err = 0;
	struct posix_acl *acl = NULL;
	struct gfs_sbd *sdp = ip->i_sbd;

	if ((current->fsuid != ip->i_di.di_uid) && !capable(CAP_FOWNER))
                return -EPERM;
	if (ip->i_di.di_type == GFS_FILE_LNK)
                return -EOPNOTSUPP;
        if (!access && ip->i_di.di_type != GFS_FILE_DIR)
                return -EACCES;
	if (!sdp->sd_args.ar_posixacls)
                return -EOPNOTSUPP;

	if (value) {
		acl = posix_acl_from_xattr(value, size);
		if (IS_ERR(acl))
                	return PTR_ERR(acl);
        	else if (acl) {
                	err = posix_acl_valid(acl);
			posix_acl_release(acl);
		}
	}
	return err;	
}

void
gfs_acl_set_mode(struct gfs_inode *ip, struct posix_acl *acl)
{
	struct inode *inode;
	mode_t mode;

	inode = gfs_iget(ip, NO_ST_CREATE);
	mode = inode->i_mode;
	posix_acl_equiv_mode(acl, &mode);
	inode->i_mode = mode;
	mark_inode_dirty(inode);
	iput(inode);
	gfs_inode_attr_out(ip);
}
	

/**
 * gfs_replace_acl - replace the value of the ea to the value of the acl
 *
 * NOTE: The new value must be the same size as the old one. 
 */
int
gfs_replace_acl(struct inode *inode, struct posix_acl *acl, int access,
		struct gfs_ea_location location)
{
	struct gfs_easet_io req;
	int err;
	int size;
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;
	void *data;

	size = posix_acl_to_xattr(acl, NULL, 0);
	GFS_ASSERT(size == GFS_EA_DATA_LEN(location.ea),
		   printk("new acl size = %d, ea size = %u\n", size,
			  GFS_EA_DATA_LEN(location.ea));
	    );
	data = gmalloc(size);
	posix_acl_to_xattr(acl, data, size);

	req.es_data = data;
	req.es_name = (access) ? GFS_POSIX_ACL_ACCESS : GFS_POSIX_ACL_DEFAULT;
	req.es_data_len = size;
	req.es_name_len =
	    (access) ? GFS_POSIX_ACL_ACCESS_LEN : GFS_POSIX_ACL_DEFAULT_LEN;
	req.es_cmd = GFS_EACMD_REPLACE;
	req.es_type = GFS_EATYPE_SYS;
	err = replace_ea(sdp, ip, location.ea, &req);
	if (!err)
		gfs_trans_add_bh(sdp, ip->i_gl, location.bh);

	kfree(data);

	return err;
}

/**
 * gfs_setacl - store the posix acl as an extended attribute
 *
 * inode->i_sem must be locked when calling this function
 */
int
gfs_setacl(struct inode *inode, int access, struct posix_acl *acl)
{
	struct gfs_easet_io req;
	int err = 0;
	int size = 0;
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;
	void *data = NULL;

	if (S_ISLNK(inode->i_mode)) {
		err = -EOPNOTSUPP;
		goto out;
	}
	if (!access && !S_ISDIR(inode->i_mode)) {
		err = -EACCES;
		goto out;
	}

	size = posix_acl_to_xattr(acl, NULL, 0);
	if (size < 0) {
		err = size;
		goto out;
	}
	data = gmalloc(size);
	if (!data) {
		err = -ENOMEM;
		goto out;
	}
	err = posix_acl_to_xattr(acl, data, size);

	req.es_data = data;
	req.es_name = (access) ? GFS_POSIX_ACL_ACCESS : GFS_POSIX_ACL_DEFAULT;
	req.es_data_len = size;
	req.es_name_len =
	    (access) ? GFS_POSIX_ACL_ACCESS_LEN : GFS_POSIX_ACL_DEFAULT_LEN;
	req.es_cmd = GFS_EACMD_SET;
	req.es_type = GFS_EATYPE_SYS;
	err = gfs_set_eattr(sdp, ip, &req, gfs_get_user_cred());

	kfree(data);

      out:
	return err;
}

/**
 * gfs_findacl - returns the requested posix acl
 *
 * this function does not log the inode. It assumes that a lock is already
 * held on it.
 */
int
gfs_findacl(struct gfs_inode *ip, int access, struct posix_acl **acl_ptr,
	    struct gfs_ea_location *location)
{
	int err = 0;
	struct gfs_sbd *sdp = ip->i_sbd;
	void *data;
	uint32_t avail_size;
	struct posix_acl *acl;

	avail_size = sdp->sd_sb.sb_bsize - sizeof (struct gfs_meta_header);
	*acl_ptr = NULL;

	if (ip->i_di.di_eattr == 0)
		goto out;

	err =
	    find_eattr(ip,
		       (access) ? GFS_POSIX_ACL_ACCESS : GFS_POSIX_ACL_DEFAULT,
		       (access) ? GFS_POSIX_ACL_ACCESS_LEN :
		       GFS_POSIX_ACL_DEFAULT_LEN, GFS_EATYPE_SYS, location);
	if (err <= 0)
		goto out;
	err = 0;
	data = gmalloc(GFS_EA_DATA_LEN(location->ea));
	if (GFS_EA_IS_UNSTUFFED(location->ea))
		err = read_unstuffed(data, ip, sdp, location->ea, avail_size,
				     gfs_ea_memcpy);
	else
		gfs_ea_memcpy(data, GFS_EA_DATA(location->ea),
			      GFS_EA_DATA_LEN(location->ea));
	if (err)
		goto out_alloc;

	acl = posix_acl_from_xattr(data, GFS_EA_DATA_LEN(location->ea));
	if (IS_ERR(acl))
		err = PTR_ERR(acl);
	else
		*acl_ptr = acl;

      out_alloc:
	kfree(data);
	if (err)
		gfs_drelse(location->bh);

      out:
	return err;
}

int
gfs_getacl(struct inode *inode, int access, struct posix_acl **acl_ptr)
{
	struct gfs_eaget_io req;
	int size;
	int err = 0;
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;
	struct gfs_holder i_gh;
	struct gfs_sbd *sdp = ip->i_sbd;
	void *data;
	struct posix_acl *acl;

	*acl_ptr = NULL;

	if (!sdp->sd_args.ar_posixacls)
		goto out;

	req.eg_name = (access) ? GFS_POSIX_ACL_ACCESS : GFS_POSIX_ACL_DEFAULT;
	req.eg_name_len =
	    (access) ? GFS_POSIX_ACL_ACCESS_LEN : GFS_POSIX_ACL_DEFAULT_LEN;
	req.eg_type = GFS_EATYPE_SYS;
	req.eg_len = NULL;
	req.eg_data = NULL;
	req.eg_data_len = 0;

	err = gfs_ea_read_permission(&req, ip, gfs_get_user_cred());
	if (err)
		goto out;

	err = gfs_glock_nq_init(ip->i_gl, LM_ST_SHARED, 0, &i_gh);
	if (err)
		goto out;

	if (ip->i_di.di_eattr == 0)
		goto out_gunlock;

	size = get_ea(sdp, ip, &req, gfs_ea_memcpy);
	if (size < 0) {
		if (size != -ENODATA)
			err = size;
		goto out_gunlock;
	}

	data = gmalloc(size);
	req.eg_data = data;
	req.eg_data_len = size;

	size = get_ea(sdp, ip, &req, gfs_ea_memcpy);
	if (size < 0) {
		err = size;
		goto out_free;
	}

	acl = posix_acl_from_xattr(data, size);
	if (IS_ERR(acl))
		err = PTR_ERR(acl);
	else
		*acl_ptr = acl;

      out_free:
	kfree(data);

      out_gunlock:
	gfs_glock_dq_uninit(&i_gh);

      out:
	return err;
}

int
gfs_setup_new_acl(struct gfs_inode *dip, unsigned int *mode, unsigned int type,
		  struct posix_acl **acl_ptr)
{
	int err = 0;
	mode_t access_mode = *mode;
	struct posix_acl *acl = NULL;
	struct gfs_ea_location location;

	*acl_ptr = NULL;

	if (type == GFS_FILE_LNK)
		goto out;

	err = gfs_findacl(dip, FALSE, &acl, &location);
	if (err)
		goto out;
	if (acl == NULL) {
		*mode = *mode & ~current->fs->umask;
		goto out;
	}
	gfs_drelse(location.bh);

	if (type == GFS_FILE_DIR) {
		*acl_ptr = acl;
		goto out;
	}

	err = posix_acl_create_masq(acl, &access_mode);
	*mode = access_mode;
	if (err > 0) {
		*acl_ptr = acl;
		err = 0;
		goto out;
	}

	posix_acl_release(acl);

      out:
	return err;
}

/**
 * gfs_init_default_acl - initializes the default acl
 *
 * NOTE: gfs_init_access_acl must be called first
 */
int
gfs_create_default_acl(struct gfs_inode *dip, struct gfs_inode *ip, void *data,
		       int size)
{
	struct gfs_easet_io req;
	struct gfs_sbd *sdp = ip->i_sbd;
	int err;
	struct gfs_ea_location avail;

	memset(&avail, 0, sizeof (struct gfs_ea_location));

	req.es_data = data;
	req.es_name = GFS_POSIX_ACL_DEFAULT;
	req.es_data_len = size;
	req.es_name_len = GFS_POSIX_ACL_DEFAULT_LEN;
	req.es_cmd = GFS_EACMD_CREATE;
	req.es_type = GFS_EATYPE_SYS;

	err = find_sys_space(dip, ip, size, &avail);
	if (err)
		goto out;

	avail.ea = prep_ea(avail.ea);
	err = write_ea(sdp, dip, ip, avail.ea, &req);
	if (!err)
		gfs_trans_add_bh(sdp, ip->i_gl, avail.bh);

	gfs_drelse(avail.bh);

      out:
	return err;
}

/**
 * gfs_init_access_acl - initialized the access acl
 *
 * NOTE: This must be the first extended attribute that is created for
 *       this inode.
 */
int
gfs_init_access_acl(struct gfs_inode *dip, struct gfs_inode *ip, void *data,
		    int size)
{
	struct gfs_easet_io req;
	int err;

	req.es_data = data;
	req.es_name = GFS_POSIX_ACL_ACCESS;
	req.es_data_len = size;
	req.es_name_len = GFS_POSIX_ACL_ACCESS_LEN;
	req.es_cmd = GFS_EACMD_CREATE;
	req.es_type = GFS_EATYPE_SYS;

	err = init_new_inode_eattr(dip, ip, &req);

	return err;
}

int
gfs_init_acl(struct gfs_inode *dip, struct gfs_inode *ip, unsigned int type,
	     struct posix_acl *acl)
{
	int err;
	void *data;
	struct gfs_sbd *sdp = ip->i_sbd;
	int size;
	struct buffer_head *dibh;

	size = posix_acl_to_xattr(acl, NULL, 0);

	data = gmalloc(size);
	posix_acl_to_xattr(acl, data, size);

	err = gfs_get_inode_buffer(ip, &dibh);
	if (err)
		goto out;

	err = gfs_init_access_acl(dip, ip, data, size);
	if (err)
		goto out_relse;

	if (type == GFS_FILE_DIR) {
		err = gfs_create_default_acl(dip, ip, data, size);
		if (err)
			goto out_relse;
	}
	gfs_trans_add_bh(sdp, ip->i_gl, dibh);
	gfs_dinode_out(&ip->i_di, (dibh)->b_data);

      out_relse:
	gfs_drelse(dibh);

      out:
	kfree(data);
	posix_acl_release(acl);
	return err;
}

int
gfs_acl_setattr(struct inode *inode)
{
	int err = 0;
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;
	struct posix_acl *acl;
	struct gfs_ea_location location;

	if (S_ISLNK(inode->i_mode))
		goto out;

	memset(&location, 0, sizeof (struct gfs_ea_location));
	err = gfs_findacl(ip, TRUE, &acl, &location);
	if (location.ea == NULL)
		goto out;

	err = posix_acl_chmod_masq(acl, inode->i_mode);
	if (err)
		goto out_release;

	err = gfs_replace_acl(inode, acl, TRUE, location);

      out_release:
	posix_acl_release(acl);
	gfs_drelse(location.bh);

      out:
	return err;
}
