/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "glock.h"

/**
 * gfs_random - Generate a random 32-bit number
 *
 * Generate a semi-crappy 32-bit pseudo-random number without using
 * floating point.
 *
 * The PRNG is from "Numerical Recipes in C" (second edition), page 284.
 *
 * Returns: a 32-bit random number
 */

uint32_t
gfs_random(void)
{
	gfs_random_number = 0x0019660D * gfs_random_number + 0x3C6EF35F;
	return gfs_random_number;
}

/**
 * hash_more_internal - hash an array of data
 * @data: the data to be hashed
 * @len: the length of data to be hashed
 * @hash: the hash from a previous call
 *
 * Take some data and convert it to a 32-bit hash.
 *
 * This is the 32-bit FNV-1a hash from:
 * http://www.isthe.com/chongo/tech/comp/fnv/
 *
 * Hash guts
 *
 * Returns: the hash
 */

static __inline__ uint32_t
hash_more_internal(const void *data, unsigned int len, uint32_t hash)
{
	unsigned char *p = (unsigned char *) data;
	unsigned char *e = p + len;
	uint32_t h = hash;

	while (p < e) {
		h ^= (uint32_t) (*p++);
		h *= 0x01000193;
	}

	return h;
}

/**
 * gfs_hash - hash an array of data
 * @data: the data to be hashed
 * @len: the length of data to be hashed
 *
 * Take some data and convert it to a 32-bit hash.
 *
 * This is the 32-bit FNV-1a hash from:
 * http://www.isthe.com/chongo/tech/comp/fnv/
 *
 * Returns: the hash
 */

uint32_t
gfs_hash(const void *data, unsigned int len)
{
	uint32_t h = 0x811C9DC5;
	h = hash_more_internal(data, len, h);
	return h;
}

/**
 * gfs_hash_more - hash an array of data
 * @data: the data to be hashed
 * @len: the length of data to be hashed
 * @hash: the hash from a previous call
 *
 * Take some data and convert it to a 32-bit hash.
 *
 * This is the 32-bit FNV-1a hash from:
 * http://www.isthe.com/chongo/tech/comp/fnv/
 *
 * This version let's you hash together discontinuous regions.
 * For example, to compute the combined hash of the memory in
 * (data1, len1), (data2, len2), and (data3, len3) you:
 *
 *   h = gfs_hash(data1, len1);
 *   h = gfs_hash_more(data2, len2, h);
 *   h = gfs_hash_more(data3, len3, h);
 *
 * Returns: the hash
 */

uint32_t
gfs_hash_more(const void *data, unsigned int len, uint32_t hash)
{
	uint32_t h;
	h = hash_more_internal(data, len, hash);
	return h;
}

/**
 * gfs_sort - Sort base array using bubble sort algorithm
 * @array: the input array
 * @num: number of elements in array
 * @size: size of each element in array
 * @compare: fxn to compare array elements (returns negative for lt, 0 for eq, and positive for gt)
 *
 * Sorts the array passed in using the compar fxn to compare elements using
 * the bubble sort algorithm
 */

void
gfs_sort(void *array, unsigned int num, unsigned int size,
	 int (*compare) (void *, void *))
{
	char buf[size];
	char *p1, *p2;
	int changed;
	unsigned int x;

	if (num > 1)
		do {
			changed = FALSE;
			p1 = (char *) array;
			p2 = (char *) array + size;

			for (x = num - 1; x--;) {
				if (compare(p1, p2) > 0) {
					memcpy(buf, p1, size);
					memcpy(p1, p2, size);
					memcpy(p2, buf, size);
					changed = TRUE;
				}

				p1 = p2;
				p2 += size;
			}
		}
		while (changed);

}

/**
 * gfs_asserti - Stop the machine
 * @assertion: the assertion that failed
 * @file: the file that called us
 * @line: the line number of the file that called us
 *
 * Don't do ENTER() and EXIT() here.
 *
 */

void
gfs_asserti(char *assertion, char *file, unsigned int line,
	    unsigned int type, void *ptr)
{
	gfs_in_panic = TRUE;

	dump_stack();

	switch (type) {
	case GFS_ASSERT_TYPE_SBD: {
		struct gfs_sbd *sdp = (struct gfs_sbd *) ptr;
		panic("GFS: Assertion failed on line %d of file %s\n"
		      "GFS: assertion: \"%s\"\n"
		      "GFS: time = %lu\n"
		      "GFS: fsid=%s\n",
		      line, file, assertion, (unsigned long)CURRENT_TIME,
		      sdp->sd_fsname);
	}

	case GFS_ASSERT_TYPE_GLOCK: {
		struct gfs_glock *gl = (struct gfs_glock *) ptr;
		struct gfs_sbd *sdp = gl->gl_sbd;
		panic("GFS: Assertion failed on line %d of file %s\n"
		      "GFS: assertion: \"%s\"\n"
		      "GFS: time = %lu\n"
		      "GFS: fsid=%s: glock = (%u, %"PRIu64")\n",
		      line, file, assertion, (unsigned long)CURRENT_TIME,
		      sdp->sd_fsname,
		      gl->gl_name.ln_type,
		      gfs_lockname2blk(sdp, &gl->gl_name));
	}

	case GFS_ASSERT_TYPE_INODE: {
		struct gfs_inode *ip = (struct gfs_inode *) ptr;
		struct gfs_sbd *sdp = ip->i_sbd;
		panic("GFS: Assertion failed on line %d of file %s\n"
		      "GFS: assertion: \"%s\"\n"
		      "GFS: time = %lu\n"
		      "GFS: fsid=%s: inode = %"PRIu64"/%"PRIu64"\n",
		      line, file, assertion, (unsigned long)CURRENT_TIME,
		      sdp->sd_fsname,
		      ip->i_num.no_formal_ino, ip->i_num.no_addr);
	}

	case GFS_ASSERT_TYPE_RGRPD: {
		struct gfs_rgrpd *rgd = (struct gfs_rgrpd *) ptr;
		struct gfs_sbd *sdp = rgd->rd_sbd;
		panic("GFS: Assertion failed on line %d of file %s\n"
		      "GFS: assertion: \"%s\"\n"
		      "GFS: time = %lu\n"
		      "GFS: fsid=%s: RG = %"PRIu64"\n",
		      line, file, assertion, (unsigned long)CURRENT_TIME,
		      sdp->sd_fsname,
		      rgd->rd_ri.ri_addr);
	}

	default:
		panic("GFS: Assertion failed on line %d of file %s\n"
		      "GFS: assertion: \"%s\"\n"
		      "GFS: time = %lu\n",
		      line, file, assertion, (unsigned long)CURRENT_TIME);
	}
}

/**
 * bitch_about - 
 * @sdp: the filesystem
 * @last: the last time we bitched
 * @about:
 *
 */

void
bitch_about(struct gfs_sbd *sdp, unsigned long *last, char *about)
{
	if (time_after_eq(jiffies, *last + sdp->sd_tune.gt_complain_secs * HZ)) {
		printk("GFS: fsid=%s: %s by program \"%s\"\n",
		       sdp->sd_fsname, about, current->comm);
		*last = jiffies;
	}
}

void *
gmalloc(unsigned int size)
{
	void *p;
	RETRY_MALLOC(p = kmalloc(size, GFP_KERNEL), p);
	return p;
}
