/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "bmap.h"
#include "dio.h"
#include "dir.h"
#include "eattr.h"
#include "file.h"
#include "glock.h"
#include "glops.h"
#include "inode.h"
#include "ioctl.h"
#include "quota.h"
#include "rgrp.h"
#include "super.h"
#include "trans.h"

/**
 * gfs_add_bh_to_ub - copy a buffer up to user space
 * @ub: the structure representing where to copy
 * @bh: the buffer
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
gfs_add_bh_to_ub(struct gfs_user_buffer *ub, struct buffer_head *bh)
{
	uint64_t blkno = (uint64_t) ((bh)->b_blocknr);
	int error = 0;

	if (ub->ub_offset + sizeof (uint64_t) + (bh)->b_size > ub->ub_size) {
		error = -ENOBUFS;
		goto out;
	}

	error =
	    (copy_to_user(ub->ub_data + ub->ub_offset, &blkno,
			  sizeof (uint64_t))) ? -EFAULT : 0;
	if (!error)
		error =
		    (copy_to_user(ub->ub_data + ub->ub_offset +
				  sizeof (uint64_t), (bh)->b_data,
				  (bh)->b_size)) ? -EFAULT : 0;

	if (!error)
		ub->ub_offset += sizeof (uint64_t) + (bh)->b_size;

      out:

	return error;
}

/**
 * get_meta - Read out all the metadata for a file
 * @ip: the file
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
get_meta(struct gfs_inode *ip, void *arg)
{
	struct gfs_holder i_gh;
	struct gfs_user_buffer ub;
	int error = 0;

	error =
	    (copy_from_user(&ub, arg, sizeof (struct gfs_user_buffer))) ?
	    -EFAULT : 0;
	if (error)
		goto out;

	error = gfs_glock_nq_init(ip->i_gl, LM_ST_SHARED, LM_FLAG_ANY, &i_gh);
	if (error)
		goto out;

	error = gfs_get_file_meta(ip, &ub);
	if (error)
		goto out_gunlock;

	if (ip->i_di.di_type == GFS_FILE_DIR &&
	    (ip->i_di.di_flags & GFS_DIF_EXHASH)) {
		error = gfs_get_dir_meta(ip, &ub);
		if (error)
			goto out_gunlock;
	}

	if (ip->i_di.di_eattr) {
		error = gfs_get_eattr_meta(ip, &ub);
		if (error)
			goto out_gunlock;
	}

	error =
	    (copy_to_user(arg, &ub, sizeof (struct gfs_user_buffer))) ? -EFAULT
	    : 0;

      out_gunlock:
	gfs_glock_dq_uninit(&i_gh);

      out:

	return error;
}

/**
 * file_stat - return the struct gfs_dinode of a file to user space
 * @ip: the inode
 * @arg: where to copy to
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
file_stat(struct gfs_inode *ip, void *arg)
{
	struct gfs_holder i_gh;
	struct gfs_dinode di;
	int error;

	error = gfs_glock_nq_init(ip->i_gl, LM_ST_SHARED, LM_FLAG_ANY, &i_gh);
	if (error)
		goto out;

	memcpy(&di, &ip->i_di, sizeof (struct gfs_dinode));

	gfs_glock_dq_uninit(&i_gh);

	error =
	    (copy_to_user(arg, &di, sizeof (struct gfs_dinode))) ? -EFAULT : 0;

      out:

	return error;
}

/**
 * do_get_super - Dump the superblock into a buffer
 * @sb: The superblock
 * @ptr: The buffer pointer
 *
 * Returns: 0 or error code
 */

static int
do_get_super(struct gfs_sbd *sdp, void *arg)
{
	struct gfs_holder sb_gh;
	struct gfs_sb sb;
	struct buffer_head *bh;
	int error;

	error = gfs_glock_nq_num(sdp,
				 GFS_SB_LOCK, &gfs_meta_glops,
				 LM_ST_SHARED, 0, &sb_gh);
	if (error)
		goto out;

	error =
	    gfs_dread(sdp, GFS_SB_ADDR >> sdp->sd_fsb2bb_shift, sb_gh.gh_gl,
		      DIO_START | DIO_WAIT, &bh);
	if (error) {
		gfs_glock_dq_uninit(&sb_gh);
		goto out;
	}

	gfs_sb_in(&sb, (bh)->b_data);

	gfs_drelse(bh);

	gfs_glock_dq_uninit(&sb_gh);

	error = (copy_to_user(arg, &sb, sizeof (struct gfs_sb))) ? -EFAULT : 0;

      out:

	return error;
}

/**
 * jt2ip - convert the file type in a jio struct to the right hidden ip
 * @sdp: the filesystem
 * @jt: the gfs_jio_structure
 *
 * Returns: The inode structure for the correct hidden file
 */

static struct gfs_inode *
jt2ip(struct gfs_sbd *sdp, struct gfs_jio *jt)
{
	struct gfs_inode *ip = NULL;

	switch (jt->jio_file) {
	case GFS_HIDDEN_JINDEX:
		ip = sdp->sd_jiinode;
		break;

	case GFS_HIDDEN_RINDEX:
		ip = sdp->sd_riinode;
		break;

	case GFS_HIDDEN_QUOTA:
		ip = sdp->sd_qinode;
		break;

	case GFS_HIDDEN_LICENSE:
		ip = sdp->sd_linode;
		break;
	}

	return ip;
}

/**
 * jread_ioctl - Read from a journaled data file via ioctl
 * @sdp: the filesystem
 * @arg: The argument from ioctl
 *
 * Returns: Amount of data copied or error
 */

static int
jread_ioctl(struct gfs_sbd *sdp, void *arg)
{
	struct gfs_jio jt;
	struct gfs_inode *ip;
	struct gfs_holder i_gh;
	int error;

	error =
	    (copy_from_user(&jt, arg, sizeof (struct gfs_jio))) ? -EFAULT : 0;
	if (error)
		goto out;

	ip = jt2ip(sdp, &jt);
	if (!ip) {
		error = -EINVAL;
		goto out;
	}

	GFS_ASSERT_INODE(gfs_is_jdata(ip), ip,);

	if (!access_ok(VERIFY_WRITE, jt.jio_data, jt.jio_size)) {
		error = -EFAULT;
		goto out;
	}

	error = gfs_glock_nq_init(ip->i_gl, LM_ST_SHARED, 0, &i_gh);
	if (error)
		goto out;

	error =
	    gfs_readi(ip, jt.jio_data, jt.jio_offset, jt.jio_size,
		      gfs_copy2user);
	if (error >= 0) {
		jt.jio_count = error;
		error = 0;
	}

	gfs_glock_dq_uninit(&i_gh);

	if (!error)
		error =
		    (copy_to_user(arg, &jt, sizeof (struct gfs_jio))) ? -EFAULT
		    : 0;

      out:

	return error;
}

/**
 * jwrite_ioctl - Write to a journaled file via ioctl
 * @sdp: the filesystem
 * @arg: The argument from ioctl
 *
 * Returns: Amount of data copied or error
 */

static int
jwrite_ioctl(struct gfs_sbd *sdp, void *arg)
{
	struct gfs_jio jt;
	struct gfs_inode *ip;
	struct gfs_alloc *al = NULL;
	struct gfs_holder i_gh;
	unsigned int data_blocks, ind_blocks;
	int alloc_required;
	int error;

	error =
	    (copy_from_user(&jt, arg, sizeof (struct gfs_jio))) ? -EFAULT : 0;
	if (error)
		goto out;

	ip = jt2ip(sdp, &jt);
	if (!ip) {
		error = -EINVAL;
		goto out;
	}

	GFS_ASSERT_INODE(gfs_is_jdata(ip), ip,);

	if (!access_ok(VERIFY_READ, jt.jio_data, jt.jio_size)) {
		error = -EFAULT;
		goto out;
	}

	gfs_write_calc_reserv(ip, jt.jio_size, &data_blocks, &ind_blocks);

	error =
	    gfs_glock_nq_init(ip->i_gl, LM_ST_EXCLUSIVE,
			      LM_FLAG_PRIORITY | GL_SYNC, &i_gh);
	if (error)
		goto out;

	error =
	    gfs_write_alloc_required(ip, jt.jio_offset, jt.jio_size,
				     &alloc_required);
	if (error)
		goto out_gunlock;

	if (alloc_required) {
		al = gfs_get_alloc(ip);

		error =
		    gfs_hold_quota_structs(ip, NO_QUOTA_CHANGE,
					   NO_QUOTA_CHANGE);
		if (error)
			goto out_alloc;

		al->al_metablks = ind_blocks + data_blocks;
		al->al_trans = TRANS_JWRITE;

		error = gfs_inplace_reserv(ip);
		if (error)
			goto out_qs;

		/*  Trans may require:
		   All blocks for a RG bitmap, all the "data" blocks, whatever indirect blocks we
		   need, a modified dinode, and a quota change  */

		error =
		    gfs_trans_begin(sdp, TRANS_JWRITE,
				    1 + al->al_rgd->rd_ri.ri_length +
				    ind_blocks + data_blocks, 1);
		if (error)
			goto out_relse;
	} else {
		/*  Trans may require:
		   All the "data" blocks and a modified dinode.  */

		error = gfs_trans_begin(sdp, TRANS_JWRITE, 1 + data_blocks, 0);
		if (error)
			goto out_relse;
	}

	error =
	    gfs_writei(ip, jt.jio_data, jt.jio_offset, jt.jio_size,
		       gfs_copy_from_user);
	if (error >= 0) {
		jt.jio_count = error;
		error = 0;
	}

	gfs_trans_end(sdp);

      out_relse:
	if (alloc_required) {
		GFS_ASSERT_INODE(error || al->al_metaalloced, ip,);
		gfs_inplace_relse(ip);
	}

      out_qs:
	if (alloc_required)
		gfs_unhold_quota_structs(ip);

      out_alloc:
	if (alloc_required)
		gfs_put_alloc(ip);

      out_gunlock:
	ip->i_gl->gl_vn++;
	gfs_glock_dq_uninit(&i_gh);

	if (!error)
		error =
		    (copy_to_user(arg, &jt, sizeof (struct gfs_jio))) ? -EFAULT
		    : 0;

      out:
	return error;
}

/**
 * jstat_ioctl - Stat to a journaled file via ioctl
 * @sdp: the filesystem
 * @arg: The argument from ioctl
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
jstat_ioctl(struct gfs_sbd *sdp, void *arg)
{
	struct gfs_jio jt;
	struct gfs_inode *ip;
	struct gfs_holder i_gh;
	struct gfs_dinode di;
	int error;

	error =
	    (copy_from_user(&jt, arg, sizeof (struct gfs_jio))) ? -EFAULT : 0;
	if (error)
		goto out;

	ip = jt2ip(sdp, &jt);
	if (!ip) {
		error = -EINVAL;
		goto out;
	}

	if (jt.jio_size < sizeof (struct gfs_dinode)) {
		error = -EINVAL;
		goto out;
	}

	error = gfs_glock_nq_init(ip->i_gl, LM_ST_SHARED, LM_FLAG_ANY, &i_gh);
	if (error)
		goto out;

	memcpy(&di, &ip->i_di, sizeof (struct gfs_dinode));

	gfs_glock_dq_uninit(&i_gh);

	error =
	    (copy_to_user(jt.jio_data, &di, sizeof (struct gfs_dinode))) ?
	    -EFAULT : 0;

      out:

	return error;
}

/**
 * jtrunc_ioctl - Truncate to a journaled file via ioctl
 * @sdp: the filesystem
 * @arg: The argument from ioctl
 * @cred: the user's credentials
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
jtrunc_ioctl(struct gfs_sbd *sdp, void *arg, int *cred)
{
	struct gfs_jio jt;
	struct gfs_inode *ip;
	struct gfs_holder i_gh;
	int error;

	error =
	    (copy_from_user(&jt, arg, sizeof (struct gfs_jio))) ? -EFAULT : 0;
	if (error)
		goto out;

	ip = jt2ip(sdp, &jt);
	if (!ip) {
		error = -EINVAL;
		goto out;
	}

	error = gfs_glock_nq_init(ip->i_gl, LM_ST_EXCLUSIVE, GL_SYNC, &i_gh);
	if (error)
		goto out;

	error = gfs_truncatei(ip, jt.jio_offset, NULL, cred);

	ip->i_gl->gl_vn++;
	gfs_glock_dq_uninit(&i_gh);

      out:

	return error;
}

/**
 * lock_dump - copy out info about the GFS' lock space
 * @sdp: the filesystem
 * @arg: a pointer to a struct gfs_user_buffer in user space
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
lock_dump(struct gfs_sbd *sdp, void *arg)
{
	struct gfs_user_buffer ub;
	int error;

	error =
	    (copy_from_user(&ub, arg, sizeof (struct gfs_user_buffer))) ?
	    -EFAULT : 0;
	if (error)
		goto out;

	error = gfs_dump_lockstate(sdp, &ub);
	if (error)
		goto out;

	error =
	    (copy_to_user(arg, &ub, sizeof (struct gfs_user_buffer))) ? -EFAULT
	    : 0;

      out:

	return error;
}

/**
 * stat_gfs_ioctl - Do a GFS specific statfs
 * @sdp: the filesystem
 * @arg: the struct gfs_usage structure
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
stat_gfs_ioctl(struct gfs_sbd *sdp, void *arg)
{
	struct gfs_usage usage;
	int error = 0;

	error = gfs_stat_gfs(sdp, &usage, TRUE);
	if (!error)
		error =
		    (copy_to_user(arg, &usage, sizeof (struct gfs_usage))) ?
		    -EFAULT : 0;

	return error;
}

/**
 * reclaim_ioctl - ioctl called to perform metadata reclaimation
 * @sdp: the filesystem
 * @arg: a pointer to a struct gfs_reclaim_stats in user space
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
reclaim_ioctl(struct gfs_sbd *sdp, void *arg)
{
	struct gfs_reclaim_stats stats;
	int error;

	memset(&stats, 0, sizeof (struct gfs_reclaim_stats));

	error = gfs_reclaim_metadata(sdp, &stats);
	if (!error)
		error =
		    (copy_to_user((struct gfs_reclaim_stats *) arg, &stats,
				  sizeof (struct gfs_reclaim_stats))) ? -EFAULT
		    : 0;

	return error;
}

/**
 * get_tune - pass the current tuneable parameters up to user space
 * @sdp: the filesystem
 * @arg: a pointer to a struct gfs_tune in user space
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
get_tune(struct gfs_sbd *sdp, void *arg)
{
	int error;

	error =
	    (copy_to_user((struct gfs_tune *) arg, &sdp->sd_tune,
			  sizeof (struct gfs_tune))) ? -EFAULT : 0;

	return error;
}

/**
 * set_tune - replace the current tuneable parameters with a set from user space
 * @sdp: the filesystem
 * @arg: a pointer to a struct gfs_tune in user space
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
set_tune(struct gfs_sbd *sdp, void *arg)
{
	struct gfs_tune gt;
	int error = 0;

	error =
	    (copy_from_user
	     (&gt, (struct gfs_tune *) arg,
	      sizeof (struct gfs_tune))) ? -EFAULT : 0;
	if (!error) {
		if (gt.gt_tune_version != GFS_TUNE_VERSION) {
			printk
			    ("GFS: fsid=%s: invalid version of tuneable parameters\n",
			     sdp->sd_fsname);
			error = -EINVAL;
		} else
			memcpy(&sdp->sd_tune, &gt, sizeof (struct gfs_tune));
	}

	return error;
}

/**
 * gfs_set_flag - set/clear a flag on an inode
 * @ip: the inode
 * @cmd: GFS_SET_FLAG or GFS_CLEAR_FLAG
 * @arg: the flag to change (in user space)
 * @cred: the credentials of the caller
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
gfs_set_flag(struct gfs_inode *ip, unsigned int cmd, void *arg, int *cred)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_holder i_gh;
	struct buffer_head *dibh;
	uint32_t flag;
	int error;

	error = (copy_from_user(&flag, arg, sizeof (uint32_t))) ? -EFAULT : 0;
	if (error)
		goto out;

	error = gfs_glock_nq_init(ip->i_gl, LM_ST_EXCLUSIVE, 0, &i_gh);
	if (error)
		goto out;

	error = gfs_permission(ip, GFS_IS_FOWNER, cred);
	if (error)
		goto out_gunlock;

	switch (flag) {
	case GFS_DIF_EXHASH:
	case GFS_DIF_UNUSED:
	case GFS_DIF_EA_INDIRECT:
		error = -EINVAL;
		goto out_gunlock;

	case GFS_DIF_JDATA:
		if (ip->i_di.di_type != GFS_FILE_REG || ip->i_di.di_size) {
			error = -EINVAL;
			goto out_gunlock;
		}
		break;

	case GFS_DIF_DIRECTIO:
		if (ip->i_di.di_type != GFS_FILE_REG) {
			error = -EINVAL;
			goto out_gunlock;
		}
		break;

	case GFS_DIF_IMMUTABLE:
	case GFS_DIF_APPENDONLY:
	case GFS_DIF_NOATIME:
	case GFS_DIF_SYNC:
		/*  FixMe!!!  */
		error = -ENOSYS;
		goto out_gunlock;

	case GFS_DIF_INHERIT_DIRECTIO:
	case GFS_DIF_INHERIT_JDATA:
		if (ip->i_di.di_type != GFS_FILE_DIR) {
			error = -EINVAL;
			goto out_gunlock;
		}
		break;

	default:
		error = -EINVAL;
		goto out_gunlock;
	}

	error = gfs_trans_begin(sdp, TRANS_SET_FLAG, 1, 0);
	if (error)
		goto out_gunlock;

	error = gfs_get_inode_buffer(ip, &dibh);
	if (error)
		goto out_trans_end;

	if (cmd == GFS_SET_FLAG)
		ip->i_di.di_flags |= flag;
	else
		ip->i_di.di_flags &= ~flag;

	gfs_trans_add_bh(sdp, ip->i_gl, dibh);
	gfs_dinode_out(&ip->i_di, (dibh)->b_data);

	gfs_drelse(dibh);

      out_trans_end:
	gfs_trans_end(sdp);

      out_gunlock:
	gfs_glock_dq_uninit(&i_gh);

      out:

	return error;
}

/**
 * get_counters - return usage counters to user space
 * @sdp: the filesystem
 * @arg: the counter structure to fill
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
get_counters(struct gfs_sbd *sdp, void *arg)
{
	struct gfs_counters gc;
	int error;

	memset(&gc, 0, sizeof (struct gfs_counters));
	gc.gc_counters_version = GFS_COUNTERS_VERSION;
	gc.gc_locks = atomic_read(&sdp->sd_glock_count);
	gc.gc_locks_held = atomic_read(&sdp->sd_glock_held_count);
	gc.gc_inodes = atomic_read(&sdp->sd_inode_count);
	gc.gc_unlinked_inodes = atomic_read(&sdp->sd_unlinked_ic_count);
	gc.gc_quota_ids = atomic_read(&sdp->sd_quota_count);
	gc.gc_iclog_buffers = sdp->sd_log_buffers;
	gc.gc_log_seg_free = sdp->sd_log_seg_free;
	gc.gc_log_seg_total = sdp->sd_jdesc.ji_nsegment;
	gc.gc_mh_cache_count = atomic_read(&sdp->sd_mh_cache_count);
	gc.gc_dependencies = atomic_read(&sdp->sd_dependencies);
	gc.gc_reclaim_count = atomic_read(&sdp->sd_reclaim_count);
	gc.gc_log_wraps = sdp->sd_log_wrap;
	gc.gc_fh2dentry_misses = atomic_read(&sdp->sd_fh2dentry_misses);
	gc.gc_reclaimed = atomic_read(&sdp->sd_reclaimed);
	gc.gc_glock_nq_calls = atomic_read(&sdp->sd_glock_nq_calls);
	gc.gc_glock_dq_calls = atomic_read(&sdp->sd_glock_dq_calls);
	gc.gc_glock_prefetch_calls = atomic_read(&sdp->sd_glock_prefetch_calls);
	gc.gc_lm_lock_calls = atomic_read(&sdp->sd_lm_lock_calls);
	gc.gc_lm_unlock_calls = atomic_read(&sdp->sd_lm_unlock_calls);
	gc.gc_lm_callbacks = atomic_read(&sdp->sd_lm_callbacks);

	error =
	    (copy_to_user(arg, &gc, sizeof (struct gfs_counters))) ? -EFAULT :
	    0;
	if (!error) {
		atomic_set(&sdp->sd_fh2dentry_misses, 0);
		atomic_set(&sdp->sd_reclaimed, 0);
		atomic_set(&sdp->sd_glock_nq_calls, 0);
		atomic_set(&sdp->sd_glock_dq_calls, 0);
		atomic_set(&sdp->sd_glock_prefetch_calls, 0);
		atomic_set(&sdp->sd_lm_lock_calls, 0);
		atomic_set(&sdp->sd_lm_unlock_calls, 0);
		atomic_set(&sdp->sd_lm_callbacks, 0);
	}

	return error;
}

/**
 * gfs_ioctli - filesystem independent ioctl function
 * @ip: the inode the ioctl was on
 * @cmd: the ioctl number
 * @arg: the argument (still in user space)
 * @cred: the user's credentials
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
gfs_ioctli(struct gfs_inode *ip, unsigned int cmd, void *arg, int *cred)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	int error = 0;

	switch (cmd) {
	case GFS_STACK_PRINT:
#ifdef DEBUG_STACK
		gfs_stack_print();
		break;
#else
		printk("GFS: stack tracking not compiled in\n");
		error = -ENOSYS;
		break;
#endif				/*  DEBUG_STACK  */

	case GFS_GET_META:
		error = get_meta(ip, arg);
		break;

	case GFS_FILE_STAT:
		error = file_stat(ip, arg);
		break;

	case GFS_SHRINK:
		error = gfs_permission(ip, GFS_MAY_ADMIN, cred);
		if (error)
			break;
		gfs_clear_ghash(sdp, FALSE);
		break;

	case GFS_GET_ARGS:
		error =
		    (copy_to_user(arg, &sdp->sd_args, sizeof (struct gfs_args)))
		    ? -EFAULT : 0;
		break;

	case GFS_GET_LOCKSTRUCT:
		error =
		    (copy_to_user(arg, &sdp->sd_lockstruct,
				  sizeof (struct lm_lockstruct))) ? -EFAULT : 0;
		break;

	case GFS_GET_SUPER:
		error = do_get_super(sdp, arg);
		break;

	case GFS_JREAD:
		error = gfs_permission(ip, GFS_MAY_ADMIN, cred);
		if (error)
			break;
		error = jread_ioctl(sdp, arg);
		break;

	case GFS_JWRITE:
		error = gfs_permission(ip, GFS_MAY_ADMIN, cred);
		if (error)
			break;
		error = jwrite_ioctl(sdp, arg);
		break;

	case GFS_JSTAT:
		error = jstat_ioctl(sdp, arg);
		break;

	case GFS_JTRUNC:
		error = gfs_permission(ip, GFS_MAY_ADMIN, cred);
		if (error)
			break;
		error = jtrunc_ioctl(sdp, arg, cred);
		break;

	case GFS_LOCK_DUMP:
		error = gfs_permission(ip, GFS_MAY_ADMIN, cred);
		if (error)
			break;
		error = lock_dump(sdp, arg);
		break;

	case GFS_STATGFS:
		error = stat_gfs_ioctl(sdp, arg);
		break;

	case GFS_FREEZE:
		error = gfs_permission(ip, GFS_MAY_ADMIN, cred);
		if (error)
			break;
		error = gfs_freeze_fs(sdp);
		break;

	case GFS_UNFREEZE:
		error = gfs_permission(ip, GFS_MAY_ADMIN, cred);
		if (error)
			break;
		gfs_unfreeze_fs(sdp);
		break;

	case GFS_RECLAIM_METADATA:
		error = gfs_permission(ip, GFS_MAY_ADMIN, cred);
		if (error)
			break;
		error = reclaim_ioctl(sdp, arg);
		break;

	case GFS_QUOTA_SYNC:
		error = gfs_permission(ip, GFS_MAY_ADMIN, cred);
		if (error)
			break;
		error = gfs_quota_sync(sdp);
		break;

	case GFS_QUOTA_REFRESH:
		error = gfs_permission(ip, GFS_MAY_ADMIN, cred);
		if (error)
			break;
		error = gfs_quota_refresh(sdp, arg);
		break;

	case GFS_QUOTA_GET:
		/*  Permissions handled later  */
		error = gfs_quota_get(sdp, arg, cred);
		break;

	case GFS_GET_TUNE:
		error = get_tune(sdp, arg);
		break;

	case GFS_SET_TUNE:
		error = gfs_permission(ip, GFS_MAY_ADMIN, cred);
		if (error)
			break;
		error = set_tune(sdp, arg);
		break;

	case GFS_EATTR_GET:
		/*  Permissions handled later  */
		error = gfs_get_eattr_ioctl(sdp, ip, arg, cred);
		break;

	case GFS_EATTR_SET:
		/*  Permissions handled later  */
		error = gfs_set_eattr_ioctl(sdp, ip, arg, cred);
		break;

	case GFS_WHERE_ARE_YOU:
		{
			unsigned int x = GFS_MAGIC;
			error =
			    (copy_to_user(arg, &x, sizeof (unsigned int))) ?
			    -EFAULT : 0;
		}
		break;

	case GFS_SET_FLAG:
	case GFS_CLEAR_FLAG:
		/*  Permissions handled later  */
		error = gfs_set_flag(ip, cmd, arg, cred);
		break;

	case GFS_GET_COUNTERS:
		error = get_counters(sdp, arg);
		break;

	case GFS_FILE_FLUSH:
		gfs_glock_force_drop(ip->i_gl);
		break;

	default:
		error = -ENOTTY;
		break;
	}

	GFS_ASSERT_INODE(error <= 0, ip,
			 printk("cmd = %u, error = %d\n", cmd, error);
	    );

	return error;
}
