/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "dio.h"
#include "glock.h"
#include "glops.h"
#include "inode.h"
#include "linux_inode.h"
#include "log.h"
#include "linux_page.h"
#include "recovery.h"
#include "rgrp.h"
#include "super.h"

/**
 * sync_meta - sync out the metadata for this glock
 * @gl: the glock
 * @flags: GOF_*
 *
 */

static void
sync_meta(struct gfs_glock *gl, int flags)
{
	if (!(flags & DIO_METADATA))
		return;

	if (test_bit(GLF_DIRTY, &gl->gl_flags)) {
		gfs_log_flush_glock(gl);
		gfs_sync_buf(gl, flags | DIO_CHECK);
	}

	clear_bit(GLF_DIRTY, &gl->gl_flags);
	clear_bit(GLF_SYNC, &gl->gl_flags);
}

/**
 * inval_meta - invalidate the metadata for this glock
 * @gl: the glock
 * @flags: 
 *
 */

static void
inval_meta(struct gfs_glock *gl, int flags)
{
	if (!(flags & DIO_METADATA))
		return;

	gfs_inval_buf(gl);
	gl->gl_vn++;
}

/**
 * demote_ok_meta - check to see if it's ok to unlock a glock
 * @gl: the glock
 *
 * Returns: TRUE if it's ok
 */

static int
demote_ok_meta(struct gfs_glock *gl)
{
	int demote;
	demote = (gl->gl_arch.gl_aspace->i_mapping->nrpages) ? FALSE : TRUE;
	return demote;
}

/**
 * xmote_inode_th - promote/demote a glock
 * @gl: the glock
 * @state: the requested state
 * @flags: the flags passed into gfs_glock()
 *
 */

static void
xmote_inode_th(struct gfs_glock *gl, unsigned int state, int flags)
{
	if (gl->gl_state != LM_ST_UNLOCKED)
		gfs_inval_pte(gl);
	gfs_glock_xmote_th(gl, state, flags);
}

/**
 * xmote_inode_bh - promote/demote a glock
 * @gl: the glock
 *
 * This will be really broken when (no_formal_ino != no_addr)
 *
 */

static void
xmote_inode_bh(struct gfs_glock *gl)
{
	struct gfs_sbd *sdp = gl->gl_sbd;
	struct gfs_holder *gh = gl->gl_req_gh;
	struct buffer_head *bh;
	uint64_t block;
	int error;

	if (gl->gl_state != LM_ST_UNLOCKED
	    && (!gh || !(gh->gh_flags & GL_SKIP))) {
		block = gfs_lockname2blk(sdp, &gl->gl_name);

		error = gfs_dread(sdp, block, gl, DIO_START, &bh);
		if (!error)
			gfs_drelse(bh);
	}
}

/**
 * drop_inode_th - unlock a glock
 * @gl: the glock
 *
 */

static void
drop_inode_th(struct gfs_glock *gl)
{
	gfs_inval_pte(gl);
	gfs_glock_drop_th(gl);
}

/**
 * sync_inode - Sync the dirty data for a inode glock
 * @gl: the glock
 * @flags: DIO_INVISIBLE or 0
 *
 */

static void
sync_inode(struct gfs_glock *gl, int flags)
{
	int meta = (flags & DIO_METADATA);
	int data = (flags & DIO_DATA);

	if (test_bit(GLF_DIRTY, &gl->gl_flags)) {
		if (meta) {
			gfs_log_flush_glock(gl);
			gfs_sync_buf(gl, flags | DIO_CHECK);
		}
		if (data)
			gfs_sync_page(gl, flags | DIO_CHECK);
	}

	if (meta && data) {
		if (!(flags & DIO_INVISIBLE))
			clear_bit(GLF_DIRTY, &gl->gl_flags);
		clear_bit(GLF_SYNC, &gl->gl_flags);
	}
}

/**
 * inval_inode - prepare a inode glock to be released
 * @gl: the glock
 * @flags: 
 *
 */

static void
inval_inode(struct gfs_glock *gl, int flags)
{
	struct gfs_inode *ip = (struct gfs_inode *) gl->gl_object;
	int meta = (flags & DIO_METADATA);
	int data = (flags & DIO_DATA);

	if (meta) {
		if (ip)
			GFS_ASSERT_INODE(!ip->i_arch.i_cache[0], ip,);
		gfs_inval_buf(gl);
		gl->gl_vn++;
	}
	if (data)
		gfs_inval_page(gl);
}

/**
 * demote_ok_inode - check to see if it's ok to unlock a glock
 * @gl: the glock
 *
 * Returns: TRUE if it's ok
 */

static int
demote_ok_inode(struct gfs_glock *gl)
{
	struct gfs_sbd *sdp = gl->gl_sbd;
	int demote = FALSE;

	if (!gl->gl_object && !gl->gl_arch.gl_aspace->i_mapping->nrpages)
		demote = TRUE;
	else if (!sdp->sd_args.ar_localcaching &&
		 time_after_eq(jiffies,
			       gl->gl_stamp + sdp->sd_tune.gt_demote_secs * HZ))
		demote = TRUE;

	return demote;
}

/**
 * lock_inode - operation done after an inode lock is locked by a process
 * @gl: the glock
 * @flags: the flags passed into gfs_glock()
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
lock_inode(struct gfs_glock *gl, int flags)
{
	struct gfs_inode *ip = (struct gfs_inode *) gl->gl_object;
	int error = 0;

	if ((flags & GL_DATA_B) &&
	    test_bit(GLF_DID_DIRECT_WRITE, &gl->gl_flags)) {
		gfs_sync_page(gl, 0);
		gfs_inval_page(gl);
	} else if ((flags & (GL_DATA_DR | GL_DATA_DW)) &&
		   !test_bit(GLF_DID_DIRECT_READ, &gl->gl_flags) &&
		   !test_bit(GLF_DID_DIRECT_WRITE, &gl->gl_flags))
		gfs_sync_page(gl, 0);

	if (flags & GL_DATA_DR)
		set_bit(GLF_DID_DIRECT_READ, &gl->gl_flags);
	else if (flags & GL_DATA_DW)
		set_bit(GLF_DID_DIRECT_WRITE, &gl->gl_flags);
	else if (flags & GL_DATA_B) {
		clear_bit(GLF_DID_DIRECT_READ, &gl->gl_flags);
		clear_bit(GLF_DID_DIRECT_WRITE, &gl->gl_flags);
	}

	if (ip && ip->i_vn != gl->gl_vn) {
		error = gfs_copyin_dinode(ip);
		if (!error)
			gfs_inode_attr_in(ip);
	}

	return error;
}

/**
 * unlock_inode - operation done before an inode lock is unlocked by a process
 * @gl: the glock
 * @flags: the flags passed into gfs_gunlock()
 *
 */

static void
unlock_inode(struct gfs_glock *gl, int flags)
{
	struct gfs_inode *ip = (struct gfs_inode *) gl->gl_object;

	if (ip && test_bit(GLF_DIRTY, &gl->gl_flags))
		gfs_inode_attr_in(ip);

	if (ip)
		gfs_flush_meta_cache(ip);

	return;
}

/**
 * xmote_rgrp_th - promote/demote a glock
 * @gl: the glock
 * @state: the requested state
 * @flags: the flags passed into gfs_glock()
 *
 */

static void
xmote_rgrp_th(struct gfs_glock *gl, unsigned int state, int flags)
{
	struct gfs_rgrpd *rgd = (struct gfs_rgrpd *) gl->gl_object;

	GFS_ASSERT_GLOCK(rgd && gl->gl_lvb, gl,);

	if (test_and_clear_bit(RDF_RB_DIRTY, &rgd->rd_flags)) {
		rgd->rd_rb.rb_owner = 0;
		gfs_rgrp_lvb_out(&rgd->rd_rb, gl->gl_lvb);
	}

	gfs_zap_mh_cache(rgd);

	gfs_glock_xmote_th(gl, state, flags);

	return;
}

/**
 * xmote_rgrp_bh - promote/demote a glock
 * @gl: the glock
 *
 */

static void
xmote_rgrp_bh(struct gfs_glock *gl)
{
	struct gfs_rgrpd *rgd = (struct gfs_rgrpd *) gl->gl_object;

	GFS_ASSERT_GLOCK(rgd && gl->gl_lvb, gl,);

	if (gl->gl_state != LM_ST_UNLOCKED && rgd->rd_rb_vn != gl->gl_vn) {
		gfs_rgrp_lvb_in(&rgd->rd_rb, gl->gl_lvb);
		rgd->rd_rb_vn = gl->gl_vn;
	}

	return;
}

/**
 * drop_rgrp_th - unlock a glock
 * @gl: the glock
 *
 */

static void
drop_rgrp_th(struct gfs_glock *gl)
{
	struct gfs_rgrpd *rgd = (struct gfs_rgrpd *) gl->gl_object;

	GFS_ASSERT_GLOCK(rgd && gl->gl_lvb, gl,);

	if (test_and_clear_bit(RDF_RB_DIRTY, &rgd->rd_flags)) {
		rgd->rd_rb.rb_owner = 0;
		gfs_rgrp_lvb_out(&rgd->rd_rb, gl->gl_lvb);
	}

	gfs_zap_mh_cache(rgd);

	gfs_glock_drop_th(gl);

	return;
}

/**
 * demote_ok_rgrp - check to see if it's ok to unlock a glock
 * @gl: the glock
 *
 * Returns: TRUE if it's ok
 */

static int
demote_ok_rgrp(struct gfs_glock *gl)
{
	struct gfs_rgrpd *rgd = (struct gfs_rgrpd *) gl->gl_object;
	int demote = TRUE;

	if (gl->gl_arch.gl_aspace->i_mapping->nrpages)
		demote = FALSE;
	else if (rgd && !list_empty(&rgd->rd_mh_cache))	/*  Don't bother with lock here  */
		demote = FALSE;

	return demote;
}

/**
 * lock_rgrp - operation done after an rgrp lock is locked by a process
 * @gl: the glock
 * @flags: the flags passed into gfs_glock()
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
lock_rgrp(struct gfs_glock *gl, int flags)
{
	struct gfs_rgrpd *rgd = (struct gfs_rgrpd *) gl->gl_object;
	int error = 0;

	GFS_ASSERT_GLOCK(rgd && gl->gl_lvb, gl,);

	if (!(flags & GL_SKIP))
		error = gfs_rgrp_read(rgd);

	return error;
}

/**
 * unlock_rgrp - operation done before an rgrp lock is unlocked by a process
 * @gl: the glock
 * @flags: the flags passed into gfs_gunlock()
 *
 */

static void
unlock_rgrp(struct gfs_glock *gl, int flags)
{
	struct gfs_rgrpd *rgd = (struct gfs_rgrpd *) gl->gl_object;

	GFS_ASSERT_GLOCK(rgd && gl->gl_lvb, gl,);

	if (!(flags & GL_SKIP)) {
		gfs_rgrp_relse(rgd);

		if (test_bit(GLF_DIRTY, &gl->gl_flags))
			gfs_rgrp_lvb_fill(rgd);
	}

	return;
}

/**
 * xmote_trans_th - promote/demote a metadata glock
 * @gl: the glock
 * @state: the requested state
 * @flags: the flags passed into gfs_glock()
 *
 */

static void
xmote_trans_th(struct gfs_glock *gl, unsigned int state, int flags)
{
	struct gfs_sbd *sdp = gl->gl_sbd;
	int error;

	if (gl->gl_state != LM_ST_UNLOCKED &&
	    test_bit(SDF_JOURNAL_LIVE, &sdp->sd_flags)) {
		gfs_log_flush(sdp);
		gfs_sync_dev(sdp);
		gfs_pull_tail(sdp);

		error = gfs_shutdown_log(sdp);
		GFS_ASSERT_SBD(!error, sdp,);	/*  FixMe!!!  */
	}

	gfs_glock_xmote_th(gl, state, flags);

	return;
}

/**
 * xmote_trans_bh - promote/demote a metadata glock
 * @gl: the glock
 *
 */

static void
xmote_trans_bh(struct gfs_glock *gl)
{
	struct gfs_sbd *sdp = gl->gl_sbd;
	struct gfs_glock *j_gl = sdp->sd_journal_gh.gh_gl;
	struct gfs_log_header head;
	int error;

	if (gl->gl_state != LM_ST_UNLOCKED &&
	    test_bit(SDF_JOURNAL_LIVE, &sdp->sd_flags)) {
		j_gl->gl_ops->go_sync(j_gl, DIO_METADATA | DIO_DATA);
		j_gl->gl_ops->go_inval(j_gl, DIO_METADATA | DIO_DATA);

		error = gfs_find_jhead(sdp, &sdp->sd_jdesc, j_gl, &head);
		GFS_ASSERT_SBD(!error, sdp,);	/*  FixMe!!!  */
		GFS_ASSERT_SBD(head.lh_flags & GFS_LOG_HEAD_UNMOUNT, sdp,);

		/*  Initialize some head of the log stuff  */
		sdp->sd_sequence = head.lh_sequence;
		sdp->sd_log_head = head.lh_first + 1;
	}

	return;
}

/**
 * drop_trans_th - prepare the transaction glock to be released
 * @gl: the glock
 *
 * We want to sync the device even with localcaching.  Remember
 * that localcaching journal replay only marks buffers dirty.
 */

static void
drop_trans_th(struct gfs_glock *gl)
{
	struct gfs_sbd *sdp = gl->gl_sbd;
	int error;

	if (test_bit(SDF_JOURNAL_LIVE, &sdp->sd_flags)) {
		gfs_log_flush(sdp);
		gfs_sync_dev(sdp);
		gfs_pull_tail(sdp);

		error = gfs_shutdown_log(sdp);
		GFS_ASSERT_SBD(!error, sdp,);	/*  FixMe!!!  */
	}

	gfs_glock_drop_th(gl);

	return;
}

/**
 * demote_ok_nondisk - check to see if it's ok to unlock a glock
 * @gl: the glock
 *
 * Returns: TRUE if it's ok
 */

static int
demote_ok_nondisk(struct gfs_glock *gl)
{
	return FALSE;
}

/**
 * demote_ok_quota - check to see if it's ok to unlock a glock
 * @gl: the glock
 *
 * Returns: TRUE if it's ok
 */

static int
demote_ok_quota(struct gfs_glock *gl)
{
	int demote;
	demote = !atomic_read(&gl->gl_lvb_count);
	return demote;
}

struct gfs_glock_operations gfs_meta_glops = {
      go_xmote_th:gfs_glock_xmote_th,
      go_drop_th:gfs_glock_drop_th,
      go_sync:sync_meta,
      go_inval:inval_meta,
      go_demote_ok:demote_ok_meta,
      go_type:LM_TYPE_META
};

struct gfs_glock_operations gfs_inode_glops = {
      go_xmote_th:xmote_inode_th,
      go_xmote_bh:xmote_inode_bh,
      go_drop_th:drop_inode_th,
      go_sync:sync_inode,
      go_inval:inval_inode,
      go_demote_ok:demote_ok_inode,
      go_lock:lock_inode,
      go_unlock:unlock_inode,
      go_type:LM_TYPE_INODE
};

struct gfs_glock_operations gfs_rgrp_glops = {
      go_xmote_th:xmote_rgrp_th,
      go_xmote_bh:xmote_rgrp_bh,
      go_drop_th:drop_rgrp_th,
      go_sync:sync_meta,
      go_inval:inval_meta,
      go_demote_ok:demote_ok_rgrp,
      go_lock:lock_rgrp,
      go_unlock:unlock_rgrp,
      go_type:LM_TYPE_RGRP
};

struct gfs_glock_operations gfs_trans_glops = {
      go_xmote_th:xmote_trans_th,
      go_xmote_bh:xmote_trans_bh,
      go_drop_th:drop_trans_th,
      go_type:LM_TYPE_NONDISK
};

struct gfs_glock_operations gfs_iopen_glops = {
      go_xmote_th:gfs_glock_xmote_th,
      go_drop_th:gfs_glock_drop_th,
      go_callback:gfs_iopen_callback,
      go_type:LM_TYPE_IOPEN
};

struct gfs_glock_operations gfs_flock_glops = {
      go_xmote_th:gfs_glock_xmote_th,
      go_drop_th:gfs_glock_drop_th,
      go_type:LM_TYPE_FLOCK
};

struct gfs_glock_operations gfs_plock_head_glops = {
      go_xmote_th:gfs_glock_xmote_th,
      go_drop_th:gfs_glock_drop_th,
      go_type:LM_TYPE_PLOCK_HEAD
};

struct gfs_glock_operations gfs_plock_glops = {
      go_xmote_th:gfs_glock_xmote_th,
      go_drop_th:gfs_glock_drop_th,
      go_type:LM_TYPE_PLOCK
};

struct gfs_glock_operations gfs_nondisk_glops = {
      go_xmote_th:gfs_glock_xmote_th,
      go_drop_th:gfs_glock_drop_th,
      go_demote_ok:demote_ok_nondisk,
      go_type:LM_TYPE_NONDISK
};

struct gfs_glock_operations gfs_quota_glops = {
      go_xmote_th:gfs_glock_xmote_th,
      go_drop_th:gfs_glock_drop_th,
      go_demote_ok:demote_ok_quota,
      go_type:LM_TYPE_QUOTA
};
