/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "bmap.h"
#include "dio.h"
#include "file.h"
#include "glock.h"
#include "glops.h"
#include "inode.h"

/**
 * gfs_mark_vfs_dirty - set or clear a dirty flag in the vfs super structure
 * @sdp: the filesystem
 * @dirty: TRUE if there is stuff in the incore log that needs to be synced
 *
 */

void
gfs_mark_vfs_dirty(struct gfs_sbd *sdp, int dirty)
{

	sdp->sd_vfs->s_dirt = dirty;

}

/**
 * gfs_invalidate_inodes - get rid of excess inodes
 * @sdp: the filesystem
 * @dirty: 0 if all inodes are cleared, -EBUSY if some are still pinned.
 *
 */

int
gfs_invalidate_inodes(struct gfs_sbd *sdp)
{
	int error = 0;

	if (invalidate_inodes(sdp->sd_vfs))
		error = -EBUSY;

	return error;
}

/**
 * gfs_device_name - fill in a string with the device name
 * @sdp: the filesystem
 * @devname: the string
 * @size: the size of the string
 *
 */

void
gfs_device_name(struct gfs_sbd *sdp, char *devname, unsigned int size)
{
	char major[11], minor[11];

	if (sprintf(major, "%u", (uint32_t) MAJOR(sdp->sd_vfs->s_dev)) +
	    sprintf(minor, "%u", (uint32_t) MINOR(sdp->sd_vfs->s_dev)) +
	    4 <= size)
		sprintf(devname, "(%s,%s)", major, minor);
	else if (size)
		*devname = 0;

}

/**
 * gmalloc_glock - malloc a struct gfs_glock structure
 * @sdp: the filesystem
 * @glops: the type of glock we want
 *
 * Returns: the structure
 */

struct gfs_glock *
gmalloc_glock(struct gfs_sbd *sdp, struct gfs_glock_operations *glops)
{
	struct inode *aspace = NULL;
	struct gfs_glock *gl;

	if (glops == &gfs_inode_glops || glops == &gfs_rgrp_glops
	    || glops == &gfs_meta_glops)
		aspace = gfs_get_aspace(sdp);

	if (gfs_glock_in_aspace && aspace)
		gl = (struct gfs_glock *) &aspace->u;
	else {
		RETRY_MALLOC(gl =
			     kmem_cache_alloc(gfs_glock_cachep, GFP_KERNEL),
			     gl);
	}

	memset(gl, 0, sizeof (struct gfs_glock));

	gl->gl_arch.gl_aspace = aspace;
	INIT_LIST_HEAD(&gl->gl_arch.gl_dirty_buffers);
	INIT_LIST_HEAD(&gl->gl_arch.gl_ail_bufs);

	return gl;
}

/**
 * gfree_glock - free a struct gfs_glock structure
 * @p the glock to be freed
 *
 */

void
gfree_glock(struct gfs_glock *gl)
{
	struct inode *aspace = gl->gl_arch.gl_aspace;

	if (!(gfs_glock_in_aspace && aspace))
		kmem_cache_free(gfs_glock_cachep, gl);
	if (aspace)
		gfs_put_aspace(aspace);
}

/**
 * gmalloc_inode - malloc a struct gfs_inode structure
 *
 * Returns: the structure
 */

struct gfs_inode *
gmalloc_inode(void)
{
	struct gfs_inode *ip;

	RETRY_MALLOC(ip = kmem_cache_alloc(gfs_inode_cachep, GFP_KERNEL), ip);
	memset(ip, 0, sizeof (struct gfs_inode));

	spin_lock_init(&ip->i_arch.i_lock);
	INIT_LIST_HEAD(&ip->i_arch.i_mmap_shared);

	return ip;
}

/**
 * gfree_inode - free a struct gfs_inode structure
 * @ip the inode to be freed
 *
 */

void
gfree_inode(struct gfs_inode *ip)
{
	gfs_flush_meta_cache(ip);
	kmem_cache_free(gfs_inode_cachep, ip);
}

/**
 * gmalloc_bufdata - malloc a struct gfs_bufdata structure
 *
 * Returns: the structure
 */

struct gfs_bufdata *
gmalloc_bufdata(void)
{
	struct gfs_bufdata *bd;

	RETRY_MALLOC(bd = kmem_cache_alloc(gfs_bufdata_cachep, GFP_KERNEL), bd);
	memset(bd, 0, sizeof (struct gfs_bufdata));

	return bd;
}

/**
 * gfree_bufdata - free a struct gfs_bufdata structure
 * @p the bufdata to be freed
 *
 */

void
gfree_bufdata(struct gfs_bufdata *bd)
{
	kmem_cache_free(gfs_bufdata_cachep, bd);
}

/**
 * gmalloc_mh_cache - malloc a struct gfs_mh_cache structure
 *
 * Returns: the structure
 */

struct gfs_mh_cache *
gmalloc_mh_cache(void)
{
	struct gfs_mh_cache *bd;

	RETRY_MALLOC(bd =
		     kmem_cache_alloc(gfs_mh_cache_cachep, GFP_KERNEL), bd);
	memset(bd, 0, sizeof (struct gfs_mh_cache));

	return bd;
}

/**
 * gfree_mh_cache - free a struct gfs_mh_cache structure
 * @p the mh_cache to be freed
 *
 */

void
gfree_mh_cache(struct gfs_mh_cache *bd)
{
	kmem_cache_free(gfs_mh_cache_cachep, bd);
}
