/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

#include "linux_mount.h"

/**
 * gfs_make_args - Parse mount arguments
 * @data:
 * @args:
 *
 * Return: 1 if data was valid, 0 if data is not valid
 */

int
gfs_make_args(char *data, struct gfs_args *args)
{
	char *opt, *p;
	int do_free = FALSE;
	int ret = 1;

	/*  If someone preloaded options, use those instead  */

	down(&gfs_mount_args_lock);
	if (gfs_mount_args) {
		data = gfs_mount_args;
		gfs_mount_args = NULL;
		do_free = TRUE;
	}
	up(&gfs_mount_args_lock);

	/*  Set some defaults  */

	memset(args, 0, sizeof (struct gfs_args));

	args->ar_num_glockd = GFS_GLOCKD_DEFAULT;

	/*  Split the options into tokens with the "," character and process
	   them  */

	if (data) {
		for (opt = data; (opt = strtok(opt, ",")); opt = NULL) {
			/* This splits the token in half if there is an "=" in it */
			p = strchr(opt, '=');
			if (p)
				*p++ = 0;

			/*  GFS options  */

			if (!strcmp(opt, "lockproto")) {
				if (!p) {
					printk
					    ("GFS: need argument to lockproto\n");
					ret = 0;
					break;
				}
				strncpy(args->ar_lockproto, p, 256);
				args->ar_lockproto[255] = 0;
			}

			else if (!strcmp(opt, "locktable")) {
				if (!p) {
					printk
					    ("GFS: need argument to locktable\n");
					ret = 0;
					break;
				}
				strncpy(args->ar_locktable, p, 256);
				args->ar_locktable[255] = 0;
			}

			else if (!strcmp(opt, "hostdata")) {
				if (!p) {
					printk
					    ("GFS: need argument to hostdata\n");
					ret = 0;
					break;
				}
				strncpy(args->ar_hostdata, p, 256);
				args->ar_hostdata[255] = 0;
			}

			else if (!strcmp(opt, "ignore_local_fs"))
				args->ar_ignore_local_fs = TRUE;

			else if (!strcmp(opt, "localflocks"))
				args->ar_localflocks = TRUE;

			else if (!strcmp(opt, "localcaching"))
				args->ar_localcaching = TRUE;

			else if (!strcmp(opt, "upgrade"))
				args->ar_upgrade = TRUE;

			else if (!strcmp(opt, "num_glockd") && p) {
				if (!p) {
					printk
					    ("GFS: need argument to num_glockd\n");
					ret = 0;
					break;
				}
				sscanf(p, "%u", &args->ar_num_glockd);
				if (!
				    (0 < args->ar_num_glockd
				     && args->ar_num_glockd <=
				     GFS_GLOCKD_MAX)) {
					printk
					    ("GFS: 0 < num_glockd <= %u  (not %u)\n",
					     GFS_GLOCKD_MAX,
					     args->ar_num_glockd);
					ret = 0;
					break;
				}
			}

			else if (!strcmp(opt, "acl"))
				args->ar_posixacls = TRUE;

			/*  Unknown  */

			else {
				ret = 0;
				break;
			}
		}
	}

	if (!ret)
		printk("GFS: invalid mount option(s)\n");

	/*  Free the option space  */

	if (do_free)
		kfree(data);

	return ret;
}

/**
 * gfs_proc_write - Read in some mount options
 * @file: unused
 * @buffer: a buffer of mount options
 * @count: the length of the mount options
 * @data: unused
 *
 * Called when someone writes to /proc/fs/gfs.
 * It allows you to specify mount options when you can't do it
 * from mount.  i.e. from a inital ramdisk
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
gfs_proc_write(struct file *file, const char *buffer,
	       unsigned long count, void *data)
{
	int error;
	char *p;

	MOD_INC_USE_COUNT;
	down(&gfs_mount_args_lock);

	if (gfs_mount_args) {
		kfree(gfs_mount_args);
		gfs_mount_args = NULL;
	}

	if (!count) {
		error = 0;
		goto fail;
	}

	gfs_mount_args = gmalloc(count + 1);

	error = (copy_from_user(gfs_mount_args, buffer, count)) ? -EFAULT : 0;
	if (error)
		goto fail_free;

	gfs_mount_args[count] = 0;

	/*  Get rid of extra newlines  */

	for (p = gfs_mount_args; *p; p++)
		if (*p == '\n')
			*p = 0;

	up(&gfs_mount_args_lock);
	MOD_DEC_USE_COUNT;

	return count;

      fail_free:
	kfree(gfs_mount_args);
	gfs_mount_args = NULL;

      fail:
	up(&gfs_mount_args_lock);
	MOD_DEC_USE_COUNT;
	return error;
}
