/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

/*
 * Plocks are kept in per-inode lists.  They are also kept in per-node lists.
 * So, every plock exists in a list particular to a specific inode and the
 * particular node which holds the plock.  Plocks from different processes
 * on one node are kept in the same list for the node.
 *
 * The plock lists are contructed of locks in a special lock space designated
 * for plocks.  The lvb's on the locks are used to store the lock details like
 * range, type, pid of owner, and a pointer for the linked list of plocks on
 * the inode.
 *
 * Within the plock lock space, the lock number to be used for a specific plock
 * is constructed as follows:
 * o 16 bits to identify the node of the proccess holding the plock (jid)
 * o 32 bits to identify the inode to which the plock applies (gid)
 * o 16 bits to identify the particular plock of all plocks on the inode (lid)
 *
 * If a separate, complete lock space was used for each node's plocks, the
 * first lock number component (node id) could be eliminated.  This is not
 * done because the quantity of separate lock spaces is limited.  So, the
 * 16 bit node id limits the number of nodes which can use plocks to 2^16
 * which is still a rather large cluster.
 *
 * The inode identifier is also limited to 32 bits.  This means that inode
 * numbers greater than 2^32 cannot have plocks.  Inode numbers are equal
 * to the sector on the volume where the inode exists, so inodes created
 * beyond the range of 2 TB cannot have plocks.  It would not be difficult
 * to change this limitation so that any inode could have plocks, but the
 * new limitation would be that a maximum of 2^32 inodes could have plocks
 * existing on them at once.
 *
 * The final limitation is the 16 bits used to identify a particular plock
 * of the many which can exist on an inode at once.  This value is called
 * the lock id (lid).  Because unique plock lists are kept per-node, each
 * node can have a maximum of 2^16 plocks outstanding on any one inode at
 * once.
 *
 * When a node holding plocks dies, all the node's locks, including those used
 * for plocks, are cleared during recovery.  When recovery completes, another
 * node blocking and retrying on a plock from a failed node will succeed in
 * acquiring it.  The semantics should be the same as those between two
 * processes on the same machine.
 *
 * Note that the PID in pe_pid field of a struct gfs_plock_elem structure can't
 * be a a process's PID.  When using the clone() system call, it's possible
 * for a process to inherit the plocks of the parent process.  In other
 * words, the child (with a different PID) can hold plocks on a file
 * long after the parent process had died.  It's possible (though unlikely)
 * that another completely different process that happens to have the
 * same PID could try to acquire plocks on that same file.
 */

#include "gfs.h"
#include "glock.h"
#include "glops.h"
#include "lvb.h"
#include "plock.h"
#include "super.h"

#define PLOCK_JHEAD_NUM(jid, gid) (((uint64_t)jid << 48) | ((uint64_t)gid << 16))
#define PLOCK_ELEM_NUM(jid, gid, lid) (((uint64_t)jid << 48) | \
                                       ((uint64_t)gid << 16) | lid)
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define MAX(a, b) ((a) > (b) ? (a) : (b))
struct plock_pid {
	struct list_head pp_list;
	unsigned int pp_count;

	unsigned long pp_owner;
	long pp_process_pid;
	uint32_t pp_lvb_pid;
};

/**
 * find_pp_by_owner - find the struct plock_pid for a given owner in a filesystem
 * @sdp: the filesystem
 * @owner: the owner identifier
 *
 * Returns: the pp structure if it exists, NULL otherwise
 */

static struct plock_pid *
find_pp_by_owner(struct gfs_sbd *sdp, unsigned long owner)
{
	struct list_head *tmp, *head;
	struct plock_pid *pp = NULL;

	for (head = &sdp->sd_plock_list, tmp = head->next;
	     tmp != head; tmp = tmp->next) {
		pp = list_entry(tmp, struct plock_pid, pp_list);
		if (pp->pp_owner == owner)
			break;
		pp = NULL;
	}

	return pp;
}

/**
 * find_pp_by_lvb_pid - find the struct plock_pid for a given lvb_pid in a filesystem
 * @sdp: the filesystem
 * @lvb_pid: the uint32_t pid value in a plock element LVB
 *
 * Returns: the pp structure if it exists, NULL otherwise
 */

static struct plock_pid *
find_pp_by_lvb_pid(struct gfs_sbd *sdp, uint32_t lvb_pid)
{
	struct list_head *tmp, *head;
	struct plock_pid *pp = NULL;

	for (head = &sdp->sd_plock_list, tmp = head->next;
	     tmp != head; tmp = tmp->next) {
		pp = list_entry(tmp, struct plock_pid, pp_list);
		if (pp->pp_lvb_pid == lvb_pid)
			break;
		pp = NULL;
	}

	return pp;
}

/**
 * lvb_pid_to_process_pid - convert a lvb_pid to the pid of a real process
 * @sdp: the filesystem
 * @lvb_pid: the PID in the pe_pid field of a struct gfs_plock_elem structure
 *
 * The panics if the struct plock_pid structure for the lvb_pid isn't found.
 *
 * Returns: the process's PID
 */

static long
lvb_pid_to_process_pid(struct gfs_sbd *sdp, uint32_t lvb_pid)
{
	struct plock_pid *pp;
	long process_pid = -1;

	spin_lock(&sdp->sd_plock_lock);

	pp = find_pp_by_lvb_pid(sdp, lvb_pid);
	if (pp)
		process_pid = pp->pp_process_pid;

	spin_unlock(&sdp->sd_plock_lock);

	GFS_ASSERT_SBD(pp, sdp,);

	return process_pid;
}

/**
 * get_pp - find (or possible create) a struct plock_pid structure for a owner
 * @sdp: the filesystem
 * @owner: a unsigned long representing the owner of plocks
 * @create: TRUE if the pp should be created
 *
 * Returns: the struct plock_pid structure, or NULL if it doesn't exist and !@create
 */

static struct plock_pid *
get_pp(struct gfs_sbd *sdp, unsigned long owner, int create)
{
	struct plock_pid *pp = NULL, *new_pp = NULL;

	for (;;) {
		spin_lock(&sdp->sd_plock_lock);

		pp = find_pp_by_owner(sdp, owner);
		if (pp)
			pp->pp_count++;
		else if (new_pp) {
			pp = new_pp;
			new_pp = NULL;
			do {
				pp->pp_lvb_pid = gfs_random();
			} while (find_pp_by_lvb_pid(sdp, pp->pp_lvb_pid));
			list_add(&pp->pp_list, &sdp->sd_plock_list);
		}

		spin_unlock(&sdp->sd_plock_lock);

		if (pp || !create)
			break;

		new_pp = gmalloc(sizeof (struct plock_pid));

		memset(new_pp, 0, sizeof (struct plock_pid));
		new_pp->pp_count = 1;

		new_pp->pp_owner = owner;
		new_pp->pp_process_pid = current->pid;
	}

	if (new_pp)
		kfree(new_pp);

	return pp;
}

/**
 * hold_pp - hold a reference on a struct plock_pid structure
 * @sdp: the filesystem
 * @pp: the structure
 *
 */

static void
hold_pp(struct gfs_sbd *sdp, struct plock_pid *pp)
{

	spin_lock(&sdp->sd_plock_lock);
	pp->pp_count++;
	spin_unlock(&sdp->sd_plock_lock);

}

/**
 * put_pp - drop a reference on a struct plock_pid structure (free it if necessary)
 * @sdp: the filesystem
 * @pp: the structure
 *
 */

static void
put_pp(struct gfs_sbd *sdp, struct plock_pid *pp)
{

	spin_lock(&sdp->sd_plock_lock);

	GFS_ASSERT_SBD(pp->pp_count, sdp,);
	if (!--pp->pp_count)
		list_del(&pp->pp_list);
	else
		pp = NULL;

	spin_unlock(&sdp->sd_plock_lock);

	if (pp)
		kfree(pp);

}

/**
 * overlap_type - returns a value based on the type of overlap
 * @s1 - start of new lock range
 * @e1 - end of new lock range
 * @s2 - start of existing lock range
 * @e2 - end of existing lock range
 *
 */

static int
overlap_type(uint64_t s1, uint64_t e1, uint64_t s2, uint64_t e2)
{
	int ret;

	/*
	 * ---r1---
	 * ---r2---
	 */

	if (s1 == s2 && e1 == e2)
		ret = 0;

	/*
	 * --r1--
	 * ---r2---
	 */

	else if (s1 == s2 && e1 < e2)
		ret = 1;

	/*
	 *   --r1--
	 * ---r2---
	 */

	else if (s1 > s2 && e1 == e2)
		ret = 1;

	/*
	 *  --r1--
	 * ---r2---
	 */

	else if (s1 > s2 && e1 < e2)
		ret = 2;

	/*
	 * ---r1---  or  ---r1---  or  ---r1---
	 * --r2--          --r2--       --r2--
	 */

	else if (s1 <= s2 && e1 >= e2)
		ret = 3;

	/*
	 *   ---r1---
	 * ---r2---
	 */

	else if (s1 > s2 && e1 > e2)
		ret = 4;

	/*
	 * ---r1---
	 *   ---r2---
	 */

	else if (s1 < s2 && e1 < e2)
		ret = 4;

	else
		ret = -1;

	return ret;
}

/**
 * lock_pl_head
 * @ip:
 * @glh:
 *
 * Returns: 0 on success, -EXXX on error
 */

static int
lock_pl_head(struct gfs_glock *ph_gl, struct gfs_holder *ph_gh,
	     struct gfs_holder *j_gh)
{
	int error;

	error = gfs_jindex_hold(ph_gl->gl_sbd, j_gh);
	if (!error) {
		error =
		    gfs_glock_nq_init(ph_gl, LM_ST_EXCLUSIVE, GL_NOCACHE,
				      ph_gh);
		if (error)
			gfs_glock_dq_uninit(j_gh);
	}

	return error;
}

/**
 * init_plock_head
 * @ip:
 * @glh:
 * @phead:
 *
 * FIXME: resolve possible gid colisions for inodes beyond 32 bits
 */

static void
init_plock_phead(struct gfs_inode *ip, struct gfs_glock *ph_gl,
		 struct gfs_plock_phead *phead)
{

	phead->ph_magic = GFS_PLOCK_PHEAD_MAGIC;
	phead->ph_gid = (ip->i_num.no_formal_ino & 0x00000000FFFFFFFFll);

	gfs_plock_phead_out(phead, ph_gl->gl_lvb);

}

/**
 * read_plock_jhead
 * @ip:
 * @jid:
 * @gid:
 * @jhead:
 *
 * Returns 0 on success, -EXXX on error
 */

static int
read_plock_jhead(struct gfs_inode *ip, uint32_t jid, uint32_t gid,
		 struct gfs_plock_jhead *jhead)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_glock *jh_gl;
	struct gfs_holder jh_gh;
	uint64_t num;
	int error;

	num = PLOCK_JHEAD_NUM(jid, gid);

	error = gfs_get_glstruct(sdp, num, &gfs_plock_glops, ST_CREATE, &jh_gl);
	if (error)
		goto out;

	error = gfs_hold_lvb(jh_gl);
	if (error)
		goto out_glput;

	error = gfs_glock_nq_init(jh_gl, LM_ST_EXCLUSIVE, GL_NOCACHE, &jh_gh);
	if (error)
		goto out_unhold;

	gfs_plock_jhead_in(jhead, jh_gl->gl_lvb);

	gfs_glock_dq_uninit(&jh_gh);

      out_unhold:
	gfs_unhold_lvb(jh_gl);

      out_glput:
	gfs_put_glstruct(jh_gl);

      out:

	return error;
}

/**
 * read_plock_elem
 * @ip:
 * @jid:
 * @gid:
 * @lid:
 * @pl:
 *
 * Returns 0 on success, -EXXX on error
 */

static int
read_plock_elem(struct gfs_inode *ip, uint32_t jid, uint32_t gid, uint16_t lid,
		struct gfs_plock_elem *pl)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_glock *e_gl;
	struct gfs_holder e_gh;
	uint64_t num;
	int error;

	num = PLOCK_ELEM_NUM(jid, gid, lid);

	error = gfs_get_glstruct(sdp, num, &gfs_plock_glops, ST_CREATE, &e_gl);
	if (error)
		goto out;

	error = gfs_hold_lvb(e_gl);
	if (error)
		goto out_glput;

	error = gfs_glock_nq_init(e_gl, LM_ST_EXCLUSIVE, GL_NOCACHE, &e_gh);
	if (error)
		goto out_unhold;

	gfs_plock_elem_in(pl, e_gl->gl_lvb);

	gfs_glock_dq_uninit(&e_gh);

      out_unhold:
	gfs_unhold_lvb(e_gl);

      out_glput:
	gfs_put_glstruct(e_gl);

      out:

	return error;
}

/**
 * write_plock_jhead
 * @ip:
 * @gid:
 * @jhead:
 *
 * Returns 0 on success, -EXXX on error
 */

static int
write_plock_jhead(struct gfs_inode *ip, uint32_t gid,
		  struct gfs_plock_jhead *jhead)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_glock *jh_gl;
	struct gfs_holder jh_gh;
	uint64_t num;
	int error;

	num = PLOCK_JHEAD_NUM(sdp->sd_lockstruct.ls_jid, gid);

	error =
	    gfs_get_glstruct(sdp, num, &gfs_plock_glops, NO_ST_CREATE, &jh_gl);
	if (error)
		goto out;
	if (!jh_gl) {
		error = -EINVAL;
		goto out;
	}

	error = gfs_hold_lvb(jh_gl);
	if (error)
		goto out_glput;

	error = gfs_glock_nq_init(jh_gl, LM_ST_EXCLUSIVE, GL_NOCACHE, &jh_gh);
	if (error)
		goto out_unhold;

	gfs_plock_jhead_out(jhead, jh_gl->gl_lvb);

	gfs_glock_dq_uninit(&jh_gh);

      out_unhold:
	gfs_unhold_lvb(jh_gl);

      out_glput:
	gfs_put_glstruct(jh_gl);

      out:

	return error;
}

/**
 * write_plock_elem
 * @ip:
 * @gid:
 * @lid:
 * @pl:
 *
 * Returns 0 on success, -EXXX on error
 */

static int
write_plock_elem(struct gfs_inode *ip, uint32_t gid, uint16_t lid,
		 struct gfs_plock_elem *pl)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_glock *pe_gl;
	struct gfs_holder pe_gh;
	uint64_t num;
	int error;

	num = PLOCK_ELEM_NUM(sdp->sd_lockstruct.ls_jid, gid, lid);

	error =
	    gfs_get_glstruct(sdp, num, &gfs_plock_glops, NO_ST_CREATE, &pe_gl);
	if (error)
		goto out;
	if (!pe_gl) {
		error = -EINVAL;
		goto out;
	}

	error = gfs_hold_lvb(pe_gl);
	if (error)
		goto out_glput;

	error = gfs_glock_nq_init(pe_gl, LM_ST_EXCLUSIVE, GL_NOCACHE, &pe_gh);
	if (error)
		goto out_unhold;

	gfs_plock_elem_out(pl, pe_gl->gl_lvb);

	gfs_glock_dq_uninit(&pe_gh);

      out_unhold:
	gfs_unhold_lvb(pe_gl);

      out_glput:
	gfs_put_glstruct(pe_gl);

      out:

	return error;
}

/**
 * find_free_lid
 * @ip:
 * @gid:
 * @out_lid:
 * @first_lid:
 *
 * Returns 0 on success, -EXXX on error
 */

static int
find_free_lid(struct gfs_inode *ip, uint32_t gid, uint16_t * out_lid,
	      uint16_t first_lid)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_plock_elem pl;
	uint16_t lid, cur_lid, prev_lid = 0;
	int error, found, used;

	found = FALSE;
	for (lid = 1; lid < 0xFFFF; lid++) {
		/* go through entire element chain for each lid to check if it's used */

		cur_lid = first_lid;

		used = FALSE;
		while (cur_lid) {
			if (cur_lid == lid) {
				used = TRUE;
				break;
			}

			error =
			    read_plock_elem(ip, sdp->sd_lockstruct.ls_jid, gid,
					    cur_lid, &pl);
			if (error)
				goto out;

			GFS_ASSERT_INODE(pl.pe_magic == GFS_PLOCK_ELEM_MAGIC,
					 ip, gfs_plock_elem_print(&pl, TRUE);
			    );

			prev_lid = cur_lid;
			cur_lid = pl.pe_next_lid;
		}

		if (!used) {
			*out_lid = lid;
			found = TRUE;
			break;
		}
	}

	if (!found) {
		error = -1;
		goto out;
	}

	/* make the new lid the last element of list.  this function isn't
	 * used until there's one element in list, so prev_lid should
	 * always be non-zero */

	GFS_ASSERT_INODE(prev_lid, ip, gfs_plock_elem_print(&pl, TRUE);
	    );

	error =
	    read_plock_elem(ip, sdp->sd_lockstruct.ls_jid, gid, prev_lid, &pl);
	if (error)
		goto out;

	GFS_ASSERT_INODE(pl.pe_next_lid == 0, ip,
			 gfs_plock_elem_print(&pl, TRUE);
	    );
	pl.pe_next_lid = lid;
	error = write_plock_elem(ip, gid, prev_lid, &pl);

      out:

	return error;
}

/**
 * update_lid
 * @ip:
 * @gid:
 * @lid:
 * @pl:
 *
 * Returns 0 on success, -EXXX on error
 */

static int
update_lid(struct gfs_inode *ip, uint32_t gid, uint16_t lid,
	   struct gfs_plock_elem *pl)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_plock_elem newpl;
	int error;

	error =
	    read_plock_elem(ip, sdp->sd_lockstruct.ls_jid, gid, lid, &newpl);
	if (error)
		goto out;

	GFS_ASSERT_INODE(newpl.pe_magic == GFS_PLOCK_ELEM_MAGIC, ip,
			 gfs_plock_elem_print(&newpl, TRUE);
	    );

	if (pl) {
		pl->pe_next_lid = newpl.pe_next_lid;

		newpl.pe_start = pl->pe_start;
		newpl.pe_end = pl->pe_end;
		newpl.pe_excl = pl->pe_excl;
	} else {
		/* no input values means invalidate */
		newpl.pe_start = 0;
		newpl.pe_end = 0;
		newpl.pe_excl = 0;
		newpl.pe_mylid = 0;
	}

	error = write_plock_elem(ip, gid, lid, &newpl);

      out:

	return error;
}

/**
 * inval_lid
 * @ip:
 * @gid:
 * @lid:
 *
 * Returns 0 on success, -EXXX on error
 */

static int
inval_lid(struct gfs_inode *ip, uint32_t gid, uint16_t lid)
{
	return update_lid(ip, gid, lid, NULL);
}

/**
 * add_lid
 * @ip:
 * @pp:
 * @gid:
 * @start:
 * @end:
 * @ex:
 * @first_lid:
 *
 * create a new element which requires finding an lid to use and changing the
 * next_lid pointer for someone, it also means leaving a reference on the gl for
 * the lid (which needs to be later removed if this plock operation fails)
 *
 * Returns 0 on success, -EXXX on error
 */

static int
add_lid(struct gfs_inode *ip, struct plock_pid *pp,
	uint32_t gid, uint32_t start, uint32_t end, int ex, uint16_t first_lid)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_plock_elem newpl;
	struct gfs_glock *e_gl;
	uint64_t num = 0;
	uint16_t lid;
	int error;

	/* new lid added at end of list */
	error = find_free_lid(ip, gid, &lid, first_lid);
	if (error)
		goto fail;

	num = PLOCK_ELEM_NUM(sdp->sd_lockstruct.ls_jid, gid, lid);

	error = gfs_get_glstruct(sdp, num, &gfs_plock_glops, ST_CREATE, &e_gl);
	if (error)
		goto fail;

	error = gfs_hold_lvb(e_gl);
	if (error)
		goto fail_glput;

	newpl.pe_magic = GFS_PLOCK_ELEM_MAGIC;
	newpl.pe_pid = pp->pp_lvb_pid;
	newpl.pe_start = start;
	newpl.pe_end = end;
	newpl.pe_excl = ex;
	newpl.pe_mylid = lid;
	newpl.pe_next_lid = 0;

	error = write_plock_elem(ip, gid, lid, &newpl);
	if (error)
		goto fail_unhold;

	/* keep a reference, don't unhold_lvb and put_glstruct */
	hold_pp(sdp, pp);

	return 0;

      fail_unhold:
	gfs_unhold_lvb(e_gl);

      fail_glput:
	gfs_put_glstruct(e_gl);

      fail:
	return error;
}

/**
 * add_first_lid
 * @ip:
 * @pp:
 * @gid:
 * @lid:
 * @start:
 * @end:
 * @ex:
 *
 * Returns 0 on success, -EXXX on error
 */

static int
add_first_lid(struct gfs_inode *ip, struct plock_pid *pp,
	      uint32_t gid, uint16_t lid, uint64_t start, uint64_t end, int ex)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_plock_elem newpl;
	struct gfs_glock *e_gl;
	uint64_t num;
	int error;

	num = PLOCK_ELEM_NUM(sdp->sd_lockstruct.ls_jid, gid, lid);

	error = gfs_get_glstruct(sdp, num, &gfs_plock_glops, ST_CREATE, &e_gl);
	if (error)
		goto fail;

	error = gfs_hold_lvb(e_gl);
	if (error)
		goto fail_glput;

	newpl.pe_magic = GFS_PLOCK_ELEM_MAGIC;
	newpl.pe_pid = pp->pp_lvb_pid;
	newpl.pe_start = start;
	newpl.pe_end = end;
	newpl.pe_excl = ex;
	newpl.pe_mylid = lid;
	newpl.pe_next_lid = 0;

	error = write_plock_elem(ip, gid, lid, &newpl);
	if (error)
		goto fail_unhold;

	/* keep a reference, don't unhold_lvb and put_glstruct */
	hold_pp(sdp, pp);

	return 0;

      fail_unhold:
	gfs_unhold_lvb(e_gl);

      fail_glput:
	gfs_put_glstruct(e_gl);

      fail:
	return error;
}

/**
 * shrink_elem_range
 * @pl:
 * @start:
 * @end:
 *
 * Returns 0 on success, -EXXX on error
 */

static int
shrink_elem_range(struct gfs_plock_elem *pl, uint64_t start, uint64_t end)
{
	int error = 0;

	if (pl->pe_start < start)
		pl->pe_end = start - 1;

	else if (pl->pe_end > end)
		pl->pe_start = end + 1;

	else
		error = -1;

	return error;
}

/**
 * check_for_overlap - checks for overlapping lock regions
 * @start1:
 * @end1:
 * @start2:
 * @end2:
 *
 * Returns: non-zero if ranges overlap
 */

static int
check_for_overlap(uint64_t start1, uint64_t end1, uint64_t start2,
		  uint64_t end2)
{
	/* start1 between range 2 endpoints */
	if (start2 <= start1 && start1 <= end2)
		return TRUE;

	/* end1 between range 2 endpoints */
	if (start2 <= end1 && end1 <= end2)
		return TRUE;

	/* start2 between range 1 endpoints */
	if (start1 <= start2 && start2 <= end1)
		return TRUE;

	/* end2 between range 2 endpoints */
	if (start1 <= end2 && end2 <= end1)
		return TRUE;

	return FALSE;
}

/**
 * check_for_conflicts
 * @ip:
 * @pp:
 * @phead:
 * @start:
 * @end:
 * @ex:
 *
 * Returns 0 on success, -EXXX on error
 */

static int
check_for_conflicts(struct gfs_inode *ip, struct plock_pid *pp,
		    struct gfs_plock_phead *phead, uint64_t start, uint64_t end,
		    int ex)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_plock_jhead jhead;
	struct gfs_plock_elem elem;
	uint32_t jid;
	uint32_t gid = phead->ph_gid;
	uint16_t lid;
	int error = 0;

	for (jid = 0; jid < sdp->sd_journals; jid++) {
		error = read_plock_jhead(ip, jid, gid, &jhead);
		if (error)
			goto out;

		if (jhead.pj_magic != GFS_PLOCK_JHEAD_MAGIC)
			continue;

		lid = jhead.pj_lid;
		GFS_ASSERT_INODE(lid, ip, gfs_plock_jhead_print(&jhead, TRUE);
		    );

		while (lid) {
			error = read_plock_elem(ip, jid, gid, lid, &elem);
			if (error)
				goto out;

			GFS_ASSERT_INODE(elem.pe_magic == GFS_PLOCK_ELEM_MAGIC,
					 ip, gfs_plock_elem_print(&elem, TRUE);
			    );

			if (jid == sdp->sd_lockstruct.ls_jid
			    && elem.pe_pid == pp->pp_lvb_pid) {
				lid = elem.pe_next_lid;
				continue;
			}

			if (ex || elem.pe_excl) {
				error =
				    check_for_overlap(start, end, elem.pe_start,
						      elem.pe_end);
				if (error)
					goto out;
			}

			lid = elem.pe_next_lid;
		}
	}

      out:

	return error;
}

/**
 * get_conflicts
 * @ip:
 * @pp:
 * @phead:
 * @start:
 * @end:
 * @ex:
 * @pid:
 *
 * Returns 0 on no conflicts, -EXXX on error, positive if conflicts found
 */

static int
get_conflicts(struct gfs_inode *ip, struct plock_pid *pp,
	      struct gfs_plock_phead *phead,
	      uint64_t * start, uint64_t * end, int *ex, long *pid)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_plock_jhead jhead;
	struct gfs_plock_elem elem;
	uint32_t jid;
	uint32_t gid = phead->ph_gid;
	uint16_t lid;
	int error = 0;

	for (jid = 0; jid < sdp->sd_journals; jid++) {
		error = read_plock_jhead(ip, jid, gid, &jhead);
		if (error)
			goto out;

		if (jhead.pj_magic != GFS_PLOCK_JHEAD_MAGIC)
			continue;

		lid = jhead.pj_lid;
		GFS_ASSERT_INODE(lid, ip, gfs_plock_jhead_print(&jhead, TRUE);
		    );

		while (lid) {
			error = read_plock_elem(ip, jid, gid, lid, &elem);
			if (error)
				goto out;

			GFS_ASSERT_INODE(elem.pe_magic == GFS_PLOCK_ELEM_MAGIC,
					 ip, gfs_plock_elem_print(&elem, TRUE);
			    );

			if (jid == sdp->sd_lockstruct.ls_jid
			    && elem.pe_pid == pp->pp_lvb_pid) {
				lid = elem.pe_next_lid;
				continue;
			}

			if (*ex || elem.pe_excl) {
				error =
				    check_for_overlap(*start, *end,
						      elem.pe_start,
						      elem.pe_end);
				if (error) {
					*start = elem.pe_start;
					*end = elem.pe_end;
					*ex = elem.pe_excl;
					*pid =
					    (jid ==
					     sdp->sd_lockstruct.
					     ls_jid) ?
					    lvb_pid_to_process_pid(sdp,
								   elem.
								   pe_pid) :
					    elem.pe_pid;
					goto out;
				}
			}

			lid = elem.pe_next_lid;
		}
	}

      out:

	return error;
}

/**
 * remove_lid
 * @ip:
 * @pp:
 * @gid:
 * @lid:
 *
 * Returns 0 on success, -EXXX on error
 */

static int
remove_lid(struct gfs_inode *ip, struct plock_pid *pp, uint32_t gid,
	   uint16_t lid)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_glock *e_gl;
	uint64_t num;
	int error;

	num = PLOCK_ELEM_NUM(sdp->sd_lockstruct.ls_jid, gid, lid);

	error =
	    gfs_get_glstruct(sdp, num, &gfs_plock_glops, NO_ST_CREATE, &e_gl);
	if (error)
		goto out;
	if (!e_gl) {
		error = -EINVAL;
		goto out;
	}

	gfs_put_glstruct(e_gl);

	gfs_unhold_lvb(e_gl);
	gfs_put_glstruct(e_gl);
	put_pp(sdp, pp);

      out:

	return error;
}

/**
 * add_first_lock_range
 * @ip:
 * @pp:
 * @gid:
 * @start:
 * @end:
 * @ex:
 *
 * Note: we rely on the fact that a new gl/lvb will have zeroed lvb data
 * to determine that we have no existing pl elements
 *
 * Adding the first element needs to worry about the jhead, but no merging
 * with existing elements.  Adding subsequent elements doesn't need to
 * bother with jhead, but needs to try and merge with existing elements.
 *
 * Returns 0 on success, -EXXX on error
 */

static int
add_first_lock_range(struct gfs_inode *ip, struct plock_pid *pp,
		     uint32_t gid, uint64_t start, uint64_t end, int ex)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_glock *jh_gl;
	struct gfs_plock_jhead jhead;
	uint64_t num;
	int error;

	num = PLOCK_JHEAD_NUM(sdp->sd_lockstruct.ls_jid, gid);

	error = gfs_get_glstruct(sdp, num, &gfs_plock_glops, ST_CREATE, &jh_gl);
	if (error)
		goto fail;

	error = gfs_hold_lvb(jh_gl);
	if (error)
		goto fail_glput;

	jhead.pj_magic = GFS_PLOCK_JHEAD_MAGIC;
	jhead.pj_lid = 1;
	jhead.pj_jid = sdp->sd_lockstruct.ls_jid;

	error = write_plock_jhead(ip, gid, &jhead);
	if (error)
		goto fail_unhold;

	error = add_first_lid(ip, pp, gid, 1, start, end, ex);
	if (error)
		goto fail_unhold;

	/* success so leave the first reference on our jhead */

	return 0;

      fail_unhold:
	gfs_unhold_lvb(jh_gl);

      fail_glput:
	gfs_put_glstruct(jh_gl);

      fail:
	return error;
}

/**
 * cleanup_lids
 * @ip:
 * @pp:
 * @gid:
 * @first_lid:
 *
 * Loop through a list of elements removing any which have been invalidated.
 *
 * Returns 0 on success, -EXXX on error
 */

static int
cleanup_lids(struct gfs_inode *ip, struct plock_pid *pp, uint32_t gid,
	     uint16_t * first_lid)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_plock_elem elem;
	uint16_t prev_lid = 0, next_lid, lid = *first_lid;
	int error = 0;

	while (lid) {
		error =
		    read_plock_elem(ip, sdp->sd_lockstruct.ls_jid, gid, lid,
				    &elem);
		if (error)
			goto out;

		GFS_ASSERT_INODE(elem.pe_magic == GFS_PLOCK_ELEM_MAGIC, ip,
				 gfs_plock_elem_print(&elem, TRUE);
		    );

		if (!elem.pe_mylid) {
			next_lid = elem.pe_next_lid;

			error = remove_lid(ip, pp, gid, lid);
			if (error)
				goto out;

			/* update the prev lid with new next */

			if (prev_lid) {
				error =
				    read_plock_elem(ip,
						    sdp->sd_lockstruct.ls_jid,
						    gid, prev_lid, &elem);
				if (error)
					goto out;

				elem.pe_next_lid = next_lid;

				error =
				    write_plock_elem(ip, gid, prev_lid, &elem);
				if (error)
					goto out;
			}

			if (*first_lid == lid)
				*first_lid = next_lid;

			lid = next_lid;
		} else {
			prev_lid = lid;
			lid = elem.pe_next_lid;
		}
	}

      out:

	return error;
}

/**
 * add_another_lock_range
 * @ip:
 * @pp:
 * @gid:
 * @first_lid:
 * @start:
 * @end:
 * @ex:
 *
 * Returns 0 on success, -EXXX on error
 */

static int
add_another_lock_range(struct gfs_inode *ip, struct plock_pid *pp,
		       uint32_t gid,
		       uint16_t * first_lid, uint64_t start, uint64_t end,
		       int ex)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_plock_elem elem;
	uint16_t lid = *first_lid;
	int error = 0;

	while (lid) {
		error =
		    read_plock_elem(ip, sdp->sd_lockstruct.ls_jid, gid, lid,
				    &elem);
		if (error)
			goto out;

		GFS_ASSERT_INODE(elem.pe_magic == GFS_PLOCK_ELEM_MAGIC, ip,
				 gfs_plock_elem_print(&elem, TRUE);
		    );

		error =
		    check_for_overlap(start, end, elem.pe_start, elem.pe_end);
		if (!error) {
			lid = elem.pe_next_lid;
			continue;
		}

		/* processes cannot modify one anothers' elements */
		if (elem.pe_pid != pp->pp_lvb_pid) {
			lid = elem.pe_next_lid;
			continue;
		}

		/* existing range (RE) overlaps new range (RN) */

		switch (overlap_type(start, end, elem.pe_start, elem.pe_end)) {
		case 0:
			/* ranges the same - just update the existing element */

			if (elem.pe_excl == ex)
				goto out;

			elem.pe_excl = ex;
			error = update_lid(ip, gid, lid, &elem);
			goto out;

		case 1:
			/* RN within RE and starts or ends on RE boundary - if RN & RE
			 * are same type nothing to do, otherwise make RN a new element
			 * and modify range of RE */

			if (elem.pe_excl == ex)
				goto out;

			error =
			    add_lid(ip, pp, gid, start, end, ex, *first_lid);
			if (error)
				goto out;

			shrink_elem_range(&elem, start, end);

			error = update_lid(ip, gid, lid, &elem);
			goto out;

		case 2:
			/* RN within RE - if RN & RE are same type nothing to do,
			 * otherwise make RN a new element, modify RE element to be
			 * front fragment, add another new element for back fragment */

			if (elem.pe_excl == ex)
				goto out;

			error =
			    add_lid(ip, pp, gid, start, end, ex, *first_lid);
			if (error)
				goto out;

			error =
			    add_lid(ip, pp, gid, end + 1, elem.pe_end,
				    elem.pe_excl, *first_lid);
			if (error)
				goto out;

			elem.pe_end = start - 1;
			error = update_lid(ip, gid, lid, &elem);
			goto out;

		case 3:
			/* RE within RN - inval RE (will be removed) and keep going on
			 * checks of RN which will be added at the end */

			error = inval_lid(ip, gid, lid);
			if (error)
				goto out;

			lid = elem.pe_next_lid;
			continue;

		case 4:
			/* front of RE in RN, or end of RE in RN - if both ranges are same
			 * type, inval (remove) RE, expand RN, and continue checks on RN
			 * which will be added at the end; otherwise update RE to be the
			 * non-common fragment and continue processing RN */

			if (elem.pe_excl == ex) {
				error = inval_lid(ip, gid, lid);
				if (error)
					goto out;
				start = MIN(start, elem.pe_start);
				end = MAX(end, elem.pe_end);
			} else {
				shrink_elem_range(&elem, start, end);
				error = update_lid(ip, gid, lid, &elem);
				if (error)
					goto out;
			}
			lid = elem.pe_next_lid;
			continue;

		default:
			error = -1;
			goto out;
		}
	}

	/* RN was not processed as a part of modifying any RE */

	error = add_lid(ip, pp, gid, start, end, ex, *first_lid);
	if (error)
		goto out;

	/* Some lid's may have been invalidated during processing to avoid
	 * hassles with removing list elements while cycling through the
	 * list.  Now go through list and remove any invalidated lids.  If
	 * the first lid is removed the new first lid value must be passed
	 * up. */

	error = cleanup_lids(ip, pp, gid, first_lid);

      out:

	return error;
}

/**
 * add_lock_range
 * @ip: the inode
 * @pp:
 * @phead: holds gid
 * @start: start of lock range
 * @end: end of lock range
 * @ex: exclusive flag
 * @old:
 * @new:
 *
 * Returns 0 on success, -EXXX on error
 */

static int
add_lock_range(struct gfs_inode *ip, struct plock_pid *pp,
	       struct gfs_plock_phead *phead,
	       uint64_t start, uint64_t end, int ex, int *before, int *after)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_plock_jhead jhead;
	uint32_t gid = phead->ph_gid;
	int error;

	error = read_plock_jhead(ip, sdp->sd_lockstruct.ls_jid, gid, &jhead);
	if (error)
		goto out;

	if (jhead.pj_magic != GFS_PLOCK_JHEAD_MAGIC) {
		/* add_first_lock_range will init our jhead */

		*before = 0;
		error = add_first_lock_range(ip, pp, gid, start, end, ex);
		*after = (error) ? 0 : 1;
	} else {
		*before = 1;
		*after = 1;

		error =
		    add_another_lock_range(ip, pp, gid, &jhead.pj_lid, start,
					   end, ex);

		/* adding the lock may have changed the first lid */

		if (!error)
			write_plock_jhead(ip, gid, &jhead);
	}

      out:

	return error;
}

/**
 * del_lock_range
 * @ip:
 * @pp:
 * @phead:
 * @start:
 * @end:
 * @old:
 * @new:
 *
 * Returns 0 on success, -EXXX on error
 */

static int
del_lock_range(struct gfs_inode *ip, struct plock_pid *pp,
	       struct gfs_plock_phead *phead,
	       uint64_t start, uint64_t end, int *before, int *after)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_plock_jhead jhead;
	struct gfs_plock_elem elem;
	struct gfs_glock *jh_gl;
	uint32_t gid = phead->ph_gid;
	uint16_t lid;
	uint64_t num;
	int need_cleanup = FALSE;
	int error = -1;

	*before = -1;
	*after = -1;

	error = read_plock_jhead(ip, sdp->sd_lockstruct.ls_jid, gid, &jhead);
	if (error)
		goto fail;

	error = 0;
	if (jhead.pj_magic != GFS_PLOCK_JHEAD_MAGIC)
		goto fail;

	lid = jhead.pj_lid;

	while (lid) {
		error =
		    read_plock_elem(ip, sdp->sd_lockstruct.ls_jid, gid, lid,
				    &elem);
		if (error)
			goto fail;

		GFS_ASSERT_INODE(elem.pe_magic == GFS_PLOCK_ELEM_MAGIC, ip,
				 gfs_plock_elem_print(&elem, TRUE);
		    );

		error =
		    check_for_overlap(start, end, elem.pe_start, elem.pe_end);
		if (!error) {
			lid = elem.pe_next_lid;
			continue;
		}

		/* processes cannot modify one anothers' elements */
		if (elem.pe_pid != pp->pp_lvb_pid) {
			lid = elem.pe_next_lid;
			continue;
		}

		/* existing range (RE) overlaps new range (RN) */

		switch (overlap_type(start, end, elem.pe_start, elem.pe_end)) {
		case 0:
			/* ranges the same - just remove the existing element */

			error = inval_lid(ip, gid, lid);
			if (error)
				goto fail;
			need_cleanup = TRUE;
			goto done;

		case 1:
			/* RN within RE and starts or ends on RE boundary -
			 * shrink and update RE */

			shrink_elem_range(&elem, start, end);
			error = update_lid(ip, gid, lid, &elem);
			if (error)
				goto fail;
			goto done;

		case 2:
			/* RN within RE - shrink and update RE to be front fragment,
			 * and add a new element for back fragment */

			error =
			    add_lid(ip, pp, gid, end + 1, elem.pe_end,
				    elem.pe_excl, jhead.pj_lid);
			if (error)
				goto fail;

			elem.pe_end = start - 1;
			error = update_lid(ip, gid, lid, &elem);
			if (error)
				goto fail;
			goto done;

		case 3:
			/* RE within RN - remove RE, then continue checking because RN
			 * could cover other extent elements */

			error = inval_lid(ip, gid, lid);
			if (error)
				goto fail;
			lid = elem.pe_next_lid;
			need_cleanup = TRUE;
			continue;

		case 4:
			/* front of RE in RN, or end of RE in RN - shrink and update
			 * RE, then continue because RN could cover other extent elements */

			shrink_elem_range(&elem, start, end);
			error = update_lid(ip, gid, lid, &elem);
			if (error)
				goto fail;
			lid = elem.pe_next_lid;
			continue;

		default:
			error = -1;
			goto fail;
		}
	}

      done:
	error = 0;

	if (need_cleanup) {
		error = cleanup_lids(ip, pp, gid, &jhead.pj_lid);
		if (error)
			goto fail;
	}

	/* when all elements are removed, first lid will be 0 */
	if (jhead.pj_lid == 0) {
		/* remove last reference to our jhead */
		num = PLOCK_JHEAD_NUM(sdp->sd_lockstruct.ls_jid, gid);
		error =
		    gfs_get_glstruct(sdp, num, &gfs_plock_glops, NO_ST_CREATE,
				     &jh_gl);
		if (error)
			goto fail;
		if (!jh_gl) {
			error = -EINVAL;
			goto fail;
		}
		gfs_put_glstruct(jh_gl);
		gfs_unhold_lvb(jh_gl);
		gfs_put_glstruct(jh_gl);
		*after = 0;
	} else {
		/* first lid pointer in jhead may have changed */
		error = write_plock_jhead(ip, gid, &jhead);
		*after = 1;
	}

      fail:
	return error;
}

/**
 * gfs_plock_get
 * @ip: The inode
 * @owner:
 * @start:
 * @end:
 * @ex:
 * @pid:
 *
 * Returns: -EXXX on error, 0 if nothing blocks request, 1 if request conflicts
 */

int
gfs_plock_get(struct gfs_inode *ip, unsigned long owner, uint64_t * start,
	      uint64_t * end, int *ex, long *pid)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_glock *ph_gl;
	struct gfs_holder j_gh, ph_gh;
	struct gfs_plock_phead phead;
	struct plock_pid *pp;
	int error;

	pp = get_pp(sdp, owner, ST_CREATE);

	error =
	    gfs_get_glstruct(sdp, ip->i_num.no_formal_ino,
			     &gfs_plock_head_glops, ST_CREATE, &ph_gl);
	if (error)
		goto out;

	error = gfs_hold_lvb(ph_gl);
	if (error)
		goto out_glput;

	error = lock_pl_head(ph_gl, &ph_gh, &j_gh);
	if (error)
		goto out_unhold;

	gfs_plock_phead_in(&phead, ph_gl->gl_lvb);

	if (phead.ph_magic == GFS_PLOCK_PHEAD_MAGIC)
		error = get_conflicts(ip, pp, &phead, start, end, ex, pid);

	gfs_glock_dq_uninit(&ph_gh);
	gfs_glock_dq_uninit(&j_gh);

      out_unhold:
	gfs_unhold_lvb(ph_gl);

      out_glput:
	gfs_put_glstruct(ph_gl);

      out:
	put_pp(sdp, pp);

	return error;
}

/**
 * gfs_punlock - unlock lock range
 * @ip:
 * @owner:
 * @start:
 * @end:
 *
 * Returns: 0 on success, -EXXX on error
 *
 * before: T/F plock range elements existed before proceedure
 * after: T/F plock range elements exist after proceedure
 */

int
gfs_punlock(struct gfs_inode *ip, unsigned long owner, uint64_t start,
	    uint64_t end)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_glock *ph_gl;
	struct gfs_holder j_gh, ph_gh;
	struct gfs_plock_phead phead;
	int before = 0, after = 0;
	struct plock_pid *pp;
	int error;

	pp = get_pp(sdp, owner, NO_ST_CREATE);
	if (!pp)
		goto out;

	error =
	    gfs_get_glstruct(sdp, ip->i_num.no_formal_ino,
			     &gfs_plock_head_glops, NO_ST_CREATE, &ph_gl);
	if (error)
		goto fail;
	if (!ph_gl) {
		error = -EINVAL;
		goto fail;
	}

	error = gfs_hold_lvb(ph_gl);
	if (error)
		goto fail_glput;

	error = lock_pl_head(ph_gl, &ph_gh, &j_gh);
	if (error)
		goto fail_unhold;

	gfs_plock_phead_in(&phead, ph_gl->gl_lvb);

	if (phead.ph_magic != GFS_PLOCK_PHEAD_MAGIC)	/*  No elements  */
		goto fail_unlock;

	error = del_lock_range(ip, pp, &phead, start, end, &before, &after);
	if (error)
		goto fail_unlock;

	gfs_glock_dq_uninit(&ph_gh);

	gfs_glock_dq_uninit(&j_gh);

	gfs_unhold_lvb(ph_gl);
	gfs_put_glstruct(ph_gl);

	if (!after) {
		gfs_unhold_lvb(ph_gl);
		gfs_put_glstruct(ph_gl);
	}

	put_pp(sdp, pp);

      out:

	return 0;

      fail_unlock:
	gfs_glock_dq_uninit(&ph_gh);
	gfs_glock_dq_uninit(&j_gh);

      fail_unhold:
	gfs_unhold_lvb(ph_gl);

      fail_glput:
	gfs_put_glstruct(ph_gl);

      fail:
	put_pp(sdp, pp);

	return error;
}

/**
 * gfs_plock - lock lock range
 * @ip:
 * @owner:
 * @wait:
 * @exclusive:
 * @start:
 * @end:
 *
 * Returns: 0 on success, -EXXX on failure
 *
 * before: T/F plock range elements existed before proceedure
 * after: T/F plock range elements exist after proceedure
 */

int
gfs_plock(struct gfs_inode *ip, unsigned long owner, int wait, int ex,
	  uint64_t start, uint64_t end)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_glock *ph_gl;
	struct gfs_holder j_gh, ph_gh;
	struct gfs_plock_phead phead;
	int before = 0, after = 0;
	struct plock_pid *pp;
	int error;

	pp = get_pp(sdp, owner, ST_CREATE);

	error =
	    gfs_get_glstruct(sdp, ip->i_num.no_formal_ino,
			     &gfs_plock_head_glops, ST_CREATE, &ph_gl);
	if (error)
		goto fail;

	error = gfs_hold_lvb(ph_gl);
	if (error)
		goto fail_glput;

      retry:
	error = lock_pl_head(ph_gl, &ph_gh, &j_gh);
	if (error)
		goto fail_unhold;

	gfs_plock_phead_in(&phead, ph_gl->gl_lvb);

	if (phead.ph_magic != GFS_PLOCK_PHEAD_MAGIC)
		init_plock_phead(ip, ph_gl, &phead);

	error = check_for_conflicts(ip, pp, &phead, start, end, ex);
	if (error) {
		if (wait) {
			gfs_glock_dq_uninit(&ph_gh);
			gfs_glock_dq_uninit(&j_gh);

			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(2 * HZ);

			if (signal_pending(current)) {
				error = -ERESTARTSYS;
				goto fail_unhold;
			}

			goto retry;
		}
		error = -EAGAIN;
		goto fail_unlock;
	}

	error = add_lock_range(ip, pp, &phead, start, end, ex, &before, &after);
	if (error)
		goto fail_unlock;

	gfs_glock_dq_uninit(&ph_gh);

	gfs_glock_dq_uninit(&j_gh);

	if (before) {
		gfs_unhold_lvb(ph_gl);
		gfs_put_glstruct(ph_gl);
	}

	put_pp(sdp, pp);

	return 0;

      fail_unlock:
	gfs_glock_dq_uninit(&ph_gh);
	gfs_glock_dq_uninit(&j_gh);

      fail_unhold:
	gfs_unhold_lvb(ph_gl);

      fail_glput:
	gfs_put_glstruct(ph_gl);

      fail:
	put_pp(sdp, pp);

	return error;
}
