/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#define POOL_RELEASE_NAME "v6.0.2"

#ifdef MODVERSIONS
#  ifndef __GENKSYMS__
#    include <linux/modversions.h>
#  endif			/* __GENKSYMS__ */
#endif				/*  MODVERSIONS  */

#include <linux/module.h>

#include <linux/init.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/vmalloc.h>
#include <linux/file.h>
#include <asm/uaccess.h>
#include <linux/smp_lock.h>
#include <linux/blk.h>
#include <linux/hdreg.h>
#include <linux/sched.h>
#include <linux/genhd.h>
#include <linux/devfs_fs_kernel.h>

#include "pool.h"

int mp_type = 1;
int mp_stripe = 0;
int pool_major=0;
MODULE_PARM(mp_type, "i");
MODULE_PARM(mp_stripe, "i");
MODULE_PARM(pool_major, "i");
MODULE_PARM_DESC(mp_type, "Multipath type: (none=0 | fail-over=1 "
		 "| Round Robin=2)");
MODULE_PARM_DESC(mp_stripe, "Stripe size across paths (in kB)");
MODULE_PARM_DESC(pool_major, "Major number to use for pool devices (use 0 for dynamic major)");
#define DEVICE_REQUEST do_pool_request

pool_t pools[POOL_MAX_DEVICES];
static devfs_handle_t pool_devfs_dir;
static devfs_handle_t pool_devfs_entries[POOL_MAX_DEVICES];

#define POOL_WRITE 1
#define POOL_READ  0

/* one of these structures exists for each open device */
typedef struct pool_open_device_s {
	uint32_t count;
	list_t list;
	kdev_t kdev;
	struct block_device *bdev;
} pod_t;

struct semaphore pod_sema;
list_t pod_list;

typedef struct pool_buffer_head_s {
	list_t pbh_list;
	int pbh_direction;
	kdev_t pbh_failed_device;	/* used only on I/O failure */
	int pbh_minor;		/* minor number of the pool */
	uint64_t pbh_start_time;
	struct buffer_head *pbh_original;
	struct buffer_head pbh_request;
} pool_buffer_head_t;

static spinlock_t pbh_reserve_lock;
static pool_buffer_head_t pbh_reserve;

typedef struct pool_thread_s {
	struct task_struct *pt_id;
	unsigned long pt_timer;
	atomic_t pt_run;
	list_t pt_requests;
	spinlock_t pt_reqlock;
	struct completion pt_completion;
} pool_thread_t;

static pool_thread_t pool_thread;

int pool_size[POOL_MAX_DEVICES];	/* in KB */
static struct hd_struct pool_hd_struct[POOL_MAX_DEVICES];
static int pool_hard_blksizes[POOL_MAX_DEVICES];
static int pool_soft_blksizes[POOL_MAX_DEVICES];
static int pool_maxreadahead[POOL_MAX_DEVICES];
static char *pool_names[POOL_MAX_DEVICES];

static void pool_geninit(struct gendisk *);
static int pool_open(struct inode *inode, struct file *file);
static int pool_release(struct inode *inode, struct file *file);
static int pool_map(kdev_t dev, kdev_t * rdev, unsigned long *rsector,
		    unsigned long size, int rw);
int pool_ioctl(struct inode *inode, struct file *file,
	       unsigned int cmd, unsigned long arg);

static struct block_device_operations pool_blk_dops = {
      open:pool_open,
      release:pool_release,
      ioctl:pool_ioctl,
};

static struct gendisk pool_gendisk = {
      major:0,			/* set in init_pool */
      major_name:"pool",
      minor_shift:0,
      max_p:1,
      part:pool_hd_struct,
      sizes:pool_size,
      nr_real:POOL_MAX_DEVICES,
      fops:&pool_blk_dops,
};

static void
pool_end_io(struct buffer_head *bh, int uptodate)
{
	int reissue = 0;
	int rw;
	int j, k, l;
	pool_t *pp = NULL;
	subpool_t *sp = NULL;
	device_t *dev = NULL;
	struct buffer_head *original_bh;
	pool_buffer_head_t *pool_bh;

	/* Note :: if the mapping function failed, bh->b_rdev will be zero **
	 ** and there will never be a device match, found will not be set   **
	 ** and the request will fail...................................... */
	pool_bh = (pool_buffer_head_t *) bh->b_private;
	original_bh = (struct buffer_head *) pool_bh->pbh_original;
	rw = pool_bh->pbh_direction;
	pp = &(pools[pool_bh->pbh_minor]);
	pool_bh->pbh_failed_device = 0;
	if (!uptodate) {
		printk("Pool: IO request to device, (%d,%d) blk #%" PRIu64
		       ", failed.\n", MAJOR(bh->b_rdev), MINOR(bh->b_rdev),
		       (uint64_t) bh->b_rsector);
		if (pp->p_mptype) {
			for (j = 0; j < pp->p_num_spools; j++) {
				sp = &(pp->p_spools[j]);
				for (k = 0; k < sp->sp_devs; k++) {
					dev = &(sp->sp_devlist[k]);
					for (l = 0; l < dev->d_num_paths; l++) {
						if (dev->d_dev[l] == bh->b_rdev) {
							if (l <
							    dev->
							    d_num_good_paths) {
								write_lock(&
									   (dev->
									    d_rwlock));
								printk
								    ("Pool: Faulty device found in:\n"
								     "\t Pool    :: %s\n"
								     "\t Subpool :: %d\n"
								     "\t Device  :: %d\n"
								     "\t Path    :: %d\n",
								     pp->p_name,
								     j, k, l);

								if (dev->
								    d_num_good_paths
								    > 1) {
									kdev_t
									    tmp_dev;
									printk
									    ("Pool: Mulitpathing update :: "
									     "Discarding faulty path (%d,%d).\n",
									     MAJOR
									     (dev->
									      d_dev
									      [l]),
									     MINOR
									     (dev->
									      d_dev
									      [l]));
									reissue++;
									tmp_dev
									    =
									    dev->
									    d_dev
									    [l];
									dev->
									    d_dev
									    [l]
									    =
									    dev->
									    d_dev
									    [dev->
									     d_num_good_paths
									     -
									     1];
									dev->
									    d_dev
									    [dev->
									     d_num_good_paths
									     -
									     1]
									    =
									    tmp_dev;
									dev->
									    d_num_good_paths--;
									pool_bh->
									    pbh_failed_device
									    =
									    tmp_dev;
									/* We must now break out of the loop to prevent the device **
									 ** from being detected again later (where we put it) in the**
									 ** array...................................................*/
									write_unlock
									    (&
									     (dev->
									      d_rwlock));
									break;
								} else {
									printk
									    ("Pool: No paths to fall back on...  Request failing.\n");
									write_unlock
									    (&
									     (dev->
									      d_rwlock));
								}
							} else {
								if (dev->
								    d_num_good_paths)
								{
									printk
									    ("Pool: I/O error on path that has been shut down... reissuing.\n");
									reissue++;
								}
							}
						}
					}
				}
			}
		}
		if (reissue) {
			/* remake the original request, only this time it should map to **
			 ** a different underlying device............................... */
			bh->b_rdev = original_bh->b_rdev;
			bh->b_rsector = original_bh->b_rsector;
			if (pool_map(bh->b_dev, &(bh->b_rdev), &(bh->b_rsector),
				     bh->b_size >> 9, rw) < 0) {
				pfree(pool_bh, sizeof (pool_buffer_head_t));
				buffer_IO_error(original_bh);
			} else {
				spin_lock(&(pool_thread.pt_reqlock));
				list_add(&(pool_bh->pbh_list),
					 &(pool_thread.pt_requests));
				wake_up_process(pool_thread.pt_id);
				spin_unlock(&(pool_thread.pt_reqlock));
			}
		} else {
#ifdef __SARD_AVAILABLE__
			uint64_t now = jiffies;
			struct hd_struct *hd =
			    &(pool_gendisk.part[pool_bh->pbh_minor]);

			hd->aveq +=
			    (hd->ios_in_flight * (now - hd->last_queue_change));
			hd->last_queue_change = now;

			if (hd->ios_in_flight)
				hd->io_ticks += (now - hd->last_idle_time);
			hd->last_idle_time = now;
			--hd->ios_in_flight;
#endif				/* __SARD_AVAILABLE__ */
			pfree(pool_bh, sizeof (pool_buffer_head_t));
			buffer_IO_error(original_bh);
		}
	} else {
#ifdef __SARD_AVAILABLE__
		/* finish iostats in hd_struct */
		uint64_t now = jiffies;
		struct hd_struct *hd = &(pool_gendisk.part[pool_bh->pbh_minor]);

		if (rw) {
			hd->wr_ticks += jiffies - pool_bh->pbh_start_time;
			hd->wr_ios++;
		} else {
			hd->rd_ticks += jiffies - pool_bh->pbh_start_time;
			hd->rd_ios++;
		}

		hd->aveq += (hd->ios_in_flight * (now - hd->last_queue_change));
		hd->last_queue_change = now;

		if (hd->ios_in_flight)
			hd->io_ticks += (now - hd->last_idle_time);

		hd->last_idle_time = now;
		--hd->ios_in_flight;
#endif				/* __SARD_AVAILABLE__ */
		original_bh->b_end_io(original_bh, uptodate);
		if (pool_bh != &pbh_reserve) {
			pfree(pool_bh, sizeof (pool_buffer_head_t));
		} else {
			spin_lock(&pbh_reserve_lock);
			memset(pool_bh, 0, sizeof (pool_buffer_head_t));
			spin_unlock(&pbh_reserve_lock);
		}
	}
}

/**
 * do_pool_request
 *
 * Stub routine to let Linux know we are a block device and we actually do
 * exist.
 */
static void
do_pool_request(request_queue_t * q)
{
	printk("Pool: request function called, this should never be called\n");
}

/**
 * pool_open
 * @inode
 * @file
 *
 * This gets called for every open.  We basically just ensure the call is
 * to a valid minor number and increment the open counts accordingly.  In
 * addition, if this is the first open on the device, all the lower level
 * device comprising the pool are sent open calls.  (This ensures once a
 * user has the pool open lower level driver modules are not unloaded.)  A
 * special case is the opening of minor device 0.  This is the poolcomm
 * device and is the only pool_t that accepts assemble and remvove ioctls.
 * It is the only minor number that always exists and is not dynamically
 * assigned.  In addtion, access to it is serialized at the open level.
 * (That means we don't let more than one process open it at the same
 * time.)  This prevents multiple copies of passemble from interfering.
 * Note that I believe it is still possible for a process to open poolcomm
 * and fork.  This essentially bypasses the precaution taken here.  DO NOT
 * do this. (cjs)  Only one open (locally) is allowed on a pool to 
 * prevent unaccounted cache blocks.
 *
 * Returns: Linux errno
 */
static int
pool_open(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);
	pool_t *pp;
	int error = 0;

	if (minor >= POOL_MAX_DEVICES) {
		error = -ENXIO;
		goto fail;
	}

	pp = &pools[minor];

	/* if this is an open on the comm device handle it specially. */

	if (minor == 0) {
		down_write(&(pp->p_rwsema));

		/* make sure the comm device exists */
		ASSERT((pp->p_flags & PF_FREE) == 0);

		/* check to see if device is busy */
		if (!(pp->p_flags & PF_VALID)) {
			up_write(&(pp->p_rwsema));
			error = -EBUSY;
			goto fail;
		}

		if (pp->p_oblk > 0) {
			up_write(&(pp->p_rwsema));
			error = -EBUSY;
			goto fail;
		}

		if (pp->p_oblk != -1) {
			pp->p_oblk++;
		} else {
			up_write(&(pp->p_rwsema));
			error = -ENXIO;
			goto fail;
		}

		up_write(&(pp->p_rwsema));
		MOD_INC_USE_COUNT;

		return (0);
	}

	/* regular open */

	down_write(&(pp->p_rwsema));

	if (pp->p_flags & PF_FREE || !(pp->p_flags & PF_LOWEROPEN)) {
		up_write(&(pp->p_rwsema));
		error = -ENXIO;
		goto fail;
	}

	if ((pp->p_flags & PF_READONLY) && (file->f_mode & FMODE_WRITE)) {
		up_write(&(pp->p_rwsema));
		error = -EROFS;
		goto fail;
	}

	ASSERT(pp->p_spools != NULL);

	pp->p_oblk++;

	up_write(&(pp->p_rwsema));
	MOD_INC_USE_COUNT;

	return (0);

      fail:

	return (error);
}

/**
 * pool_release
 * @inode
 * @file
 *
 * Returns: Linux errno
 */
static int
pool_release(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);
	pool_t *pp;

	pp = &pools[minor];

	down_write(&(pp->p_rwsema));

	ASSERT((pp->p_flags & PF_VALID) != 0);
	ASSERT((pp->p_flags & PF_FREE) == 0);
	ASSERT(pp->p_oblk > 0);

	pp->p_oblk--;

	up_write(&(pp->p_rwsema));
	MOD_DEC_USE_COUNT;

	return (0);
}

/**
 * pool_map
 * @dev
 * @rdev
 * @rsector
 * @size
 * @rw
 *
 * rsector and size are in basic blocks and are relative to the
 * logical pool device.  pool_blkmap() is called to translate the 
 * logical device and logical block into the actual device and block
 * for the request.
 *
 * Returns: rdev and rsector are changed to reflect the mapping, Linux
 * errno on error, 0 if OK.
 */
static int
pool_map(kdev_t dev, kdev_t * rdev, unsigned long *rsector,
	 unsigned long size, int rw)
{
	pool_t *pp = NULL;
	device_t *p_dev;
	int bc;
	uint64_t dev_sector;
	int error = 0;
	int minor = MINOR(dev);
	static uint32_t blk_count = 0;

	if (minor == 0) {	/* handle writes to poolcomm */
		printk("Pool: map: can not write to poolcomm.\n");
		error = -EPERM;
		goto fail;
	}

	pp = &pools[minor];

	down_read(&(pp->p_rwsema));

	if (pp->p_flags & PF_FREE || !(pp->p_flags & PF_VALID) ||
	    !(pp->p_flags & PF_LOWEROPEN)) {
		up_read(&(pp->p_rwsema));
		printk("Pool: map: No such minor device %d\n", minor);
		error = -EIO;
		goto fail;
	}

	/* Simple bound test */

	up_read(&(pp->p_rwsema));

	if ((*rsector + size) > pp->p_capacity) {
		down_write(&(pp->p_rwsema));
		if(pp->p_flags & PF_GROWING){
			up_write(&(pp->p_rwsema));
			while(1){
				up_read(&(pp->p_rwsema));
				if(!(pp->p_flags & PF_GROWING)){
					down_read(&(pp->p_rwsema));
					break;
				}
				down_read(&(pp->p_rwsema));
				schedule();
			}
		} else {
			pp->p_flags |= PF_GROWING;
			up_write(&(pp->p_rwsema));
			error = assemble_pools(pp->p_name, POOL_GROW);
			down_write(&(pp->p_rwsema));
			pp->p_flags &= ~PF_GROWING;
			up_write(&(pp->p_rwsema));
			if (error) {
				error = -ENOSPC;
				goto fail;
			}
		}
		if ((*rsector + size) > pp->p_capacity) {	/* recheck capacity */
			error = -ENOSPC;
			goto fail;
		}
	}

	if (pool_blkmap
	    (pp, *rsector, size << BBSHIFT, &dev_sector, &bc, &p_dev)) {
		error = -EIO;
		goto fail;
	}

	if (sizeof (unsigned long) <= 4 && dev_sector > UINT_MAX) {
		printk
		    ("Pool: map: attempted access beyond 32 bit sector limit.\n");
		error = -EIO;
		goto fail;
	}

	blk_count++;
	read_lock(&(p_dev->d_rwlock));
	if (pp->p_mptype > 1) {
		*rdev =
		    p_dev->d_dev[(blk_count / pp->p_mpstripe) %
				 p_dev->d_num_good_paths];
	} else {		/* No multipath and failover work the same here */
		*rdev = p_dev->d_dev[0];
	}
	read_unlock(&(p_dev->d_rwlock));
	*rsector = dev_sector;

	/* do statistics */
	spin_lock(&pp->p_statslock);

	if (rw == READ || rw == READA) {
		pp->p_stats.s_reads++;
		pp->p_stats.s_readblks += size;
	} else {
		pp->p_stats.s_writes++;
		pp->p_stats.s_writeblks += size;
	}

	spin_unlock(&pp->p_statslock);

	return (0);

      fail:
	/* clear the rdev field on error, for proper operation of pool_end_io */
	/*  *rdev = 0; */
	return (error);

}

/**
 * pool_make_request_fn
 * @q
 * @rw: read or write (use macros)
 * @bh: buffer_head to be remapped
 *
 * This funtion simply formats the arguments for pool_map.
 *
 * Returns: 0 on failure, 1 if OK.
 */
static int
pool_make_request_fn(request_queue_t * q, int rw, struct buffer_head *bh)
{
	pool_buffer_head_t *pool_bh;
	struct buffer_head *our_bh;
	int error = 0;

	pool_bh = kmalloc(sizeof (pool_buffer_head_t), GFP_NOIO);
	while (!pool_bh) {
		if (rw) {	/* push writes through special buffer if necessary */
			spin_lock(&pbh_reserve_lock);
			if (pbh_reserve.pbh_original) {	/* in use */
				spin_unlock(&pbh_reserve_lock);
			} else {
				pool_bh = &pbh_reserve;
				pool_bh->pbh_original = bh;
				spin_unlock(&pbh_reserve_lock);
				break;
			}
		}
		schedule();
		pool_bh = kmalloc(sizeof (pool_buffer_head_t), GFP_NOIO);
	}

	memcpy(&(pool_bh->pbh_request), bh, sizeof (struct buffer_head));

	if (rw) {
		pool_bh->pbh_direction = POOL_WRITE;
	} else {
		pool_bh->pbh_direction = POOL_READ;
	}
	pool_bh->pbh_original = bh;
	pool_bh->pbh_minor = MINOR(bh->b_dev);
	pool_bh->pbh_request.b_end_io = pool_end_io;
	pool_bh->pbh_request.b_private = (void *) pool_bh;

	our_bh = &(pool_bh->pbh_request);
	error = pool_map(our_bh->b_dev, &(our_bh->b_rdev), &(our_bh->b_rsector),
			 our_bh->b_size >> 9, rw);
	if (error < 0) {
		pfree(pool_bh, sizeof (pool_buffer_head_t));
		buffer_IO_error(bh);
		return 0;	/* failure, buffer_IO_error has been called, don't recurse */
	} else {
#ifdef __SARD_AVAILABLE__
		/* begin iostats in hd_struct */
		uint64_t now = jiffies;
		struct hd_struct *hd = &(pool_gendisk.part[pool_bh->pbh_minor]);

		pool_bh->pbh_start_time = now;
		if (rw)
			hd->wr_sectors += bh->b_size >> 9;	/* number of disk sectors */
		else
			hd->rd_sectors += bh->b_size >> 9;

		hd->aveq += (hd->ios_in_flight * (now - hd->last_queue_change));
		hd->last_queue_change = now;

		if (hd->ios_in_flight)
			hd->io_ticks += (now - hd->last_idle_time);

		hd->last_idle_time = now;
		++hd->ios_in_flight;
#endif
		generic_make_request(rw, our_bh);
		return 0;	/* all ok, mapping done, we'll call generic_make_request */
	}
}

/**
 * pool_ioctl
 * @arg
 * @size
 *
 * Implements all pool ioctl's.  This includes assemble/remove,
 * statistics gathering, and other miscellaneous ioctls.  Note there should
 * be only one retrun from this function to handle the return protocol
 * correctly.  If an ioctl causes an internal routine to be called, the
 * parameters are copied to kernel space before the call is made.  NOTE:
 * structures that are in user space and are pointed to by the structure
 * passed as the ioctl arguement are not copied into kernel space before the
 * call.  Some ioctls will only work on the /dev/poolcomm device (minor=0)
 *
 * Returns: 0 if OK, Linux errno on error.
 */
int
pool_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	   unsigned long arg)
{
	kdev_t dev = inode->i_rdev;
	int minor = MINOR(dev);
	pool_t *pp = NULL;
	int rc = 0;

	if (minor < 0 || minor >= POOL_MAX_DEVICES)
		return (-ENXIO);

	pp = &pools[minor];

	down_read(&(pp->p_rwsema));
	if (pp->p_flags & PF_FREE) {
		up_read(&(pp->p_rwsema));
		return (-ENXIO);
	}
	ASSERT(POOL_BUSY(pp));
	ASSERT(pp->p_flags & PF_VALID);
	up_read(&(pp->p_rwsema));

	/* Do commands */
	switch (cmd) {

	case POOL_CLEAR_STATS:
		{
			spin_lock(&pp->p_statslock);

			memset(&(pp->p_stats), 0, sizeof (pool_stat_t));

			spin_unlock(&pp->p_statslock);
			break;
		}

	case POOL_DUMP_TOTALSTATS:
		{
			pool_stat_t tmp_stat;

			spin_lock(&pp->p_statslock);
			memcpy(&tmp_stat, &(pp->p_stats), sizeof (pool_stat_t));
			spin_unlock(&pp->p_statslock);

			if (copy_to_user
			    ((void *) arg, &tmp_stat, sizeof (pool_stat_t)))
				rc = -EFAULT;

			break;
		}

	case BLKGETSIZE:	/*  Size in 512-byte blocks  */
		{
			unsigned long value = pp->p_capacity;

			if (!minor) {
				rc = -EINVAL;
				break;
			}
			if (!arg)
				return (-EINVAL);

			if (copy_to_user
			    ((void *) arg, &value, sizeof (unsigned long)))
				rc = -EFAULT;
			break;
		}

	case BLKGETSIZE64:	/*  Size in bytes  */
		{
			uint64_t value = pp->p_capacity << 9;

			if (!minor) {
				rc = -EINVAL;
				break;
			}
			if (!arg)
				return (-EINVAL);

			if (copy_to_user
			    ((void *) arg, &value, sizeof (uint64_t)))
				rc = -EFAULT;
			break;
		}

		/* Linux specific IOCTL's */

	case HDIO_GETGEO:
		printk("Pool: ioctl: HDIO_GETGEO\n");
		return -ENXIO;

	case BLKRASET:
		if (arg > 0xff) {
			rc = -EINVAL;
			break;
		}

		read_ahead[MAJOR(dev)] = arg;
		break;

	case BLKRAGET:

		copy_to_user((void *) arg, &read_ahead[MAJOR(dev)],
			     sizeof (long));
		break;

	case BLKFLSBUF:
		fsync_dev(dev);
		invalidate_buffers(dev);
		return 0;

	case POOL_ASSEMBLE:
		{
			char poolname[POOL_NAME_SIZE];
			int rtn = 0;

			if (minor) {
				return (-EINVAL);
			}
			if (copy_from_user
			    (poolname, (void *) arg, POOL_NAME_SIZE) != 0) {
				return (-EFAULT);
			}

			rtn = assemble_pools(poolname, POOL_ASSEMBLE);

			return rtn;

			break;
		}

	case POOL_REMOVE:
		{
			uint32_t m;	/* minor number of device to remove */

			if (minor) {
				return (-EINVAL);
			}

			if (copy_from_user((void *) &m, (void *) arg,
					   sizeof (uint32_t)) != 0) {
				rc = -EFAULT;
				break;
			}

			rc = remove_pool(pool_gendisk.major, m);
			break;
		}

	case POOL_GROW:
		{
			int rtn = 0;
			char poolname[POOL_NAME_SIZE];

			if (minor) {
				return (-EINVAL);
			}

			if (copy_from_user(poolname, (void *) arg,
					   POOL_NAME_SIZE) != 0) {
				return (-EFAULT);
			}

			rtn = assemble_pools(poolname, POOL_GROW);

			/* must update sizes */
			if (rtn >= 0) {
				pool_size[minor] = pools[minor].p_capacity >> 1;
				pool_gendisk.part[minor].nr_sects =
				    pools[minor].p_capacity >> 1;
				return 0;
			} else {
				return rtn;
			}
			break;
		}

	case POOL_COUNT:
		{
			int count = 0;

			rc = count_pool(&count);

			if (copy_to_user
			    ((void *) arg, (void *) &count,
			     sizeof (int)) != 0) {
				if (!rc)
					rc = -EFAULT;
			}

			break;
		}

	case POOL_LIST:
		{
			pool_list_t list;

			if (minor) {
				return (-EINVAL);
			}

			if (copy_from_user((void *) &list, (void *) arg,
					   sizeof (pool_list_t)) != 0) {
				return (-EFAULT);
			}

			rc = list_pool(pool_gendisk.major, &list);

			if (copy_to_user((void *) arg, (void *) &list,
					 sizeof (pool_list_t)) != 0) {
				if (!rc)
					rc = -EFAULT;
			}
			break;
		}

	case POOL_SPLIST:
		{
			pool_splist_t list;

			if (copy_from_user(&list, (void *) arg,
					   sizeof (list)) != 0) {
				return -EFAULT;
			}

			if (list.spl_count <= 0) {
				rc = -EFAULT;
				break;
			}

			rc = list_spool(pp, &list);

			if (copy_to_user((void *) arg, (void *) &list,
					 sizeof (list)) != 0) {
				if (!rc)
					rc = -EFAULT;
			}
			break;
		}
	case POOL_DEVLIST:
		{
			pool_devlist_t list;

			if (copy_from_user(&list, (void *) arg,
					   sizeof (list)) != 0) {
				return -EFAULT;
			}

			if (list.pdl_count <= 0) {
				rc = -EFAULT;
				break;
			}

			rc = list_pooldevs(pp, &list);

			if (copy_to_user((void *) arg, (void *) &list,
					 sizeof (list)) != 0) {
				if (!rc)
					rc = -EFAULT;
			}
			break;
		}
	case POOL_GET_VERSION:
		{
			uint64_t version = POOL_VERSION;
			if (copy_to_user
			    ((void *) arg, (void *) &version,
			     sizeof (uint64_t)) != 0) {
				rc = -EFAULT;
			}
			break;
		}
	case POOL_MP_RESTORE:
		{
			if (!minor) {
				rc = -EINVAL;
				break;
			}
			if (restore_pool_paths(pp)) {
				rc = -EFAULT;
			}

			break;
		}
	case POOL_TUNE_MP:
		{
			int i, tmp_stripe;
			if (!minor) {
				rc = -EINVAL;
				break;
			}
			if (copy_from_user
			    (&tmp_stripe, (void *) arg, sizeof (int)) != 0) {
				rc = -EFAULT;
			}

			if (tmp_stripe < 0) {
				down_write(&(pp->p_rwsema));

				/* Turn off multipathing */
				if (pp->p_mptype != 0)
					printk
					    ("Pool: Multipathing type changed to \"None\" for pool, %s\n",
					     pp->p_name);
				pp->p_mptype = 0;
				pp->p_mpstripe = -1;
				up_write(&(pp->p_rwsema));

			} else {
				down_write(&(pp->p_rwsema));
				/* We don't need to set p_mptype here because the
				   pool_update_multipath call does it for us later based on the
				   the p_mpstripe setting.
				 */
				if (!tmp_stripe) {
					/* Switch to Fail-over Multipathing */
					pp->p_mpstripe = tmp_stripe;
				} else {
					/* Switch to Round-Robin Multipathing */
					if (tmp_stripe <= 64) {
						tmp_stripe = 64;
					} else {
						for (i = 128;
						     i < 1 * 1024 * 1024;
						     i *= 2) {
							if (tmp_stripe <= i) {
								tmp_stripe = i;
								break;
							}
						}
						if (i == 1 * 1024 * 1024) {
							tmp_stripe = i;
						}
					}
					pp->p_mpstripe = tmp_stripe * 2;
				}

				/* this prints which mp type we switch to */
				/*      pool_update_multipath(rc, pp); */
				if (pp->p_mpstripe >= 0) {
					if (!pp->p_mptype) {
						printk
						    ("Pool: Enabling multipathing (%s) for pool, %s.\n",
						     (pp->
						      p_mpstripe) ?
						     "Round Robin" :
						     "Fail-over", pp->p_name);
					} else if (pp->p_mptype !=
						   ((pp->p_mpstripe) ? 2 : 1)) {
						printk
						    ("Pool: Switching to multipathing (%s) for pool, %s.\n",
						     (pp->
						      p_mpstripe) ?
						     "Round Robin" :
						     "Fail-over", pp->p_name);
					}
					pp->p_mptype = (pp->p_mpstripe) ? 2 : 1;
				}

				/* so we only need to print the stripe size here */
				if (!rc)
					switch (tmp_stripe) {
					case 0:
						break;
					case 64:
						printk
						    ("Pool: Multipathing stripe :: 64 kB - the minumum.\n");
						break;
					case 1048576:
						printk
						    ("Pool: Multipathing stripe :: 1 GB - the maximum.\n");
						break;
					default:
						printk
						    ("Pool: Multipathing stripe :: %d kB.\n",
						     tmp_stripe);
						break;
					}
				up_write(&(pp->p_rwsema));

			}

			break;
		}
	default:
		rc = -EINVAL;
	}

	return (rc);
}

static void
pool_geninit(struct gendisk *gdisk)
{
	int i;
	for (i = 0; i < POOL_MAX_DEVICES; i++) {
		pool_gendisk.part[i].start_sect = -1;	/* avoid partition check */
		pool_gendisk.part[i].nr_sects = 0;
		pool_names[i] = NULL;
	}
}

static __inline__ void
do_daemonize(char *name)
{
	unsigned long flags;

	lock_kernel();

	daemonize();

	current->tty = NULL;
	current->tty_old_pgrp = 0;

	spin_lock_irqsave(&current->sighand->siglock, flags);
	flush_signals(current);
	sigfillset(&current->blocked);
	spin_unlock_irqrestore(&current->sighand->siglock, flags);

	sprintf(current->comm, name);

	unlock_kernel();
}

static void
check_pool_paths(void)
{
	int i;

	for (i = 1; i < POOL_MAX_DEVICES; i++) {
		restore_pool_paths(&(pools[i]));
	}
}

static void
pool_mpd(void)
{
	pool_buffer_head_t *req;

	do_daemonize("pool_mpd");
	pool_thread.pt_id = current;
	atomic_set(&pool_thread.pt_run, 1);
	INIT_LIST_HEAD(&pool_thread.pt_requests);
	spin_lock_init(&pool_thread.pt_reqlock);
	spin_lock(&pool_thread.pt_reqlock);
	pool_thread.pt_timer = jiffies;
	complete(&pool_thread.pt_completion);

	do {
		spin_unlock(&pool_thread.pt_reqlock);
		while (1) {
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(5 * HZ);
			spin_lock(&pool_thread.pt_reqlock);
			if (((jiffies - pool_thread.pt_timer) >= 86400 * HZ) ||
			    !list_empty(&pool_thread.pt_requests) ||
			    !atomic_read(&pool_thread.pt_run)) {
				break;
			} else {
				spin_unlock(&pool_thread.pt_reqlock);
			}
		}

		if (list_empty(&pool_thread.pt_requests) &&
		    atomic_read(&pool_thread.pt_run)) {
			spin_unlock(&pool_thread.pt_reqlock);
			check_pool_paths();
			pool_thread.pt_timer = jiffies;

			spin_lock(&pool_thread.pt_reqlock);
		}

		while (!list_empty(&pool_thread.pt_requests)) {
			req = list_entry(pool_thread.pt_requests.next,
					 pool_buffer_head_t, pbh_list);
			list_del(&req->pbh_list);
			spin_unlock(&pool_thread.pt_reqlock);
			if (req->pbh_failed_device)
			  bdclose(req->pbh_failed_device);
			generic_make_request(req->pbh_direction,
					     &req->pbh_request);
			spin_lock(&pool_thread.pt_reqlock);
		}
	} while (atomic_read(&pool_thread.pt_run));

	complete(&pool_thread.pt_completion);
	/* Undaemonize */
}

static int
pool_mpd_bounce(void *data)
{
	pool_mpd();
	return 0;
}

/**
 * init_pool
 *
 * This routine is called by init_module().  It does the 
 * important elements required for Linux block driver initialization
 * including registering itself as a block device in the kernel and
 * setting up the blk_dev[] function pointers.  Pool locks are also
 * initialized. 
 *
 * Returns: Linux errno
 */
int
init_pool(void)
{
	int i;
	int error;
	unsigned int major;
	int result;

	printk("Pool %s (built %s %s) installed\n",
	       POOL_RELEASE_NAME, __DATE__, __TIME__);

	if (mp_type != 1) {
		if ((mp_type != 0) && (mp_type != 2)) {
			mp_type = 1;	/* Put in Fail-over mode */
		}
		printk("Pool:\tMultipathing type   :: %s\n",
		       (mp_type == 0) ? "None" :
		       (mp_type ==
			1) ? "Single Path (Fail-over)" : "Round Robin");
		if (mp_type > 1) {
			if (mp_stripe <= 64) {
				mp_stripe = 64;
				printk
				    ("Pool:\tMultipathing stripe :: 64 kB - the minumum.\n");
			} else {
				for (i = 128; i < 1 * 1024 * 1024; i *= 2) {
					if (mp_stripe <= i) {
						printk
						    ("Pool:\tMultipathing stripe :: %d kB.\n",
						     i);
						mp_stripe = i;
						break;
					}
				}
				if (i == 1 * 1024 * 1024) {
					mp_stripe = i;
					printk
					    ("Pool:\tMultipathing stripe :: 1 GB - the maximum.\n");
				}
			}
		}
	}
	/* Since the math is really done on the basic block level */
	mp_stripe *= 2;

	/* Grab a major for pool */
	result = register_blkdev(pool_major, "pool", &pool_blk_dops);
	if (result < 0) {
		if(pool_major == 0)
			printk("Pool: Error - unable to get dynamic major number\n");
		else
			printk("Pool: Error - major number %d in use\n", pool_major);
		return (-1);
	}

	if(pool_major == 0)
		major = result;
	else
		major = pool_major;

	pool_gendisk.major = major;

	read_ahead[major] = vm_max_readahead;

	for (i = 0; i < POOL_MAX_DEVICES; i++) {
		pool_soft_blksizes[i] = BLOCK_SIZE;
		pool_hard_blksizes[i] = 0;
		pool_maxreadahead[i] = vm_max_readahead;
		pool_size[i] = 0;
	}

	hardsect_size[major] = pool_hard_blksizes;
	/* what is blksize_size used for ?? */
	blksize_size[major] = pool_soft_blksizes;

	/* Apparently, if I don't set this, the kernel assumes **
	 ** the largest value possible......................... **
	 blk_size[POOL_MAJOR]=pool_size;
	 */

	max_readahead[major] = pool_maxreadahead;

	/* initialize pools[] array */
	memset(pools, 0, POOL_MAX_DEVICES * sizeof (pool_t));

	for (i = 0; i < POOL_MAX_DEVICES; i++) {
		init_rwsem(&(pools[i].p_rwsema));
		spin_lock_init(&pools[i].p_statslock);
		pools[i].p_flags = PF_VALID | PF_FREE;
	}

	pools[0].p_flags = PF_VALID;

	/* In the 2.2 kernels, the pool_geninit function was linked into **
	 ** the gendisk structure and taken care of by the kernel.  Now   **
	 ** we must call this function on our own - it is no longer in the**
	 ** structure.................................................... */
	pool_geninit(&pool_gendisk);

	add_gendisk(&pool_gendisk);

	/* set up the devfs directory */
	pool_devfs_dir = devfs_mk_dir(NULL, "pool", NULL);
	pool_devfs_entries[0] =
	    devfs_register(NULL, "poolcomm",
			   DEVFS_FL_DEFAULT, major,
			   0, S_IFBLK | S_IRWXU, &pool_blk_dops, NULL);

	blk_init_queue(BLK_DEFAULT_QUEUE(major), DEVICE_REQUEST);
	blk_queue_make_request(BLK_DEFAULT_QUEUE(major), pool_make_request_fn);

	INIT_LIST_HEAD(&pod_list);
	init_MUTEX(&pod_sema);
	spin_lock_init(&pbh_reserve_lock);
	memset(&pbh_reserve, 0, sizeof (pool_buffer_head_t));

	/* start thread */
	init_completion(&pool_thread.pt_completion);
	error = kernel_thread(pool_mpd_bounce, NULL, 0);
	if (error < 0) {
		printk("Pool: Failed to start multipathing thread.\n");
		goto fail;
	}
	wait_for_completion(&pool_thread.pt_completion);

	return (0);

      fail:
	del_gendisk(&pool_gendisk);

	devfs_unregister(pool_devfs_entries[0]);	/* the poolcomm device */
	devfs_unregister(pool_devfs_dir);	/* the pool directory */

	unregister_blkdev(major, "pool");
	blk_cleanup_queue(BLK_DEFAULT_QUEUE(major));

	read_ahead[major] = 0;
	blk_size[major] = NULL;
	blksize_size[major] = NULL;
	return -1;
}

/**
 * bdopen - open a lower level block device
 * @dev
 *
 * Returns: 0 on success, 1 on read-only, <0 on failure
 */
int
bdopen(kdev_t dev)
{
	pool_t *pp;
	pod_t *pod;
	list_t *tmp;
	int readonly=0;
	int error = 0;

	if (MAJOR(dev) == pool_gendisk.major) {

		pp = &pools[MINOR(dev)];

		down_write(&(pp->p_rwsema));

		if (pp->p_flags & PF_FREE || !(pp->p_flags & PF_LOWEROPEN)) {
			up_write(&(pp->p_rwsema));
			printk("Pool: bdopen of pool - error 1\n");
			return -ENXIO;
		}

		if (!(pp->p_flags & PF_VALID)) {
			up_write(&(pp->p_rwsema));
			printk("Pool: bdopen of pool - error 2\n");
			return -EBUSY;
		}

		ASSERT(pp->p_spools != NULL);

		pp->p_oblk++;

		up_write(&(pp->p_rwsema));
		MOD_INC_USE_COUNT;

		return (0);
	}

	down(&pod_sema);
	list_for_each(tmp, &pod_list) {
		pod = list_entry(tmp, pod_t, list);
		if (pod->kdev == dev) {
			pod->count++;
			up(&pod_sema);
			return 0;
		}
	}

	pod = pmalloc(sizeof (pod_t));
	if (!pod) {
		up(&pod_sema);
		return -ENOMEM;
	}
	pod->count = 1;
	pod->kdev = dev;
	pod->bdev = bdget(kdev_t_to_nr(dev));
	if (!pod->bdev) {
		pfree(pod, sizeof (pod_t));
		up(&pod_sema);
		return -ENOMEM;
	}
	error = blkdev_get(pod->bdev, FMODE_READ | FMODE_WRITE, 0, BDEV_FILE);
	if(error == -EROFS){
	  readonly = 1;
	  pod->bdev = bdget(kdev_t_to_nr(dev));
	  if (!pod->bdev) {
	    pfree(pod, sizeof (pod_t));
	    up(&pod_sema);
	    return -ENOMEM;
	  }
	  error = blkdev_get(pod->bdev, FMODE_READ, 0, BDEV_FILE);
	}
	if (error) {
		/*bdput(pod->bdev); blkdev_get calls this if it fails */
		pfree(pod, sizeof (pod_t));
		up(&pod_sema);
		return error;
	}

	list_add(&(pod->list), &pod_list);
	up(&pod_sema);

	return readonly;
}

/**
 * bdclose - closes a lower level block device
 * @dev
 *
 * Returns: Linux errno
 */

int
bdclose(kdev_t dev)
{
	pool_t *pp;
	pod_t *pod;
	list_t *tmp;

	if (MAJOR(dev) == pool_gendisk.major) {
		pp = &pools[MINOR(dev)];

		/* locking must be the responsibility of a higher layer 
		   down_write(&(pp->p_rwsema)); */

		ASSERT((pp->p_flags & PF_FREE) == 0);
		ASSERT(pp->p_oblk > 0);

		if (pp->p_oblk > 0)
			pp->p_oblk--;

		/* locking must be the responsibility of a higher layer 
		   up_write(&(pp->p_rwsema)); */
		MOD_DEC_USE_COUNT;

		return (0);
	}

	down(&pod_sema);
	pod = NULL;
	list_for_each(tmp, &pod_list) {
		pod = list_entry(tmp, pod_t, list);
		if (pod->kdev == dev) {
			pod->count--;
			break;
		}
		pod = NULL;
	}

	if (!pod) {		/* multiple closes are ok */
		up(&pod_sema);
		return 0;
	}
	ASSERT(pod->count >= 0);

	if (pod->count) {
		up(&pod_sema);
		return 0;
	}

	blkdev_put(pod->bdev, BDEV_FILE);
	list_del(&(pod->list));
	pfree(pod, sizeof (pod_t));
	up(&pod_sema);

	return 0;
}

/*
 * get_device_sector_size - an abstraction for get_hardsect_size
 * @device: the device to get the sector size for
 *
 * Returns: the sector size
 */
int
get_device_sector_size(kdev_t device)
{
	return (int) get_hardsect_size(device);
}

/**
 * read_partition_list - reads devices from /proc
 * @ptr
 *
 * This routine will place all of the devices known in the gendisk
 * list into an array which will be malloced and pointed to by *ptr.
 * The number of the devices is what is returned.
 *
 * Returns: the number of devices found.
 */
int
read_partition_list(kdev_t ** ptr)
{
	struct gendisk *gp = NULL;
	kdev_t *array;
	int i, j, n, num_devs = 0;
	int device_change = 1;

	while (device_change) {
		device_change = 0;
		for (i = 0; i < MAX_BLKDEV; i++) {
			gp = get_gendisk(MKDEV(i, 0));
			if (!gp)
				continue;
			if (gp->major != pool_gendisk.major) {
				for (n = 0;
				     n < (gp->nr_real << gp->minor_shift);
				     n++) {
					if (gp->part[n].nr_sects) {
						num_devs++;
					}
				}
			}
		}

		if (!(*ptr = pmalloc(num_devs * sizeof (kdev_t)))) {
			return -ENOMEM;
		}
		array = *ptr;

		for (j = 0, i = 0; i < MAX_BLKDEV; i++) {
			gp = get_gendisk(MKDEV(i, 0));
			if (!gp)
				continue;
			if (gp->major != pool_gendisk.major) {
				for (n = 0;
				     n < (gp->nr_real << gp->minor_shift);
				     n++) {
					if (gp->part[n].nr_sects) {
						if (j >= num_devs) {
							device_change = 1;
						} else {
							array[j++] =
							    MKDEV(gp->major, n);
						}
					}
				}
			}
		}
		if (j < num_devs) {
			device_change = 1;
		}
		if (device_change) {
			pfree(array, num_devs * sizeof (kdev_t));
		}
	}
	return num_devs;
}

/**
 * scan_dev_for_pool_label - read a pool label if on the device
 * @dev
 * @label
 *
 * This routine will scan a specific device for a pool label.  If found,
 * the label will be copied into the space pointed to by (the apply named)
 * "label".
 *
 * Returns: 1 = success, 0 = no pool label, -EXXX on error
 */
int
scan_dev_for_pool_label(kdev_t dev, pool_label_t * label)
{
	int rtn = 0;
	unsigned int blksize;
	struct buffer_head *bh = NULL;

	if (bdopen(dev) < 0) {
		return 0;
	} else {
		blksize = blksize_size[MAJOR(dev)][MINOR(dev)];
		bh = getblk(dev, 0, blksize);
		mark_buffer_uptodate(bh, 0);
		ll_rw_block(READ, 1, &bh);
		wait_on_buffer(bh);
		if (!buffer_uptodate(bh)) {
			brelse(bh);
			bh = NULL;
			rtn = -EIO;
		}
	}

	if (bh) {
		pool_label_in(label, bh->b_data);

		if (label->pl_magic == POOL_MAGIC) {
			rtn = 1;
		} else {
			rtn = 0;
		}

		brelse(bh);
	}
	bdclose(dev);
	return rtn;
}

/* WARNING -- the way false checking is done now could potentially **
** fuck up forward compatibility by writing information beyond a   **
** label, which may be assumed to be zeroed in future updates!     */
extern int *blksize_size[];
int
test_for_false_multipathing(kdev_t first, kdev_t second)
{
	int error = 0;
	pool_label_t label;
	struct buffer_head *bh = NULL;
	uint32_t blksize = 0;

	return 0;		/* attention -- for now, say no false multipathing */
	blksize = blksize_size[MAJOR(first)][MINOR(first)];
	if (!blksize) {
		error = -EIO;
		goto fail;
	}
	bh = getblk(first, 0, blksize);
	mark_buffer_uptodate(bh, 0);

	ll_rw_block(READ, 1, &bh);
	wait_on_buffer(bh);
	if (!buffer_uptodate(bh)) {
		brelse(bh);
		bh = NULL;
		error = -EIO;
		goto fail;
	}
	if (bh) {
		pool_label_in(&label, bh->b_data);
		if (label.pl_magic != POOL_MAGIC) {
			brelse(bh);
			error = -EIO;	/* This device should have label, but doesn't */
			goto fail;
		} else {
			strcpy((bh->b_data) + sizeof (pool_label_t), "brassow");
			/* mark as dirty */
			ll_rw_block(WRITE, 1, &bh);
			wait_on_buffer(bh);
			brelse(bh);
			bh = NULL;
		}
	} else {
		error = -EIO;
		goto fail;
	}

  /** New information has been written to 1st device, now read from 2nd **/

	blksize = blksize_size[MAJOR(second)][MINOR(second)];
	bh = getblk(second, 0, blksize);
	mark_buffer_uptodate(bh, 0);

	ll_rw_block(READ, 1, &bh);
	wait_on_buffer(bh);
	if (!buffer_uptodate(bh)) {
		brelse(bh);
		bh = NULL;
		error = -EIO;
		goto fail;
	}
	if (bh) {
		pool_label_in(&label, bh->b_data);
		if (label.pl_magic != POOL_MAGIC) {
			brelse(bh);
			error = -EIO;	/* This device should have label, but doesn't */
			goto fail;
		} else
		    if (!strncmp
			((bh->b_data) + sizeof (pool_label_t), "brassow", 8)) {
			brelse(bh);
			error = 0;	/* true multipathing has been found */
		} else {
			brelse(bh);
			error = 1;	/* false multipathing has been detected */
		}
	}

      fail:
	return error;
}

/**
 * register_pool - registers pool with Linux /proc
 * @new: the new pool to be registered
 * @minor: it's corresponding minor number
 *
 * This function provides the Linux /proc system with info
 * on size of the device and it's name.
 *
 * Returns: Linux errno
 */
extern int *hardsect_size[MAX_BLKDEV];
int
register_pool(pool_t * new, int minor)
{
	MOD_INC_USE_COUNT;
	pool_hard_blksizes[minor] = new->p_blksize;
	if (pool_soft_blksizes[minor] < pool_hard_blksizes[minor]) {
		pool_soft_blksizes[minor] = pool_hard_blksizes[minor];
	}

	/* set up gendisk structure, which is used by /proc */
	pool_gendisk.part[minor].start_sect = 0;

	/* this is reported in 1k block sizes */
	pool_gendisk.part[minor].nr_sects = new->p_capacity >> 1;

	if (!(pool_names[minor] =
	      pmalloc(strlen("pool/") + strlen(new->p_name) + 1))) {
		return -ENOMEM;
	}
	sprintf(pool_names[minor], "pool/%s", new->p_name);

	/* make entry in devfs, if necessary */
	pool_devfs_entries[minor] =
	    devfs_register(pool_devfs_dir, new->p_name,
			   DEVFS_FL_DEFAULT, pool_gendisk.major,
			   minor, S_IFBLK | S_IRWXU, &pool_blk_dops, NULL);

	return 0;		/* success */
}

/**
 * unregister_pool - registers pool with Linux /proc
 * @minor: minor number of pool to be removed
 *
 * This function has the opposite effect as register_pool.
 *
 * Returns: Linux errno
 */
int
unregister_pool(int minor)
{
	if (pool_names[minor] != NULL) {
		MOD_DEC_USE_COUNT;
		pool_hard_blksizes[minor] = 0;
		pool_gendisk.part[minor].start_sect = 0;
		pool_gendisk.part[minor].nr_sects = 0;
		devfs_unregister(pool_devfs_entries[minor]);
		pfree(pool_names[minor], strlen(pool_names[minor]));
		pool_names[minor] = NULL;
	}
	return 0;		/* success */
}

/**
 * cleanup_pool - does all the necessary things to remove pool from the kernel
 */
void
cleanup_pool(void)
{
	int i, j, k, l;
	for (i = 0; i < POOL_MAX_DEVICES; i++) {
		down_write(&(pools[i].p_rwsema));
		if (pools[i].p_flags & PF_FREE) {
			up_write(&(pools[i].p_rwsema));
			continue;
		}

		/* sync operations to flush any dirty buffers */

		for (j = 0; j < pools[i].p_num_spools; j++) {
			for (k = 0; k < pools[i].p_spools[j].sp_devs; k++)
				for (l = 0;
				     l <
				     pools[i].p_spools[j].sp_devlist[k].
				     d_num_paths; l++) {
					read_lock(&
						  (pools[i].p_spools[j].
						   sp_devlist[k].d_rwlock));
					fsync_dev(pools[i].p_spools[j].
						  sp_devlist[k].d_dev[l]);
					read_unlock(&
						    (pools[i].p_spools[j].
						     sp_devlist[k].d_rwlock));
				}
		}

		if (!(pools[i].p_flags & PF_VALID)) {
			/* attention -- return error here if no higher level check exists */
			printk
			    ("Pool: ARRGG unloading module while still in use\n");
		}

		if (pools[i].p_flags & PF_LOWEROPEN) {
			close_devices(&pools[i]);
		}

		free_pool(&pools[i]);
		pools[i].p_flags = PF_VALID | PF_FREE;
		up_write(&(pools[i].p_rwsema));
	}

	del_gendisk(&pool_gendisk);

	devfs_unregister(pool_devfs_entries[0]);	/* the poolcomm device */
	devfs_unregister(pool_devfs_dir);	/* the pool directory */

	unregister_blkdev(pool_gendisk.major, "pool");
	blk_cleanup_queue(BLK_DEFAULT_QUEUE(pool_gendisk.major));

	/* shutdown thread */
	spin_lock(&pool_thread.pt_reqlock);
	atomic_set(&pool_thread.pt_run, 0);
	wake_up_process(pool_thread.pt_id);
	spin_unlock(&pool_thread.pt_reqlock);
	wait_for_completion(&(pool_thread.pt_completion));

	read_ahead[pool_gendisk.major] = 0;
	blk_size[pool_gendisk.major] = NULL;
	blksize_size[pool_gendisk.major] = NULL;
}

module_init(init_pool);
module_exit(cleanup_pool);

MODULE_DESCRIPTION("Pool Volume Manager " POOL_RELEASE_NAME);
MODULE_AUTHOR("Red Hat, Inc.");
MODULE_LICENSE("GPL");
