/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

/* kernel.c - linux kernel interface to pool */

#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <dirent.h>
#include <linux/kdev_t.h>
#include <ctype.h>
#include <limits.h>

#include "sisstring.h"
#include "sislogger.h"
#include "osi_user.h"
#include "pool_ioctl.h"
#include "libpool.h"
#include "libpool_internal.h"
#include "global.h"
#include "iddev.h"

/* the comm device should only be used by kernel-related functions */
static int _commfd = -1;
static pool_list_t *_pl_cache = NULL;
static int _pl_cache_dirty = 1;
static char *match_me = NULL;
static int pool_major = 0;

/* helper fxn for scandir in add_devs */
static int select_entry(const struct dirent *entry)
{
   if (strstr(entry->d_name, match_me))
      return TRUE;
   return FALSE;
}

/**
 * add_devs:  add devices in given dir to the label list
 * labels: list to add devices to
 * dir: directory to scan for block devices
 * match: optional name to match in directory
 *
 */
static void add_devs(osi_list_t *labels, char *dir, char *match)
{
   char devpath[128];
   struct label_list *label;
   struct dirent **entries;
   struct stat statbuf;
   int count;

   if (match)
      match_me = strdup(match);

   count = scandir(dir, &entries, match ? select_entry : NULL, alphasort);

   if (count > 0) {
      while (count--) {

         memset(devpath, 0, 128);
         strcpy(devpath, dir);
         strcat(devpath, entries[count]->d_name);

         if (stat(devpath, &statbuf))
            continue;

         if (!S_ISBLK(statbuf.st_mode))
            continue;

         if (get_pool_label(devpath, &label))
            continue;

         label->ll_major = MAJOR(statbuf.st_rdev);
         label->ll_minor = MINOR(statbuf.st_rdev);

         osi_list_add(&(label->ll_list), labels);

         free(entries[count]);
      }
      free(entries);
   }

   if (match)
      free(match_me);
}

static void add_special_devs(osi_list_t *labels)
{
   add_devs(labels, "/dev/gnbd/", NULL);
   add_devs(labels, "/dev/pool/", NULL);
   /* add_devs(labels, "/dev/", "dasd"); */
}

static int ignore_special_devs(char *name)
{
   if (strstr(name, "gnbd") || strstr(name, "pool"))
      return TRUE;
   return FALSE;
}


/**
 * _get_pool_count - gets the number of pools currently loaded
 * @pool_count: the variable to insert the count into
 * @emsg: storage for error message if error occurs
 *
 * Returns: positive if the comm device is not opened, negative if there is
 *          an ioctl error, and zero on success
 *
 * Scope: static
 */
static int _get_pool_count(uint32 *pool_count, struct sstring *emsg) {
  int error = 0;

  if (ioctl(_commfd, POOL_COUNT, pool_count) < 0){
    ss_sprintf(emsg, "%s(%d): Unable to retrieve pool count: %s",
               __FILE__, __LINE__, strerror(errno));
    error = errno;
  }

  return error;
}


/**
 * _calc_link_name - calculate the hard link name based on the minor number
 * @link_name: pre-allocated character string for link name
 * @minor: minor number of device to calculate hard link for
 *
 * Since the linux kernel is stupid, and won't accept longer device node
 * names, (ie, /dev/pool/poolname), we need to create hard links to the real
 * pool device with names that are displayed in /proc/partitions.
 *
 * Scope: static
 */
inline static void _calc_link_name(char *link_name, uint32 minor)
{
   strcpy(link_name, "/dev/pool");
   if(minor < 26){
      link_name[9] = 'a'+minor;
   } else {
      link_name[9] = 'a'+(minor/26 - 1);
      link_name[10]= 'a'+(minor%26);
   }
}

/**
 * _check_pool_dir - verify existance and permissions of pool directory
 * @emsg: storage for error message if error occurs
 *
 */
static int _check_pool_dir(struct sstring *emsg) {
   int error = 0;
   struct stat stat_buf;

   if(stat( "/dev/pool", &stat_buf) < 0) {
     if(errno == ENOENT) {
       if((mkdir("/dev/pool", 0700)) != 0) {
          ss_sprintf(emsg, "%s(%d): Error creating /dev/pool: %s",
                     __FILE__, __LINE__, strerror(errno));
          error = errno;
          goto end;
       }
     }
     else {
       ss_sprintf(emsg, "%s(%d): Error accessing /dev/pool: %s",
                  __FILE__, __LINE__, strerror(errno));
       error = errno;
       goto end;
     }
   }
end:
   return error;
}

/**
 * _check_block_dev - check that the pool blockdev is available
 * @name: name of pool to check for
 * @emsg: storage for error message if error occurs
 *
 */
static int _check_block_dev(char *name, struct sstring *emsg) {
   int error = 0;
   char tmp_name[POOL_NAME_SIZE+10];
   struct stat stat_buf;

   if((error = _check_pool_dir(emsg)) != 0) {
      goto end;
   }

   memset(tmp_name, 0, POOL_NAME_SIZE+10);
   strcpy(tmp_name, "/dev/pool/");
   strcat(tmp_name, name);

   if(stat(tmp_name, &stat_buf) < 0){
      if(errno == ENOENT) {
         error = 0;
         goto end;
      }
      else {
         ss_sprintf(emsg, "%s(%d): Error accessing %s: %s", __FILE__, __LINE__,
                    tmp_name, strerror(errno));
         error = errno;
         goto end;
      }
   }
   if(!S_ISBLK(stat_buf.st_mode)) {
      /* FIXME: need a better error code here */
      ss_sprintf(emsg, "%s(%d): Error: %s is not a block device",
                 __FILE__, __LINE__, tmp_name);
      error = STD_ERROR;
      goto end;
   }

   if(unlink(tmp_name) != 0) {
      ss_sprintf(emsg, "%s(%d): Unable to unlink %s: %s", __FILE__, __LINE__,
                 tmp_name, strerror(errno));
      error = errno;
      goto end;
   }

end:
   return error;
}

/**
 * _make_block_dev - create a pool blockdev
 * @name: name of pool to check for
 * @emsg: storage for error message if error occurs
 *
 *
 */
static int _make_block_dev(char *name, struct sstring *emsg) {
   int  error = 0,
        j;
   char tmp_name[POOL_NAME_SIZE+10];
   struct stat stat_buf;

   if((error = _check_pool_dir(emsg)) != 0) {
     goto end;
   }
   if(_pl_cache->pl_count != 0){
      for(j=0; j < _pl_cache->pl_count; j++){
         if(strcmp(name,_pl_cache->pl_list[j].ps_name)) {
            continue;
         }
         memset(tmp_name, 0, POOL_NAME_SIZE+10);
         strcpy(tmp_name, "/dev/pool/");
         strcat(tmp_name, _pl_cache->pl_list[j].ps_name);
         if(stat(tmp_name, &stat_buf) < 0){
            if(errno == ENOENT) {
               char link_name[20]; memset(link_name, 0, 20);
               mknod(tmp_name, S_IFBLK | S_IREAD | S_IWRITE,
                     osi_make_dev(_pl_cache->pl_list[j].ps_pool_major,
                        _pl_cache->pl_list[j].ps_pool_minor));

               /* --------------> /proc is not reporting the pool devices to be
                * pool/<assigned device name>, but as pool[bcde...].
                * Therefore, we must link those names (pool[bcd...]) in the
                * /dev/ directory to the assigned device names in
                * /dev/pool/
                * ....................
                * */

               _calc_link_name(link_name, _pl_cache->pl_list[j].ps_pool_minor);
               link(tmp_name, link_name);
            }
            else {
               error = errno;
            }
         }
      }
   }
   else {
      /* FIXME: error? */
      error = 1;
   }

end:
   return error;
}

/**
 * _invalidate_pl_cache - wipes out the pool cache
 *
 * FIXME: This should be a real cache at some point - so i shouldn't have
 * to wipe the entire thing every time, but only the parts that have changed
 *
 * Scope: static
 */
static void _invalidate_pl_cache(void)
{
   int i,
       j;

   if(_pl_cache != NULL) {
      if(_pl_cache->pl_list != NULL) {
         for(i = 0; i < _pl_cache->pl_count; i++) {
            if(_pl_cache->pl_list[i].ps_splist != NULL) {
               for(j = 0; j < _pl_cache->pl_list[i].ps_num_spools; j++) {
                  if(_pl_cache->pl_list[i].ps_splist[j].sps_devlist != NULL) {
                     free(_pl_cache->pl_list[i].ps_splist[j].sps_devlist);
                     _pl_cache->pl_list[i].ps_splist[j].sps_devlist = NULL;
                  }
               }
               free(_pl_cache->pl_list[i].ps_splist);
               _pl_cache->pl_list[i].ps_splist = NULL;
            }
         }
         free(_pl_cache->pl_list);
         _pl_cache->pl_list = NULL;
      }
      free(_pl_cache);
      _pl_cache = NULL;
   }

   return;

}

/**
 * _get_pool_list - fills in the pool cache with the basic pool info
 * @emsg: storage for error message if error occurs
 *
 * Returns: zero on success, non-zero on error
 *
 * Scope: static
 */
static int _get_pool_list(struct sstring *emsg)
{
   int error = 0;

   /* if the cache is dirty or non-existant, fill it in... */
   if(_pl_cache_dirty != 0 || _pl_cache == NULL) {
      _invalidate_pl_cache();

      _pl_cache = (pool_list_t *) malloc(sizeof(pool_list_t));
      if(_pl_cache == NULL) {
         ss_sprintf(emsg, "%s(%d): %s", __FILE__, __LINE__, strerror(errno));
         error = errno;
         goto end;
      }
      memset(_pl_cache, 0, sizeof(pool_list_t));

      if ((error = _get_pool_count(&_pl_cache->pl_count, emsg)) == 0) {
         log_debug("pool count %d", _pl_cache->pl_count);
         if(_pl_cache->pl_count <= 0) {
            goto end;
         }
         _pl_cache->pl_list = malloc(_pl_cache->pl_count *
                                     sizeof(pool_status_t));
         if(!(_pl_cache->pl_list)) {
            ss_sprintf(emsg, "%s(%d): %s", __FILE__, __LINE__,
                       strerror(errno));
            error = errno;
            free(_pl_cache);
            goto end;
         }
         memset(_pl_cache->pl_list, 0,
               sizeof(_pl_cache->pl_count * sizeof(pool_status_t)));

         if (ioctl(_commfd, POOL_LIST, _pl_cache) < 0) {
            ss_sprintf(emsg, "%s(%d): Error retrieving pool list: %s.\n",
                       __FILE__, __LINE__, strerror(errno));
            error = errno;
         }
      }
   }

end:
   return error;
}

/**
 * _get_full_pool - fills in the pool cache with all the pool info
 * @emsg: storage for error message if error occurs
 *
 * Returns: zero on success, non-zero on error
 *
 * Scope: static
 */
static int _get_full_pool(struct sstring *emsg)
{

   int              error = 0,
                    i;
   pool_status_t    *pstat = NULL;
   pool_splist_t    splist = {0};
   pool_devstatus_t *devstat = NULL;
   pool_devlist_t   total_devlist = {0};

   if((error = _get_pool_list(emsg)) != 0)
      goto end;

   /* Grab stats for each of the pools in the list */
   for(i=0; i < _pl_cache->pl_count; i++){
      int   j = 0;
      int   index = 0;
      int   poolfd = -1;
      char  poolpath[POOL_LINE_MAX] = {0};
      int   dev_count = 0;

      pstat = &(_pl_cache->pl_list[i]);

      if(!pstat->ps_splist) {
         strcpy(poolpath, "/dev/pool/");
         strcat(poolpath, pstat->ps_name);
         if ((poolfd = open(poolpath, O_RDONLY)) < 0)  {
            /* i could check for ENOENT & ENOTDIR here, but I don't see a good
               reason to do so at this point */
            if((error = _make_block_dev(pstat->ps_name, emsg)) != 0) {
               error = errno;
               goto end;
            }
            else if ((poolfd = open(poolpath, O_RDONLY)) < 0)  {
               ss_sprintf(emsg, "%s:%d - Couldn't open pool device "
                          " \"%s\": %s",
                          __FILE__, __LINE__, poolpath, strerror(errno));
               goto end;
            }
         }

         /* setup the subpool info */
         splist.spl_count = pstat->ps_num_spools;
         splist.spl_list = (pool_spstatus_t *)
            malloc(pstat->ps_num_spools * sizeof(pool_spstatus_t));
         if(splist.spl_list == NULL) {
            ss_sprintf(emsg, "%s(%d): %s", __FILE__, __LINE__, strerror(errno));
            error = errno;
            _invalidate_pl_cache();
            goto end;
         }
         memset(splist.spl_list, 0,
                splist.spl_count * sizeof(pool_spstatus_t));

         /* the list of subpools is what we care about long term */
         if (ioctl(poolfd, POOL_SPLIST, &splist) < 0){
            ss_sprintf(emsg, "%s(%d): Error retrieving subpool list for %s:"
                       " %s",
                       __FILE__, __LINE__, poolpath, strerror(errno));
            error = errno;
            goto end;
         }
         pstat->ps_splist = splist.spl_list;

         /* we need to scan the entire subpool list of the number of devices,
            since that's what the POOL_DEVLIST IOCTL passes back */
         for(j=0, total_devlist.pdl_count = 0; j < splist.spl_count; j++){
            total_devlist.pdl_count += splist.spl_list[j].sps_numdevs;
         }

         total_devlist.pdl_list = (pool_devstatus_t *)
            malloc(total_devlist.pdl_count * sizeof(pool_devstatus_t));
         if(total_devlist.pdl_list == NULL) {
            ss_sprintf(emsg, "%s(%d): %s", __FILE__, __LINE__,
                       strerror(errno));
            error = errno;
            _invalidate_pl_cache();
            goto end;
         }
         memset(total_devlist.pdl_list, 0,
               (total_devlist.pdl_count * sizeof(pool_devstatus_t)));

         if (ioctl(poolfd, POOL_DEVLIST, &total_devlist) < 0){
            ss_sprintf(emsg, "%s(%d): Error retrieving device list for %s:"
                       " %s", __FILE__, __LINE__, poolpath, strerror(errno));
            error = errno;
            goto end;
         }

         for(j=0; j < splist.spl_count; j++){
            dev_count = splist.spl_list[j].sps_numdevs;
            devstat = (pool_devstatus_t *)
               malloc(sizeof(pool_devstatus_t) * dev_count);
            if(devstat == NULL) {
               ss_sprintf(emsg, "%s(%d): %s", __FILE__, __LINE__,
                          strerror(errno));
               error = errno;
               _invalidate_pl_cache();
               goto end;
            }
            splist.spl_list[j].sps_devlist = devstat;
            memset(devstat, 0, (dev_count * sizeof(pool_devstatus_t)));
            memcpy(devstat, &(total_devlist.pdl_list[index]),
                  (dev_count * sizeof(pool_devstatus_t)));
            index += dev_count;
         }
         free(total_devlist.pdl_list);
      }
      close(poolfd);
   }

end:
   if(!error)
         _pl_cache_dirty = 0;


   return error;

}

/**
 * _get_pool_status - find the named pool and return its kernel structure
 * @name: name of pool to find
 * @pool: pointer to pool structure
 * @emsg: storage for error message if error occurs
 *
 * Returns: zero on success, non-zero on error
 *
 * Scope: static
 */
static int _get_pool_status(char *name, pool_status_t **pool,
                            struct sstring *emsg)
{
   int i = 0;
   int error = 0;

   if(_pl_cache_dirty || !_pl_cache) {
      _invalidate_pl_cache();
      if((error = _get_full_pool(emsg)) != 0) {
         goto end;
      }
   }
   log_debug("Looking for %s", name);
   for(i=0; i < _pl_cache->pl_count; i++) {

      if(!(strcmp(name, _pl_cache->pl_list[i].ps_name))) {
         *pool = &(_pl_cache->pl_list[i]);
         log_debug("Looking for %s, found %s", name, (*pool)->ps_name);
         goto end;
      }
   }
   log_debug("Pool %s not found by pool module", name);

end:
   return error;
}


/**
 * clean_pool_dir - get rid of unused pool device nodes
 *
 * Description: unlinks pool device nodes for pools that are no longer
 * assembled.
 *
 * Returns: zero on success, non-zero on error
 *
 * Scope: external
 */
int clean_pool_dir(struct sstring *emsg){
   int           error = 0;
   int           i;
   int           errcount = 0;
   char          path[POOL_LINE_MAX];
   char          link_name[20];
   DIR           *dir;
   struct dirent *file;
   struct stat   statbuf;

  if((dir = opendir("/dev/pool/")) == NULL){
    error = errno;
  }
  else {
     rewinddir(dir);
     while ((file = readdir(dir)) != NULL){
        if(!strcmp(".", file->d_name) || !strcmp("..", file->d_name)){
           continue;
        }
        strcpy(path, "/dev/pool/");
        strcat(path, file->d_name);

        if((error = _get_full_pool(emsg)) != 0) {
           break;
        }

        /* The cache must be valid */
        if(!_pl_cache){
           error = STD_ERROR;
           ss_sprintf(emsg, "%s(%d): Error accessing in-kernel pool cachelist:"
                      __FILE__, __LINE__, strerror(errno));
           break;
        }

        if(stat(path, &statbuf) == 0) {
           /* don't unlink anything but block devices */
           if((S_ISBLK(statbuf.st_mode)) != 1) {
              continue;
           }
           /* If the device isn't a pool device (major number does not equal
              pool_major) don't unlink it */
           if((uint32) MAJOR(statbuf.st_rdev) != pool_major)
              continue;

           /* If we match the minor number of the device node with a pool
            * that we found in the pool cache, it's loaded and the device
            * node should not be removed */
           for(i = 0; i < _pl_cache->pl_count; i++){
              if(_pl_cache->pl_list[i].ps_pool_minor ==
                 (uint32) MINOR(statbuf.st_rdev)){
                 break;
              }
           }
           if(i != _pl_cache->pl_count){ /* pool loaded, skip unlink */
              continue;
           }

           if(unlink(path) != 0) {
              if(errno != ENOENT && errno != ENOTDIR) {
                 errcount++;
                 error |= 1;
              }
           }

           /* remove the link from /dev/pool/pool# -> /dev/poolXX */
           memset(link_name, 0, 20);
           _calc_link_name(link_name, MINOR(statbuf.st_rdev));
           if(unlink(link_name) != 0) {
              if(errno != ENOENT && errno != ENOTDIR) {
                 errcount++;
                 error |= 1;
              }
           }
        }
        else {
           errcount++;
           error |= 1;
        }

     }
     closedir(dir);
  }

  if(error != 0 && errcount > 0) {
     ss_sprintf(emsg, "%d errors attempting to clean old device nodes from"
                " /dev/pool\nPlease verify the device nodes in /dev/pool",
                errcount);
  }
  return error;
}



/**
 * _make_pool_nodes - create the pool nodes based on the pool cache
 * @msg: storage for error message if error occurs
 *
 * iterates through the pool cache and creates nodes based on its
 * contents.  Does not refresh the cache.
 *
 * Returns: zero on success, non-zero on error
 *
 * FIXME: Error checking!
 *
 * FIXME: Should we check for the /dev/pool directory and the ability to
 *        write to it before we assemble a pool?
 *
 * Scope: static
 */
static int _make_pool_nodes(struct sstring *emsg)
{

   int    j,
          additions = 0,
          found = 0,
          error = 0;
   char   tmp_name[POOL_NAME_SIZE];
   struct stat stat_buf;

   if((error = _check_pool_dir(emsg)) != 0) {
     goto end;
   }
   if(_pl_cache->pl_count){
      for(j=0; j < _pl_cache->pl_count; j++){
         memset(tmp_name, 0, POOL_NAME_SIZE);
         strcpy(tmp_name, "/dev/pool/");
         strcat(tmp_name, _pl_cache->pl_list[j].ps_name);
         if(stat(tmp_name, &stat_buf) < 0){
            if(errno == ENOENT) {
               char link_name[20]; memset(link_name, 0, 20);
               mknod(tmp_name, S_IFBLK | S_IREAD | S_IWRITE,
                     osi_make_dev(_pl_cache->pl_list[j].ps_pool_major,
                        _pl_cache->pl_list[j].ps_pool_minor));

               /* --------------> /proc is not reporting the pool devices to be
                * pool/<assigned device name>, but as pool[bcde...].
                * Therefore, we must link those names (pool[bcd...]) in the
                * /dev/ directory to the assigned device names in
                * /dev/pool/
                * ....................
                * */

               _calc_link_name(link_name, _pl_cache->pl_list[j].ps_pool_minor);
               link(tmp_name, link_name);
               additions++;
            }
            found++;
         }
         /* FIXME: should i do something when the error isn't because of a
            no entry error? */
      }
   }
end:
   return 0;
}

/**
 * _get_devpath - find the device path based on a dev_t
 * @path: place to store the device path (not preallocated)
 * @dev: dev_t from kernel
 * @emsg: storage for error message if error occurs
 *
 * scans the labels of disks in /proc/partitions and tries to find the one
 * that matches the dev_t info.
 *
 * Returns: zero on success, non-zero on failure
 *
 * Scope: static
 */
int _get_devpath(char **path, dev_t dev, struct sstring *emsg)
{
   int          error = 1;
   int          major = MAJOR(dev),
                minor = MINOR(dev);
   int          found = 0;
   struct label_list *ll = NULL,  *current = NULL;
   osi_list_t   *tmp = NULL, *head = NULL;

   scan_labels(&ll, emsg);

   head = &(ll->ll_list);
   log_debug("Major: %d, Minor: %d", major, minor);
   osi_list_foreach(tmp, head) {
      current = osi_list_entry(tmp, struct label_list, ll_list);

      if(current->ll_major == major && current->ll_minor == minor) {
         found = 1;
         break;
      }
   }
   if(found) {
      *path = strdup(current->ll_dev_path);
      error = 0;
   }
   else {
      *path = strdup("DEVICE_NOT_FOUND");
      error = 0;
   }

   destroy_labels(ll);
   free(ll);
   return error;
}

/* Internal (to library) functions (Not in exported_symbols.sym) */

/**
 * set_pool_in_use - checks to see whether a given pool is in use.
 * @p: pool to check, in_use field gets set
 * @emsg: storage for error message if error occurs
 *
 * Default return value is 0 because if the pool can't be found, it can't
 * be in use.
 *
 * Returns: zero on success, one on can't find, > one on error
 *
 * Scope: Internal to libpool
 */
int set_pool_in_use(user_pool_t *p, struct sstring *emsg)
{
   int error = 0;
   pool_status_t *pool = NULL;

   p->in_use = 0;

   if((error = _get_pool_status(p->name, &pool, emsg)) == 0) {
      if(pool != NULL) {
         p->in_use = pool->ps_in_use;
         log_debug("In use: %d\n", pool->ps_in_use);
      }
      else {
         p->in_use = 0;
      }
   }

   return error;
}

/**
 * set_pool_assembled - checks for matching pool assembled in kernel
 * @p: pool to check
 * @assembled: variable to store result in
 * @emsg: error message structure
 *
 * Returns: 0 on success, non-zero on failure.
 *
 * Scope: Internal to libpool
 */
int set_pool_assembled(user_pool_t *p, int *assembled, struct sstring *emsg)
{
   int error = 0;
   pool_status_t *pool = NULL;

   if((error = _get_pool_status(p->name, &pool, emsg)) == 0) {
      if(pool != NULL) {
         *assembled = 1;
         log_debug("Pool %s assembled", p->name);
      }
      else {
         *assembled = 0;
      }
   }

   /* Pool module not loaded */
   if(error == EBADF) {
      error = 0;
      cstring2ss(emsg, "", 0);
      *assembled = 0;
   }

   return error;
}

/**
 * get_blk_size - get the number of blocks in a device
 * @path: device path
 * @blocks: store the number of blocks here
 * @emsg: error message structure
 *
 * gets the block size for a specified device
 *
 * Returns: zero on success, non-zero on error
 *
 * Scope: Internal to libpool
 */
int get_blk_size(char *path, uint64 *blocks, struct sstring *emsg)
{
   int error = 0;
   int fd = -1;

   if(blocks == NULL) {
      error = EFAULT;
      ss_sprintf(emsg, "passed null pointer to get_blk_size");
      goto end;
   }

   if((fd = open(path, O_RDONLY)) < 0) {
      ss_sprintf(emsg, "%s(%d): Unable to open %s for reading: %s",
            __FILE__, __LINE__, path, strerror(errno));
      error = errno;
      goto end;
   }

   if (device_size(fd, blocks) < 0)
   {
     ss_sprintf(emsg, "%s(%d): unable to retrieve size of %s: %s",
		__FILE__, __LINE__, path, strerror(errno));
     close(fd);
     error = errno;
     goto end;
   }

   *blocks >>= 9;

   if(*blocks == 0) {
      ss_sprintf(emsg, "%s(%d): %s is a zero-block block device",
            __FILE__, __LINE__, path);
      close(fd);
      error = STD_ERROR;
      goto end;
   }
   log_debug("Device %s has %"PRIu64" blocks\n", path, *blocks);
   close(fd);
end:
   return error;
}

static char *_create_procbuf(uint32 *pbsize, uint32 *count)
{
   int procfd = -1;
   char *pb = NULL;
   int tmpcount = 0;
   int tmpsize = 0;

   if((procfd = open("/proc/partitions", O_RDONLY)) == -1){
      log_err("Could not open /proc/partitions");
      goto end;
   }

   if(!(pb = (char *) malloc(sizeof(char) * (*pbsize)))) {
      log_err("Out of memory allocating buffer (%u) for /proc/partitions",
              *pbsize);
      goto end;
   }

   /* Read and realloc until we read less than we asked for */
   do {
      memset(pb, 0, sizeof(char) * (*pbsize));
      tmpsize = *pbsize;
      *count = 0;
      lseek(procfd, 0, SEEK_SET);
      while((tmpcount = read(procfd, pb+*count, tmpsize)) > 0) {
         tmpsize -= tmpcount;
         *count += tmpcount;
         if(tmpsize <= 0)
            break;
      }

      if(tmpcount < 0 && errno != EINTR) {
         free(pb);
         pb = NULL;
         goto end;
      }
      if(*count < *pbsize) {
         log_debug("count = %u, pbsize = %u", *count, *pbsize);
         /* check to see if there was more - if there was, reread the
          * whole thing in case we were interrupted by a signal - we
          * want to get all of /proc/partitions in a single read... */
         if((tmpcount = read(procfd, pb+*count, 1)) > 0) {
            log_debug("Read %u again", tmpcount);
            continue;
         }
         else
            break;
      }

      *pbsize *= 2;
      if(!(pb = (char *) realloc(pb, sizeof(char) * (*pbsize)))) {
         log_err("Out of memory reallocating buffer (%u) for /proc/partitions",
                 *pbsize);
         goto end;
      }

   } while(*pbsize < SSIZE_MAX);

  end:
   close(procfd);
   return pb;
}

/**
 * scan_labels - scans the disk and reads each label
 * @l: label list
 * @emsg: error message structure
 *
 * allocates and fills in the label list by scanning the disk
 *
 * Returns: zero on success, non-zero on error
 *
 * FIXME: What needs to happen for multipathing...do we want to detect paths
 *        here?
 *
 * Scope: Internal to libpool
 */
int scan_labels(struct label_list **l, struct sstring *emsg)
{
   int error = 0;
   int i = 0;
   char *pb = NULL, *curproc = NULL, *tmpproc = NULL;
   uint32 pbsize = 4096;
   uint32 count = 0;
   char  proc_line[POOL_LINE_MAX],
         maj[20],
         min[20],
         blocks[40],
         name[POOL_LINE_MAX],
         tmp_name[POOL_LINE_MAX];
   char	 tmp_str[POOL_LINE_MAX];

   static struct label_list *_ll_cache = NULL;
   struct label_list *label = NULL, *current = NULL;
   struct label_list *head = NULL;
   osi_list_t *tmp, *ohead;
   int   found_emcpower = 0;

   if(_ll_cache == NULL) {
      _ll_cache = head = (struct label_list*)malloc(sizeof(struct label_list));
   }
   else {
      goto cp_cache;
   }

   if(!head) {
      SETMSGSS(emsg, "Out of memory");
      error = ENOMEM;
      goto end;
   }

   memset(head, 0, sizeof(struct label_list));

   osi_list_init(&(head->ll_list));

   memset(name, 0, sizeof(name));
   memset(tmp_name, 0, sizeof(tmp_name));
   memset(proc_line, 0, sizeof(proc_line));

   if(!(pb = _create_procbuf(&pbsize, &count))) {
      error = 1;
      goto end;
   }

   curproc = pb;

   /* Skip the first two lines of /proc/partitions */
   for(i = 0; i < 2; i++) {
      /* FIXME: should i use strchr? */
      tmpproc = memchr(curproc, '\n', POOL_LINE_MAX);
      curproc = tmpproc+1;
   }

   /* Scan /proc/partitions for valid block devices */
   /* FIXME: this sucks - need a way to add block devices - of course this
    * would also mean fixing the pool kernel driver, so i dunno that it's gonna
    * be happening any time soon...*/
   while((tmpproc = memchr(curproc, '\n', POOL_LINE_MAX)) &&
         (tmpproc - pb <= count)) {
      strncpy(proc_line, curproc, tmpproc-curproc);

      sscanf(proc_line,"%s %s %s %s", maj, min, blocks, tmp_name);

      /* Need to handle emc powerpath devices specially - we don't
       * want to scan all of /dev if we don't have to */
      if (!strncmp(tmp_name, "emcpower", 8)) {
	 found_emcpower = 1;
         curproc = tmpproc + 1;
         memset(proc_line, 0, sizeof(proc_line));
         continue;
      }

      if (ignore_special_devs(tmp_name)) {
         curproc = tmpproc + 1;
         memset(proc_line, 0, sizeof(proc_line));
         continue;
      }

      if(!isalpha(tmp_name[0]) || !strcmp(tmp_name, name)){
         memset(tmp_name, 0, sizeof(tmp_name));
         curproc = tmpproc + 1;
         memset(proc_line, 0, sizeof(proc_line));
         continue;
      } else {
         memcpy(name, tmp_name, sizeof(tmp_name));
         memset(tmp_name, 0, sizeof(tmp_name));
      }

      memset(tmp_str, 0, sizeof(tmp_str));
      /* ignore the device name and calculate it from the minor number if it's
         a pool volume */
      if(atoi(maj) == pool_major) {
         char *tmppool;
         get_pool_alias(&tmppool, atoi(min), emsg);
         snprintf(tmp_str, sizeof(tmp_str), "%s", tmppool);
         destroy_pool_alias(tmppool);
      }
      else {
         snprintf(tmp_str, sizeof(tmp_str), "/dev/%s", name);
      }
      tmp_str[POOL_LINE_MAX-1] = '\0';

      error = get_pool_label(tmp_str, &label);
      if(error == 1){
         error = 0;
      }
      else if (error > 1) {
         stack;
         goto end;
      }
      label->ll_major = atoi(maj);
      label->ll_minor = atoi(min);
      osi_list_add_prev(&(label->ll_list), &(head->ll_list));
      curproc = tmpproc + 1;
      memset(proc_line, 0, sizeof(proc_line));
   }

   if(found_emcpower) {
      add_devs(&(head->ll_list), "/dev/", "emcpower");
   }

   add_special_devs(&(head->ll_list));

  cp_cache:
   *l = head = (struct label_list*) malloc(sizeof(struct label_list));
   if(!head) {
      SETMSGSS(emsg, "Out of memory");
      error = ENOMEM;
      goto end;
   }
   osi_list_init(&(head->ll_list));
   ohead = &(_ll_cache->ll_list);
   osi_list_foreach(tmp, ohead) {
      current = osi_list_entry(tmp, struct label_list, ll_list);
      label =  (struct label_list *) malloc(sizeof(struct label_list));
      memcpy(label, current, sizeof(struct label_list));
      label->ll_dev_path = strdup(current->ll_dev_path);
      if(!label->ll_dev_path) {
         SETMSGSS(emsg, "Out of memory");
         error = ENOMEM;
         goto end;
      }
      osi_list_add(&(label->ll_list), &(head->ll_list));
   }

  end:
   if(pb)
      free(pb);
   return error;
}




/*
 * get_major_nr - read the dynamic major pool is using from /proc/devices
 *
 */

int get_major_nr(struct sstring *emsg){
   char device[LINE_MAX];
   char buf[LINE_MAX];
   int major_nr;
   FILE *fp;
   fp = fopen("/proc/devices", "r");
   if (!fp){
      ss_sprintf(emsg, "could not open /proc/devices: %s\n", strerror(errno));
      return -ENOENT;
   }
   while(fgets(buf, LINE_MAX, fp) != NULL)
      if (sscanf(buf, "%d %s", &major_nr, device) == 2 &&
          strcmp(device, "pool") == 0)
         break;
   if(strcmp(device, "pool") == 0)
      return major_nr;
   ss_sprintf(emsg,
      "could not find pool registered in /proc/devices.  This\n"
      "probably means that you have not loaded the pool module.\n"
      );
   return -ENOENT;
}
/* External functions (In exported_symbols.sym) */


/**
 * open_comm - opens the pool communication device
 *
 * Returns: 0 on success, non-zero on failure.
 *
 * Scope: external
 */
int open_comm(struct sstring *emsg)
{
   int error = 0;
   struct stat s;

   if((pool_major = get_major_nr(emsg)) < 0)
      return -pool_major;

   if(stat(COMMDEV, &s) != 0)
      if(errno != ENOENT) {
         log_err("Unable to stat %s: %s", COMMDEV, strerror(errno));
         return errno;
      }

   if((!S_ISBLK(s.st_mode)) || (MAJOR(s.st_rdev) != pool_major)) {
      if(S_ISDIR(s.st_mode)) {
         log_err("%s is a directory. Move or remove it before"
                 " running the pool tools.", COMMDEV);
         return -1;
      }
      unlink(COMMDEV);
   }

   /* Only open the device if it's not already opened by the library */
   if(_commfd == -1) {
      if((_commfd = open(COMMDEV, O_RDONLY)) < 0){
         if(errno == ENOENT) {
            mknod(COMMDEV, S_IFBLK | 0600 , osi_make_dev(pool_major, 0));
            if((_commfd = open(COMMDEV, O_RDONLY)) < 0) {
               ss_sprintf(emsg, "Unable to open %s: %s",
                          COMMDEV, strerror(errno));
               error = errno;
            }
         }
         else {
            ss_sprintf(emsg, "Unable to open %s: %s",
                       COMMDEV, strerror(errno));
            error = errno;
         }
      }
   }

   return error;
}

/**
 * close_comm - closes the pool comm device
 *
 * Closes the library's comm dev file descriptor
 *
 * Scope: external
 */
int close_comm(struct sstring *emsg)
{
   int error = 0;

   if(_commfd != -1) {
      if(close(_commfd) < 0) {
         ss_sprintf(emsg, "Error closing %s: %s", COMMDEV, strerror(errno));
         error = errno;
      }
      else {
         _commfd = -1;
      }
   }

   /* free the pool cache structure */
   _invalidate_pl_cache();

   return error;
}

/**
 * check_version - verifies that the pool module matches the library
 * @emsg:
 *
 * Pool is versioned using a 64-bit number that is adjusted when the kernel
 * metadata is changed.  This checks the module version against a list of
 * version that this library is compiled against.
 *
 * Returns: zero if the versions are compatible, non-zero otherwise
 *
 * Scope: external
 */
int check_version(struct sstring *emsg)
{
   int    i;
   int    error = 0;
   uint64 ver = 0;
   uint64 valid_ver = 0;


   if(ioctl(_commfd, POOL_GET_VERSION, &ver) < 0) {
      ss_sprintf(emsg, "%s(%d): Error retrieving pool module version: %s",
            __FILE__, __LINE__, strerror(errno));
      error = errno;
      goto end;
   }

   error = STD_ERROR;
   ss_sprintf(emsg, "%s(%d): Pool module version %"PRIX64" incompatible"
              " with this version of libpool", __FILE__, __LINE__, ver);
   for(i = 0; (valid_ver = valid_module_vers[i]) && valid_ver; i++) {
      if(valid_ver == ver) {
         ss_sprintf(emsg, "");
         error = 0;
         break;
      }
   }

  end:
   return error;
}

/**
 * reintegrate_paths - reintegrate the given pool's paths
 * @name: name of pool
 * @emsg:
 *
 * Scope: external
 */
int reintegrate_paths(char *name, struct sstring *emsg)
{
   int error = 0;
   int fd;
   pool_status_t *pool = NULL;
   char path[256];

   cstring2ss(emsg, "", 0);

   if ((error = _get_pool_status(name, &pool, emsg))) {
       goto end;
   }
   if(pool == NULL) {
      error = 1;
      ss_sprintf(emsg, "Pool %s not assembled", name);
      goto end;
   }

   strcpy(path, "/dev/pool/");
   strcat(path, pool->ps_name);

   log_debug("Opening %s O_RDONLY", path);

   if((fd = open(path, O_RDONLY)) < 0) {
      ss_sprintf(emsg, "Error opening %s: %s", path, strerror(errno));
      error = errno;
      goto end;
   }

   if ((ioctl(fd, POOL_MP_RESTORE) < 0)) {
      ss_sprintf(emsg, "Pool path restore (ioctl) failed for %s: %s",
                 path, strerror(errno));
      error = errno;
      close(fd);
      goto end;
   }

   close(fd);

  end:
   return error;
}

/**
 * tune_multipath - adjusts the multipath type and stripe
 * @name: pool name
 * @mp: multipath value (-1: none, 0: failover, >0: RR)
 * @emsg:
 *
 * Scope: external
 */
int tune_multipath(char *name, int mp, struct sstring *emsg)
{
   int error = 0;
   int mpv = mp;
   int fd;
   pool_status_t *pool = NULL;
   char path[256];

   cstring2ss(emsg, "", 0);

   if ((error = _get_pool_status(name, &pool, emsg))) {
       goto end;
   }
   if(pool == NULL) {
      error = 1;
      ss_sprintf(emsg, "Pool %s not assembled", name);
      goto end;
   }

   strcpy(path, "/dev/pool/");
   strcat(path, pool->ps_name);

   log_debug("Opening %s O_RDONLY", path);

   if((fd = open(path, O_RDONLY)) < 0) {
      ss_sprintf(emsg, "Error opening %s: %s", path, strerror(errno));
      error = errno;
      goto end;
   }

   if (ioctl(fd, POOL_TUNE_MP, &mpv) < 0) {
      error = errno;
      switch (error) {
      case 0:
         ss_sprintf(emsg, "This should never be printed");
         break;
      case ENOENT:
         ss_sprintf(emsg, "License file not found in CCS.  Multipathing"
                    " disabled for %s.", name);
         break;
      case ETIMEDOUT:
         ss_sprintf(emsg, "Bad communication exchange with CCS. Verify CCS"
                    " is running.\n"
                    "Multipathing disabled for %s.", name);
         break;
      case EFAULT:
         ss_sprintf(emsg, "License file expired.  Multipathing disabled"
                    " for %s.", name);
         break;
      case EINVAL:
         ss_sprintf(emsg, "License file invalid.  Multipathing disabled"
                    " for %s.", name);
         break;
      case EBADMSG:
         ss_sprintf(emsg, "Error checking license file.  Multipathing"
                    " disabled for %s.", name);
         break;
      case EPERM:
         ss_sprintf(emsg, "License does not allow pool multipathing.\n"
                    "Multipathing set to 'none' for %s.", name);
         break;
      default:
         printf("%d\n", error);
         /* This is the EPARSE case - I'd rather not read in the ccs.h header
            just for this definition */
         ss_sprintf(emsg, "Unable to parse license file.  Multipathing"
                    " disabled for %s.", name);
         break;
      }

      close(fd);
      goto end;
   }

   close(fd);
  end:
   return error;
}

/**
 * assemble_pool - Assemble a named pool
 * @name: name of pool to assemble
 * @emsg:
 *
 * assembles a given pool and marks the pool cache dirty
 *
 * Returns: zero on success, non-zero on error
 *
 * Scope: external
 */
int assemble_pool(char *name, struct sstring *emsg)
{
   int error = 0;
   char poolname[POOL_NAME_SIZE];

   if(strlen(name) < POOL_NAME_SIZE) {
      strncpy(poolname, name, POOL_NAME_SIZE-1);
      poolname[POOL_NAME_SIZE-1] = '\0';
   }
   else {
      ss_sprintf(emsg, "Pool name must be less than %d characters.",
                 POOL_NAME_SIZE);
   }

   if(_commfd < 0){
      SETMSGSS(emsg, "comm device not opened");
      error = EBADF;
      goto end;
   }
   /* Check for file in /dev/pool that has the name of the pool - if it exists
      and is a block device, ax it - otherwise error out. */
   if((error = _check_block_dev(poolname, emsg)) != 0) {
      goto end;
   }

   /* FIXME: should i error out here? */
   if(clean_pool_dir(emsg) != 0) {
      char str[128];
      int len;
      ss_len(emsg, &len);
      if(len > 0) {
         ss2cstring(str, 128, emsg);
         log_warn(str);
         cstring2ss(emsg, "", 0);
      }
   }

   if ((ioctl(_commfd, POOL_ASSEMBLE, (void *)poolname)) < 0) {
      error = errno;
      switch(error) {
      case ENOSPC:
         ss_sprintf(emsg, "Assembling %s failed.\n"
                    "Verify that two or more of your pool labels are not on"
                    " the same device\n"
                    "and that all devices are present.", poolname);
         break;
      case ENODEV:
         ss_sprintf(emsg, "Assembling %s failed.\n"
                    "Verify that your pool labels are valid", poolname);
         break;
      case EMFILE:
         ss_sprintf(emsg, "Assembling %s failed.\n"
                    "You can only have 64 static and 63 dynamic numbered pools",
                    poolname);
         break;
      case EEXIST:
         ss_sprintf(emsg, "Assembling %s failed.\n"
                    "Fixed minor number conflicts with already assembled pool",
                    poolname);
         break;
      default:
         ss_sprintf(emsg, "Assembling %s (ioctl) failed: %s",
                    poolname, strerror(errno));
      }

      goto end;
   }
   /* Mark the cache dirty so the new pool gets read in later */
   _pl_cache_dirty = 1;

   /* Create a device node for the newly assembled pool */
   /* _make_block_dev needs the cache updated */
   if((error = _get_pool_list(emsg)) != 0) {
      goto end;
   }

   /* FIXME: should i disassemble the pool if i can't create the device node? */
   if((error = _make_block_dev(poolname, emsg)) != 0) {
      goto end;
   }

end:
   return error;
}

/**
 * grow_pool - Grow an existing pool
 * @name: name of pool to be grown
 * @emsg:
 */
int grow_pool(char *name, struct sstring *emsg)
{
   int error = 0;
   char poolname[POOL_NAME_SIZE];

   if(strlen(name) < POOL_NAME_SIZE) {
      strncpy(poolname, name, POOL_NAME_SIZE-1);
      poolname[POOL_NAME_SIZE-1] = '\0';
   }
   else {
      ss_sprintf(emsg, "Pool name must be less than %d characters.",
                 POOL_NAME_SIZE);
   }

   if(_commfd < 0) {
      ss_sprintf(emsg, "comm device not opened");
      error = EBADF;
   }
   else if ((ioctl(_commfd, POOL_GROW, (void *)poolname)) < 0) {
      error = errno;
      if(error == ENOSPC) {
         ss_sprintf(emsg, "Growing %s failed.\n"
                    "Verify that two or more of your pool labels are not on"
                    " the same device", poolname);
      }
      else if (error == ENODEV) {
         ss_sprintf(emsg, "Growing %s failed.\n"
                    "Verify that your pool labels are valid", poolname);
      }
      else {
         ss_sprintf(emsg, "Growing %s (ioctl) failed: %s",
                    poolname, strerror(errno));
      }
   }

   _pl_cache_dirty = 1;
   return error;
}

/**
 * disassemble_pool - Disassemble a named pool
 * @name: name of pool to disassemble
 *
 * Attempts to disassemble pool and marks pool cache dirty if successful
 *
 * Returns: zero on success, non-zero on error
 *
 * Scope: external
 */
int disassemble_pool(char *name, struct sstring *emsg)
{
   int error = 0;
   pool_status_t *pool = NULL;

   if(_commfd < 0){
      SETMSGSS(emsg, "comm device not opened");
      error = EBADF;
      goto end;
   }

   if((error = _get_pool_status(name, &pool, emsg))) {
      goto end;
   }
   if(pool == NULL) {
      error = 1;
      ss_sprintf(emsg, "Pool %s not assembled", name);
      goto end;
   }
   /* Actually unload the pool from the kernel */
   if(ioctl(_commfd, POOL_REMOVE, &(pool->ps_pool_minor))) {
      ss_sprintf(emsg, "Removing %s (ioctl) minor: %u failed: %s", name,
            pool->ps_pool_minor, strerror(errno));
      error = errno;
      goto end;
   }
   /* as soon as the remove is successful, the cache is dirty */
   _pl_cache_dirty = 1;
   _get_full_pool(emsg);

   /* FIXME: should i error out here? */
   if(clean_pool_dir(emsg) != 0) {
      char str[128];
      int len;
      ss_len(emsg, &len);
      if(len > 0) {
         ss2cstring(str, 128, emsg);
         log_warn(str);
         cstring2ss(emsg, "", 0);
      }
   }

  end:
   return error;
}

/**
 * clear_pool_stats - clear the pool statistics counter in the pool module
 * @name: name of pool to clear
 *
 * Clears the pool statistics counter in the pool module.
 *
 * Returns: zero on success, non-zero on error
 *
 * Scope: external
 *
 */
int clear_pool_stats(char *name, struct sstring *emsg)
{
   int error = 0;
   int fd = -1;
   pool_status_t *pool = NULL;
   char path[256];

   if(_commfd < 0){
      SETMSGSS(emsg, "comm device not opened");
      error = EBADF;
      goto end;
   }

   if ((error = _get_pool_status(name, &pool, emsg)) != 0) {
      goto end;
   }
   if(pool == NULL) {
      error = 1;
      ss_sprintf(emsg, "Pool %s not assembled", name);
      goto end;
   }
   strcpy(path, "/dev/pool/");
   strcat(path, pool->ps_name);

   log_debug("Opening %s O_RDONLY", path);
   fd = open(path, O_RDONLY);
   if(fd == -1) {
      ss_sprintf(emsg, "Unable to open %s for reading", path);
      error = errno;
      goto end;
   }

   if (ioctl(fd, POOL_CLEAR_STATS, NULL) < 0) {
      ss_sprintf(emsg, "Clearing pool statistics for %s (ioctl) failed: %s",
                 name, strerror(errno));
      error = errno;
      goto end;
   }

  end:
   return error;
}

/**
 * get_pool_stats - get current pool statistics
 * @name: name of pool
 * @stat: statistics structure to fill
 *
 * Returns: zero on success, non-zero on error
 *
 * Scope: external
 *
 */
int get_pool_stats(char *name, user_poolstat_t **stat, struct sstring *emsg)
{
   int error = 1;
   pool_stat_t		iostat;
   int fd = -1;
   pool_status_t *pool = NULL;
   char path[256];

   if ((error = _get_pool_status(name, &pool, emsg))) {
       goto end;
   }
   if(pool == NULL) {
      error = 1;
      ss_sprintf(emsg, "Pool %s not assembled", name);
      goto end;
   }

   strcpy(path, "/dev/pool/");
   strcat(path, pool->ps_name);

   log_debug("Opening %s O_RDONLY", path);

   if((fd = open(path, O_RDONLY)) < 0) {
      ss_sprintf(emsg, "Error opening %s: %s", path, strerror(errno));
      error = errno;
      goto end;
   }

   if (ioctl(fd, POOL_DUMP_TOTALSTATS, &iostat) < 0) {
      ss_sprintf(emsg, "dumping pool statistics for %s (ioctl) failed: %s.",
                 name, strerror(errno));
      error = errno;
      close(fd);
      goto end;
   }

   close(fd);

   if(error == 0) {
      *stat = (user_poolstat_t *) malloc(sizeof(user_poolstat_t));
      (*stat)->reads = iostat.s_reads;
      (*stat)->readblks = iostat.s_readblks;
      (*stat)->writes = iostat.s_writes;
      (*stat)->writeblks = iostat.s_writeblks;
   }

  end:
   return error;
}

/**
 * destroy_pool_stats - cleans up the pool stat structure
 * @stat: The statistics structure to destroy
 *
 * Frees the user_poolstat_t structure
 *
 * Scope: external
 */
void destroy_pool_stats(user_poolstat_t **stat)
{
   if(*stat)
      free(*stat);
}

/**
 * update_device_nodes - create and remove nodes based on cache
 *
 * Creates new device nodes based on assembled pools
 *
 * Returns: zero on success, non-zero on failure
 *
 * Scope: external
 */
int update_device_nodes(struct sstring *emsg)
{
   int error = 0;

   if ((error = _get_pool_list(emsg)) != 0) {
      goto end;
   }
   /* I should really have an accessor for the cache that checks the dirty flag
    * and pulls in the required information if necessary... */
   /*else if ((error = _clean_pool_dir()) != 0) {
      log_debug("Error cleaning pool nodes...\n");
   }*/
   if ((error = _make_pool_nodes(emsg)) != 0) {
      goto end;
   }

  end:
   return error;
}

/**
 * get_loaded_pools - fills diskinfo structure with currently assembled pools
 * @di: diskinfo structure to fill
 *
 * fills di.plcount and di.pl. ignores di.dl* since disklabels are irrelevant
 * for this.
 *
 * Returns: zero on sucess, non-zero on error
 *
 * Scope: external
 */
int get_loaded_pools(diskinfo_t *di, struct sstring *emsg)
{
   int i, j, k, l;
   int error = 0;
   user_pool_t *p = NULL;
   user_subpool_t *sp = NULL;
   user_device_t *dev = NULL;
   char *path = NULL;
   pool_status_t *pl_l = NULL;
   pool_spstatus_t *sp_l = NULL;
   pool_devstatus_t *dev_l = NULL;

   if ((error = _get_full_pool(emsg)) != 0) {
      goto end;
   }

   di->dl_count = 0;
   di->pl_count = _pl_cache->pl_count;

   di->pl = (user_pool_t *) malloc(sizeof(user_pool_t) * di->pl_count);
   if(di->pl == NULL) {
      ss_sprintf(emsg, "%s(%d): %s", __FILE__, __LINE__, strerror(errno));
      error = errno;
      goto end;
   }
   memset(di->pl, 0, sizeof(user_pool_t) * di->pl_count);

   /* Should i use the user_* structures for the cache?? */
   for(i = 0; i < di->pl_count; i++) {
      p = &(di->pl[i]);
      pl_l = &(_pl_cache->pl_list[i]);

      strcpy(p->name, pl_l->ps_name);
      p->id = pl_l->ps_pool_id;
      p->major = pl_l->ps_pool_major;
      p->minor = pl_l->ps_pool_minor;
      p->num_sps = pl_l->ps_num_spools;
      p->in_use = pl_l->ps_in_use;
      p->mptype = pl_l->ps_mptype;
      p->mpstripe = pl_l->ps_mpstripe;

      p->sp = (user_subpool_t *) malloc(sizeof(user_subpool_t) * p->num_sps);
      if(p->sp == NULL) {
         ss_sprintf(emsg, "%s(%d): %s", __FILE__, __LINE__, strerror(errno));
         error = errno;
         free(di->pl);
         di->pl_count = 0;
         goto end;
      }
      memset(p->sp, 0, sizeof(user_subpool_t) * p->num_sps);

      for(j = 0; j < p->num_sps; j++) {
         sp = &(p->sp[j]);
         sp_l = &(_pl_cache->pl_list[i].ps_splist[j]);

         sp->id = j;
         sp->striping = sp_l->sps_stripe;
         sp->num_devs = sp_l->sps_numdevs;
         sp->type = sp_l->sps_type;

         sp->devs = (user_device_t *)
            malloc(sizeof(user_device_t) * sp->num_devs);
         if(sp->devs == NULL) {
            ss_sprintf(emsg, "%s(%d): %s", __FILE__, __LINE__,
                       strerror(errno));
            error = errno;
            free(p->sp);
            free(di->pl);
            di->pl_count = 0;
            goto end;
         }
         memset(sp->devs, 0, sizeof(user_device_t) * sp->num_devs);

         for(k = 0; k < sp->num_devs; k++) {
            dev = &(sp->devs[k]);
            dev_l = &(_pl_cache->pl_list[i].ps_splist[j].sps_devlist[k]);
            dev->initialized = 1;
            dev->sp_id = j;
            dev->devid = k;
            dev->blocks = dev_l->pds_blocks;
            dev->path_count = dev_l->pds_num_paths;
            dev->good_path_count = dev_l->pds_num_good_paths;
            /* FIXME: do i really want to ignore errors here? */
            for(l = 0; l < dev->path_count; l++) {
               _get_devpath(&path, (dev_l->pds_dev[l]), emsg);
               (dev->dev_path)[l] = path;
               path = NULL;
            }
         }
      }
   }

end:
   return error;
}

/**
 * get_pool_alias - Figures out the pool name as /proc/partitions sees it
 * @name: pointer to hold pool alias name
 * @minor: pool minor number
 *
 * Allocates a character pointer for the pool alias name and assigns it to
 * *pool, Calculates the pool alias name based on the minor number, and fills
 * in *pool with it.
 *
 * Returns: zero on success, non-zero on failure
 *
 * Scope: external
 */
int get_pool_alias(char **name, uint32 minor, struct sstring *emsg)
{
   int error = 0;
   char *n = NULL;
   n = (char *) malloc(sizeof(char) * 20);
   if(n != NULL) {
      memset(n, 0, 20);
      _calc_link_name(n, minor);
      *name = n;
   }
   else {
      error = errno;
      SETMSGSS(emsg, strerror(error));
   }

   return error;
}

/**
 * destroy_pool_alias - frees the memory in pool
 * @pool - pool alias to destroy
 *
 * Scope: external
 *
 */
void destroy_pool_alias(char *pool)
{
   if(pool) {
      free(pool);
      pool = NULL;
   }
}

/*
 * vim: et ts=3 sw=3 noai cin :
 */

