/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

/* pool_tool.c - Manipulates pool labels */

/*
 * FIXME: add the ability to print the number of blocks lost to striping and
 *        subpool boundaries.
 */

#include <libgen.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <ctype.h>

#include "sisstring.h"
#include "sislogger.h"
#include "siscli.h"
#include "libpool.h"
#include "osi_user.h"
#include "pool_release.cf"
#include "copyright.cf"

#define STRING_SIZE 512
#define PRE_TAG "<- "
#define POST_TAG " ->"

static int _print_configs(char *poolname, struct sstring *emsg);
static int _create_pool(char *conffile, struct sstring *emsg);
static int _grow_pool(char *conffile, struct sstring *emsg);
static int _scan_and_print_labels(struct sstring *emsg);
static int _erase_pool(char *poolname, struct sstring *emsg);
static int _rename_pool(char *oldname, char *newname, struct sstring *emsg);
static int _reminor_pool(char *oldname, int minor, struct sstring *emsg);

void _usage(char *name) {

   printf("Usage:\n");
   printf("\n");
   printf("%s [options] [config file(s)/pool(s)]\n", name);
   printf("\n");
   printf("Options:\n");
   printf("  -c               Create a new pool.\n");
   printf("  -D               Enable debugging output.\n");
   printf("  -e               Erase pool.\n");
   printf("  -g               Grow pool.\n");
   printf("  -h               This usage information.\n");
   printf("  -m <num>         Change minor number on specified pool"
         " to <num>.\n");
   printf("  -O               Override prompts\n");
   printf("  -p               Print config file(s) based on disk labels.\n");
   printf("  -q               Enable quiet output.\n");
   printf("  -r <name>        Rename pool to <name>.\n");
   printf("  -s               Scan for and show all disk labels.\n");
   printf("  -V               Version information.\n");
   printf("  -v               Enable verbose output.\n");

   return;
}

void _grow_usage(char *name) {
   printf("Usage:\n");
   printf("\n");
   printf("%s [options] [config file(s)]\n", name);
   printf("\n");
   printf("Options:\n");
   printf("  -D               Enable debugging output.\n");
   printf("  -h               This usage information.\n");
   printf("  -O               Override prompts.\n");
   printf("  -q               Quiet operation.\n");
   printf("  -V               Version information.\n");
   printf("  -v               Enable verbose output.\n");

   return;
}

static void _grow_exec(int argc, char **argv)
{
   int error = 0;
   int c;
   int quiet = 0;
   int verbose = 0;
   int i = 0;
   int len = 0;
   char **arg_list = NULL;
   char msg[STRING_SIZE];
   struct sstring *emsg = NULL;

   while((c = getopt(argc, argv, "DhOqVv")) != -1 && !error) {
      switch(c) {

      case 'D':
         verbose = 4;
         break;

      case 'h':
         _grow_usage(argv[0]);
         goto end;
         break;

      case 'O':
         break;

      case 'q':
         if(verbose > 0) {
            mutually_exclusive(argv[0], c, 'v');
            error = 1;
         }
         else {
            quiet++;
         }
         break;

      case 'V':
         printf("%s %s (built %s %s)\n", argv[0],
                POOL_RELEASE_NAME, __DATE__, __TIME__);
         printf("%s\n", GFS_COPYRIGHT);
         goto end;
         break;

      case 'v':
         if(quiet > 0) {
            mutually_exclusive(argv[0], c, 'q');
            error = 1;
         }
         else {
            verbose++;
         }
         break;
      }
   }

   if(argc > optind) {
      arg_list = &(argv[optind]);
      argc = argc - optind;
   }

   if(!error) {
      if((error = ccs_initlog("pool.ccs", "pool/tool_config",
                              verbose, quiet))) {
         if((error = init_defaultlog(verbose, quiet, NULL)) != 0) {
            goto end;
         }
      }

      emsg = ss_init(STRING_SIZE-1);

      if((error = open_comm(emsg))) {
         if(error != ENXIO) {
            ss_len(emsg, &len);
            if(len > 0) {
               ss2cstring(msg, STRING_SIZE, emsg);
               log_err(msg);
            }
            goto end;
         }
         error = 0;
         cstring2ss(emsg, "", 0);
      }

      if ((error = check_version(emsg))) {
         goto end;
      }

      /* Clear out block devices in the /dev/pool directory */
      clean_pool_dir(emsg);

      if(arg_list != NULL) {
         for(i = 0; i < argc; i++) {
            if((error |= _grow_pool(arg_list[i], emsg)) == 0) {
               log_warn("Pool grown successfully from %s",
                        arg_list[i]);
            }
            else {
               ss_len(emsg, &len);
               if(len > 0) {
                  ss2cstring(msg, STRING_SIZE, emsg);
                  log_err(msg);
               }
            }
         }
      }
      else {
         log_err("Please specifiy a config file to use for growing");
         error = 1;
      }


   }

  end:
   close_comm(emsg);
   /* Until we have well defined error codes, the exit status will either
      be 0 or 1 */
   if(error != 0) {
      error = 1;
   }
   exit(error);
}

int main(int argc, char **argv)
{
   int    error = 0;
   int    c;
   int    quiet = 0;
   int    verbose = 0;
   int    i = 0;
   int    len = 0;
   uint32 minor = 0;
   char   action = 0,
        **arg_list = NULL,
         *newpool = NULL;
   char  *cmdname,
         *base = NULL;
   char   msg[STRING_SIZE];
   struct sstring *emsg = NULL;

   cmdname = strdup(argv[0]);
   base = basename(cmdname);

   if(strcmp(base, "pool_grow") == 0) {
      free(cmdname);
      _grow_exec(argc, argv);
   }
   free(cmdname);
   while((c = getopt(argc, argv, "cDeghm:Opqr:sVv")) != -1 && !error) {
      switch(c) {

         case 'c':
            if(action != 0) {
               mutually_exclusive(argv[0], c, action);
               error = 1;
            }
            else
               action = c;
            break;

         case 'D':
            verbose = 4;
            break;

         case 'e':
            if(action != 0) {
               mutually_exclusive(argv[0], c, action);
               error = 1;
            }
            else
               action = c;
            break;

         case 'g':
            if(action != 0) {
               mutually_exclusive(argv[0], c, action);
               error = 1;
            }
            else
               action = c;
            break;

         case 'h':
            _usage(argv[0]);
            return error;
            break;

         case 'm':
            if(action != 0) {
               mutually_exclusive(argv[0], c, action);
               error = 1;
            }
            else {
               action = c;
               minor = atoi(optarg);
            }
            break;

         case 'O':
            add_flag(POOL_FORCE);
            break;

         case 'p':
            if(action != 0) {
               mutually_exclusive(argv[0], c, action);
               error = 1;
            }
            else
               action = c;
            break;

         case 'q':
            if(verbose > 0) {
               mutually_exclusive(argv[0], c, 'v');
               error = 1;
            }
            else {
               quiet++;
            }
            break;

         case 'r':
            if(action != 0) {
               mutually_exclusive(argv[0], c, action);
               error = 1;
            }
            else {
               action = c;
               newpool = optarg;
            }
            break;

         case 's':
            if(action != 0) {
               mutually_exclusive(argv[0], c, action);
               error = 1;
            }
            else
               action = c;
            break;

         case 'V':
            printf("%s %s (built %s %s)\n", argv[0],
                  POOL_RELEASE_NAME, __DATE__, __TIME__);
            printf("%s\n", GFS_COPYRIGHT);
            return error;
            break;

         case 'v':
            if(quiet > 0) {
               mutually_exclusive(argv[0], c, 'q');
               error = 1;
            }
            else {
               verbose++;
               break;
            }

         case ':':
         case '?':
            fprintf(stderr, "Please use '-h' for usage.\n");
            exit(1);
            break;

         default:
            fprintf(stderr, "Bad programmer! You forgot to catch the"
                 " %c flag\n", c);
            exit(1);
            break;
      }

   }

   if(argc > optind) {
      arg_list = &(argv[optind]);
      argc = argc - optind;
   }

   if(!error) {
      if((error = ccs_initlog("pool.ccs", "pool/tool_config",
                              verbose, quiet))) {
         if((error = init_defaultlog(verbose, quiet, NULL)) != 0) {
            goto end;
         }
      }

      emsg = ss_init(STRING_SIZE-1);

      if((error = open_comm(emsg))) {
         /* ENXIO occurs when pool module isn't loaded */
         if(error != ENXIO && error != ENOENT) {
            ss_len(emsg, &len);
            if(len > 0) {
               ss2cstring(msg, STRING_SIZE, emsg);
               log_err(msg);
            }
            goto end;
         }
         error = 0;
         cstring2ss(emsg, "", 0);
      }
      else if ((error = check_version(emsg))) {
         ss_len(emsg, &len);
         if(len > 0) {
            ss2cstring(msg, STRING_SIZE, emsg);
            log_err(msg);
         }
         goto end;
      }

      /* Clear out block devices in the /dev/pool directory */
      clean_pool_dir(emsg);

      switch(action) {

            /* Create */
            case 0:
            case 'c':
               if(arg_list != NULL) {
                  for(i = 0; i < argc; i++) {
                     if((error |= _create_pool(arg_list[i], emsg)) == 0) {
                        log_warn("Pool label written successfully from %s",
                                 arg_list[i]);
                     }
                     else {
                        ss_len(emsg, &len);
                        if(len > 0) {
                           ss2cstring(msg, STRING_SIZE, emsg);
                           log_err(msg);
                        }
                     }
                  }
               }
               else {
                  log_err("Please specifiy a config file");
                  log_err("Please use -h for usage");
                  error = 1;
               }

               break;

            /* Erase */
            case 'e':
               if(arg_list != NULL) {
                  for(i = 0; i < argc; i++) {
                     if((error |= _erase_pool(arg_list[i], emsg)) != 0) {
                        ss_len(emsg, &len);
                        if(len > 0) {
                           ss2cstring(msg, STRING_SIZE, emsg);
                           log_err(msg);
                        }
                     }
                  }
               }
               else {
                  if((error = _erase_pool(NULL, emsg)) != 0) {
                     ss_len(emsg, &len);
                     if(len > 0) {
                        ss2cstring(msg, STRING_SIZE, emsg);
                        log_err(msg);
                     }
                  }
               }
               break;

            /* Grow */
            case 'g':
               if(arg_list != NULL) {
                  for(i = 0; i < argc; i++) {
                     if((error |= _grow_pool(arg_list[i], emsg)) == 0) {
                        log_warn("Pool grown successfully from %s",
                                 arg_list[i]);
                     }
                     else {
                        ss_len(emsg, &len);
                        if(len > 0) {
                           ss2cstring(msg, STRING_SIZE, emsg);
                           log_err(msg);
                        }
                     }
                  }
               }
               else {
                  log_err("Please specifiy a config file to use for growing");
                  error = 1;
               }

               break;

            /* Reminor */
            case 'm':
               if(arg_list != NULL) {
                  if(argc > 1) {
                     log_warn("Can only change the minor number for one "
                              "pool at a time");
                  }
                  if((error = _reminor_pool(arg_list[0], minor, emsg)) != 0) {
                     ss_len(emsg, &len);
                     if(len > 0) {
                        ss2cstring(msg, STRING_SIZE, emsg);
                        log_err(msg);
                     }
                  }
               }
               else {
                  log_err("No pool specified");
                  log_err("Please use '-h' for usage");
                  error = 1;
                  break;
               }

               break;

            /* Print config */
            case 'p':
               if(arg_list != NULL) {
                  for(i = 0; i < argc; i++) {
                     if((error |= _print_configs(arg_list[i], emsg)) != 0) {
                        ss_len(emsg, &len);
                        if(len > 0) {
                           ss2cstring(msg, STRING_SIZE, emsg);
                           log_err(msg);
                        }
                     }
                  }
               }
               else {
                  if((error = _print_configs(NULL, emsg)) != 0) {
                     ss_len(emsg, &len);
                     if(len > 0) {
                        ss2cstring(msg, STRING_SIZE, emsg);
                        log_err(msg);
                     }
                  }
               }
               break;

            /* Rename */
            case 'r':
               if(arg_list != NULL) {
                  if(argc > 1) {
                     log_warn("Can only change the name of one "
                              "pool at a time");
                  }
                  if((error = _rename_pool(arg_list[0], newpool, emsg)) != 0) {
                     ss_len(emsg, &len);
                     if(len > 0) {
                        ss2cstring(msg, STRING_SIZE, emsg);
                        log_err(msg);
                     }
                  }
               }
               else {
                  log_err("No pool specified");
                  log_err("Please use '-h' for usage");
                  error = 1;
                  break;
               }

               break;

            /* Scan disks for labels */
            case 's':
               if((error = _scan_and_print_labels(emsg)) != 0) {
                  ss_len(emsg, &len);
                  if(len > 0) {
                     ss2cstring(msg, STRING_SIZE, emsg);
                     log_err(msg);
                  }
               }
               break;

            default:
               fprintf(stderr, "Please use -h' for usage.\n");
               error = 1;
               break;
         }
      }

  end:
   close_comm(emsg);
   /* Until we have well defined error codes, the exit status will either
      be 0 or 1 */
   if(error != 0) {
      error = 1;
   }
   return error;

}

/* Gets a single character from the input stream and flushes the rest */
static char _get_a_char(FILE *fp) {
   int c, d;

   d = c = fgetc(fp);
   while (d != '\n' && d != EOF)
      d = fgetc (fp);

   return (char)c;
}

/*
 * _find_pool - finds a pool in a disk_info structure
 * @di: diskinfo structure to search
 * @name: name of pool to search for
 *
 * Returns: zero based index in di.pl array of found pool, or negative if no
 * pool is found.
 *
 */
static int _find_pool(diskinfo_t *di, char *name)
{
   int i;

   for(i = 0; i < di->pl_count; i++) {
         if(strcmp(di->pl[i].name, name) == 0) {
            break;
         }
      }
   if(i < di->pl_count)
      return i;
   return -1;
}

/* _write_config_to_disk - writes the given pool out to disk
 * @pl: pool structure to write to disk
 *
 * Returns: zero on success, non-zero on error
 */
static int _write_config_to_disk(user_pool_t *pl, struct sstring *emsg)
{
   int         error = 0;

   log_info("Writing labels to disk...");
   error = write_labels(pl, emsg);
   if(!error)
      log_info("ok.");
   else {
      log_info("error!");
   }
   return error;
}

/* _read_config -
 * @conffile:
 * @pl:
 */
static int _read_config(char *conffile, user_pool_t **pl, struct sstring *emsg)
{
   int error = 0;
   struct stat stat_buf;

   /* Read in the pool config */
   if(conffile) {
      if(!strcmp(conffile, "-"))
         conffile = NULL;
      else
         if(stat(conffile, &stat_buf)) {
            log_err("Config file \"%s\" not found.", conffile);
            error = -errno;
            goto end;
         }
   }

   log_info("Reading config file");
   if((error = read_pool_config(conffile, pl)) != 0) {
      log_info("failed.");
   }
   else {
      log_info("ok.");
   }
end:
   return error;
}

/**
 * _cmpdevs - compare two device paths to see if they are equal
 * @path1:
 * @path2:
 *
 * Returns: 0 on match, non-zero on no match
 */
static int _cmpdevs(char *path1, char *path2)
{
   struct stat s1, s2;

   log_debug("Comparing %s with %s", path1, path2);
   if(!strcmp(path1, path2)) {
      return 0;
   }

   /* if either stat fails we have no match (and we probably won't be able
      to do anything with that device anyway */
   if(stat(path1, &s1))
      return 1;
   if(stat(path2, &s2))
      return 1;

   if(!S_ISBLK(s1.st_mode) || !S_ISBLK(s2.st_mode)) {
      return 1;
   }

   log_debug("%s: %u, %u - %s: %u, %u", path1, major(s1.st_rdev), minor(s1.st_rdev), path2, major(s2.st_rdev), minor(s2.st_rdev));

   if((major(s1.st_rdev) == major(s2.st_rdev)) &&
      (minor(s1.st_rdev) == minor(s2.st_rdev)))
      return 0;

   return 1;

}


static int dev_grow_cmp(user_device_t *dev, user_device_t *ddev)
{
   int error = 3;
   int i;

   log_debug("%d", dev->path_count);
   for(i = 0; i < dev->path_count; i++) {
      log_debug("%s %s", dev->dev_path[i], ddev->dev_path[0]);

      if(dev->sp_id == ddev->sp_id &&
         dev->devid == ddev->devid &&
         /*dev->blocks == ddev->blocks &&*/
         /* Check all visible paths */
         !_cmpdevs(dev->dev_path[i], ddev->dev_path[0])) {
         error = 0;
         break;
      }
   }

   return error;
}

/* subpool_grow_cmp - compare two sp structures to verify growing correctness
 * @sp: new subpool
 * @dsp: ondisk subpool to check against
 * @d: device that had error (only valid on error)
 *
 */
static int subpool_grow_cmp(user_subpool_t *sp, user_subpool_t *dsp, int *d)
{
   int error = 2,
      i = 0;
   if(sp->id == dsp->id &&
         sp->striping == dsp->striping &&
         sp->type == dsp->type &&
         sp->num_devs == dsp->num_devs)
      for(i = error = 0; !error && i < dsp->num_devs; i++) {
            log_debug("%d", sp->id);

         error = dev_grow_cmp(&(sp->devs[i]), &(dsp->devs[i]));
         *d = i;
      }


   return error;
}

/* pool_grow_cmp - compare two pool structures to verify growing correctness
 * @pl: new pool
 * @dpl: ondisk pool to check against
 * @sp: subpool that had error (only valid on error)
 * @d: device that had error (only valid on error)
 */
static int pool_grow_cmp(user_pool_t *pl, user_pool_t *dpl, int *sp, int *d)
{
   int error = 1,
      i = 0;

   if(pl->major == dpl->major &&
      pl->minor == dpl->minor) {

      if(pl->num_sps > dpl->num_sps) {

         for(i = error = 0; !error && i < dpl->num_sps ; i++) {

            error = subpool_grow_cmp(&(pl->sp[i]), &(dpl->sp[i]), d);
            *sp = i;
         }
      }
   }


   return error;
}

/* _verify_pool_to_grow -
 * @pl: new pool
 * @di: ondisk pools to check against
 */
int _verify_pool_to_grow(user_pool_t *pl, diskinfo_t *di, struct sstring *emsg)
{
   int error = 0,
       i;
   int sp = 0,
       dev = 0;
   user_pool_t *dpl = NULL;

   /* look for the pool in the diskinfo structure */
   for(i = 0; i < di->pl_count; i++)
      if(!strcmp(pl->name, di->pl[i].name)) {
         dpl = &(di->pl[i]);
         break;
      }

   if(dpl != NULL) {
      if((error = pool_grow_cmp(pl, dpl, &sp, &dev)) != 0) {
         switch(error) {
         case 1:
            ss_sprintf(emsg, "The number of subpools in the new pool should"
                       " be greater than the number in the old");
            break;
         case 2:
            ss_sprintf(emsg, "When growing, you can add subpools, but you"
                       " can't change them\nSubpool %d changed in new config",
                       sp);
            break;
         case 3:
            ss_sprintf(emsg, "Device %d in subpool %d doesn't match up with"
                       " old config", dev, sp);
            break;
         default:
            ss_sprintf(emsg, "Unknown error");
         goto end;
         }
      }
   }
   else {
      ss_sprintf(emsg, "No matching configuration found on disk");
      error = 2;
      goto end;
   }

   pl->id = dpl->id;

  end:
   return error;
}


/* _print_configs - read labels and convert them to configs
 * @poolname - name of pool to print the config for

*/
static int _print_configs(char *poolname, struct sstring *emsg)
{
   int error = 0,
       i;
   diskinfo_t di = {0};

   log_info("Scanning disks for pools...");
   if((error = scan_disks(&di, emsg)) == 0) {
      if(poolname) {
         if((i = _find_pool(&di, poolname)) < 0) {
            log_info("failed.");
            log_warn("No pool named %s found.", poolname);
            error = 1;
         }
         else {
            log_info("ok.");
            print_config(&di.pl[i]);
            print_pool_labels(di.pl[i].name, emsg);
         }
      }
      else {
         if(di.pl_count <= 0) {
            log_info("failed.");
            log_warn("No pools found on disk");
            error = 1;
         }
         else if (di.pl_count == 1) {
            log_info("ok.");
            print_config(&di.pl[0]);
            print_pool_labels(di.pl[0].name, emsg);
         }
         else {
            log_info("ok.");
            log_warn("#Break the following into separate files.");
            for(i = 0; i < di.pl_count; i++) {
               log_warn("#--- begin %s --------------------", di.pl[i].name);
               print_config(&di.pl[i]);
               log_warn("#--- end %s ----------------------", di.pl[i].name);
               print_pool_labels(di.pl[i].name, emsg);
            }
         }
      }
   }
   else {
      log_info("Error.");
   }
   return error;
}


/* _check_labels - check disks for other types of labels
 * @di:
 * @pl:
 *
 * Returns: zero on success, non-zero on error
 */
static int _check_labels(diskinfo_t *di, diskinfo_t *cdi,
                         user_pool_t *pl, struct sstring *emsg)
{
   int  i, j, k, l,
      error = 0,
      found = 0;
   int set_force = 0;
   char *ptype = NULL;
   char *last_erased = NULL;
   char *pool_path = NULL;

   for(i = 0; i < pl->num_sps; i++)
      for(j = 0; j < pl->sp[i].num_devs; j++) {
         found = 0;
         for(k = 0; k < di->dl_count && !found; k++) {
            /* watch for override */
            for(l = 0; l < pl->sp[i].devs[j].path_count; l++) {
               pool_path = pl->sp[i].devs[j].dev_path[l];
               if(!_cmpdevs(pool_path, di->dl[k].dev_path)) {
                  found = 1;
                  switch(di->dl[k].lt) {

                  case error_label:
                     log_err("Error processing label for %s.  Cannot use "
                             "for pool.", pool_path);
                     error = 1;
                     goto end;

                  case known_label:

                     if(check_flag(POOL_FORCE) == 0) {
                        printf("%s has \"%s\" on it.\nAre you sure you want "
                               "to overwrite \"%s\" with a pool label on %s? "
                               "(y/n)",
                               pool_path, di->dl[k].name,
                               di->dl[k].name, pool_path);
                        if(tolower(_get_a_char(stdin)) != 'y') {
                           ss_sprintf(emsg, "User requested %s with \"%s\" "
                                      "not be overwritten.",
                                      pool_path, di->dl[k].name);
                           error = 1;
                           goto end;
                        }
                     }
                     break;

                  case pool_label:
                     if(strcmp(di->dl[k].name, pl->name) == 0) {
                        break;
                     }
                     if(last_erased != NULL) {
                        if(strcmp(di->dl[k].name, last_erased) == 0) {
                           break;
                        }
                     }

                     if(cdi && _find_pool(cdi, di->dl[k].name) >= 0) {
                        log_err("%s has an assembled pool \"%s\"\n"
                                "Disassemble and erase it"
                                " if you want to write a new pool.",
                                pool_path, di->dl[k].name);
                        error = 1;
                        goto end;
                     }
                     if(check_flag(POOL_FORCE) == 0) {
                        /* Watch for pool labels... */
                        printf("%s has pool label %s on it.\n",
                               pool_path, di->dl[k].name);
                        printf("Do you want to erase pool %s (y/n)",
                               di->dl[k].name);
                        if(tolower(_get_a_char(stdin)) != 'y') {
                           ss_sprintf(emsg, "User requested %s with pool %s"
                                      " label not be overwritten",
                                      pool_path, di->dl[k].name);
                           error = 1;
                           goto end;
                        }
                        set_force = 1;
                        add_flag(POOL_FORCE);
                     }

                     if((error = _erase_pool(di->dl[k].name, emsg)) != 0) {
                        if(set_force == 1) {
                           del_flag(POOL_FORCE);
                        }
                        goto end;
                     }

                     if(set_force == 1) {
                        del_flag(POOL_FORCE);
                     }

                     if(last_erased)
                        free(last_erased);
                     last_erased = strdup(di->dl[k].name);

                     break;

                  case unknown_label:
                     break;

                  default:
                     /* Shouldn't get here */
                     log_debug("Missed a label type in the switch statement or"
                               " invalid label type in label for %s: %d",
                               pool_path, di->dl[k].lt);
                     error = 1;
                     break;
                  }
               }
            }
         }
      }


end:
   if(ptype != NULL)
      free(ptype);
   if(last_erased)
      free(last_erased);

   return error;
}

/* _create_pool -
 * @conffile:
 *
 */
static int _create_pool(char *conffile, struct sstring *emsg)
{
   int error = 0,
       set_force = 0;
   user_pool_t *pl = NULL;
   diskinfo_t core_di = {0};
   diskinfo_t disk_di = {0};

   if((error = _read_config(conffile, &pl, emsg))) {
      goto end;
   }

   /* Verify pool isn't loaded */
   if((error = get_loaded_pools(&core_di, emsg)) != 0) {
      /* don't error out if the pool module isn't loaded (in this case) */
      if(error == EBADF) {
         cstring2ss(emsg, "", 0);
         error = 0;
      }
      else {
         goto end;
      }
   }

   if(_find_pool(&core_di, pl->name) >= 0) {
         ss_sprintf(emsg, "%s is an assembled pool - disassemble and erase %s"
               " if you want to write a new pool.", pl->name, pl->name);
         error = 1;
         goto end;
   }

   /* Check labels to make sure poolname isn't used there */
   log_info("Checking disks for %s...", pl->name);
   if((error = scan_disks(&disk_di, emsg)) != 0) {
      log_info("ERROR.");
      goto end;
   }
   log_info("ok.");

   if((error = _check_labels(&disk_di, &core_di, pl, emsg)) != 0) {
      goto end;
   }
 
   /* If the pool is on any visible disks handle potential erasure */
   if(_find_pool(&disk_di, pl->name) >= 0) {

      /* verify pool erasure */
      if (!check_flag(POOL_FORCE)) {
         printf("Pool %s exists.  Do you want to erase it? (y/n)",
                pl->name);
         if(tolower(_get_a_char(stdin)) == 'y') {
            if(!check_flag(POOL_FORCE)) {
               add_flag(POOL_FORCE);
               set_force = 1;
            }
            error = _erase_pool(pl->name, emsg);
            if(set_force)
               del_flag(POOL_FORCE);
            if(error != 0) {
               goto end;
            }
         }
         else {
            error = 1;
            goto end;
         }
      }
      else {
         /* user specified FORCE, so don't bother asking them */
         if((error = _erase_pool(pl->name, emsg)) != 0) {
            goto end;
         }
      }
   }

   error = _write_config_to_disk(pl, emsg);

end:
   destroy_di(&core_di);
   destroy_di(&disk_di);
   return error;
}

/*
 * _revert_labels - reverts ondisk labels to the labels seen (in di)
 * @name: name of pool to revert
 * @di: diskinfo structure containing pools
 * @emsg:
 *
 */
static int _revert_labels(char *pname, struct diskinfo *di,
                          struct sstring *emsg)
{
   int error = 1;
   int i;

   for(i = 0; i < di->pl_count; i++) {
      if(strcmp(pname, di->pl[i].name) == 0) {
         error = _write_config_to_disk(&(di->pl[i]), emsg);
         break;
      }
   }
   if(i == di->pl_count) {
           ss_sprintf(emsg, "Revert failed: Unable to find %s in old disk"
                      " information structure", pname);
   }

   return error;
}

/* _grow_pool - grows a pool based on the conffile
 * @conffile: config file with new subpools for growth
 *
 * Description: New config file must have the same subpools as the old
 * config file, plus at least one more.  The new subpools are they only
 * way to add space to a pool.  Pool must be assembled to grow it because
 * the parser can't detect when pool labels are written to the same device
 * (with different names) twice.
 *
 */
static int _grow_pool(char *conffile, struct sstring *emsg)
{

   int error = 0;
   int len = 0;
   user_pool_t *pl = NULL;
   diskinfo_t di = {0};
   diskinfo_t core_di = {0};
   char msg[STRING_SIZE];

   if((error = _read_config(conffile, &pl, emsg))) {
      goto end;
   }

   if((error = get_loaded_pools(&core_di, emsg))) {
      if(error == EBADF) {
         ss_sprintf(emsg, "Unable to grow if pool module is not loaded");
      }
      else {
         log_info("FAILED");
      }
      goto end;
   }

   /* Get the version of the config that's on disk */
   log_info("Scanning disks for pool %s...", pl->name);
   if((error = scan_disks(&di, emsg))) {
      log_info("FAILED");
      goto end;
   }
   log_info("ok");

   /* Verify they match (other than added subpools) */
   log_info("Checking validity of config versus ondisk information...");
   if((error = _verify_pool_to_grow(pl, &di, emsg))) {
      log_info("FAILED");
      log_err("Config file not valid for growing pool %s", pl->name);
      goto end;
   }
   log_info("ok");
   log_info("Checking labels...");

   /* check disk labels for ext2, partition tables, etc */
   if((error = _check_labels(&di, &core_di, pl, emsg)) != 0) {
      log_info("FAILED");
      goto end;
   }
   log_info("ok");

   /* Wipe ondisk info */
   destroy_di(&di);

   /* check to make sure kernel has pool assembled */
   if (_find_pool(&core_di, pl->name) >= 0) {
      /* Write out the config */
      if((error = _write_config_to_disk(pl, emsg)) != 0) {
         goto end;
      }

      log_info("Notifying kernel of change in pool size...");
      /* Grow the pool in the kernel */
      if((error = grow_pool(pl->name, emsg)) != 0) {
         log_info("FAILED");
         ss_len(emsg, &len);
         if(len > 0) {
            ss2cstring(msg, STRING_SIZE, emsg);
            log_err(msg);
            ss_sprintf(emsg, "", 0);
         }

         log_err("Attempting to revert to old labels");
         /* revert back based on ondisk labels */
         if((_revert_labels(pl->name, &core_di, emsg)) != 0) {
            ss_len(emsg, &len);
            if(len > 0) {
               ss2cstring(msg, STRING_SIZE, emsg);
               log_err(msg);
               ss_sprintf(emsg, "", 0);
            }
         }
         else {
            log_err("Revert to old labels successful");
         }
      }
      else {
         log_info("ok");
      }
   }
   else {
      error = 1;
      ss_sprintf(emsg, "Pool must be assembled prior to growing");
   }

end:
   /* Wipe the ondisk info...*/
   destroy_di(&di);
   destroy_di(&core_di);
   return error;
}

static int _rename_pool(char *oldname, char *newname, struct sstring *emsg)
{
   int        error = 1;
   diskinfo_t di = {0};

   /* We need to make sure the pool isn't assembled */
   log_info("Checking kernel for %s...", oldname);
   if((error = get_loaded_pools(&di, emsg))) {
      if(error == EBADF) {
         cstring2ss(emsg, "", 0);
         error = 0;
      }
      else {
         log_info("ERROR.");
         goto end;
      }
   }
   else if(_find_pool(&di, oldname) >= 0) {
      log_info("found.");
      log_err("Cannot rename active pool.  Deactivate %s before renaming.",
            oldname);
      error = 1;
      goto end;
   }
   /* If the oldname pool isn't assembled, continue on */
   destroy_di(&di);
   log_info("not found.");
   log_info("Scanning disks...");
   if((error = scan_disks(&di, emsg))) {
      log_info("ERROR.");
      goto end;
   }
   log_info("ok.");

   /* Verify that the old poolname is actually a pool ondisk */
   if(_find_pool(&di, oldname) < 0) {
      log_err("Pool %s not found on disk", oldname);
      error = 1;
      goto end;
   }

   /* Verify that the new poolname isn't on any disk we can currently see */
   if(_find_pool(&di, newname) >= 0) {
      log_err("Pool name %s already in use.  Please use a different name.",
            newname);
      error = 1;
      goto end;
   }
   /* Write new name to disklabel */
   if((error = rename_pool(oldname, newname, emsg)) != 0) {
   }
   else {
      log_warn("%s renamed to %s.", oldname, newname);
   }

end:
   destroy_di(&di);
   return error;
}

static int _reminor_pool(char *oldname, int minor, struct sstring *emsg)
{
   int        error = 1,
              i;
   diskinfo_t di = {0};

   /* We need to make sure the pool isn't assembled */
   log_info("Checking kernel for %s...", oldname);
   if((error = get_loaded_pools(&di, emsg))) {
      if(error == EBADF) {
         cstring2ss(emsg, "", 0);
         error = 0;
      }
      else {
         log_info("ERROR.");
         goto end;
      }
   }
   else if(_find_pool(&di, oldname) >= 0) {
      log_info("found.");
      ss_sprintf(emsg, "Cannot reminor active pool.  "
            "Deactivate %s before reminoring.",
            oldname);
      error = 1;
      goto end;
   }
   /* If the oldname pool isn't assembled, continue on */
   destroy_di(&di);
   log_info("not found.");
   log_info("Scanning disks...");
   if((error = scan_disks(&di, emsg))) {
      log_info("ERROR.");
      goto end;
   }
   log_info("ok.");

   /* Verify that the old poolname is actually a pool ondisk */
   if(_find_pool(&di, oldname) < 0) {
      ss_sprintf(emsg, "Pool %s not found", oldname);
      error = 1;
      goto end;
   }

   /* Verify that the new poolname isn't on any disk we can currently see */
   for(i = 0; i < di.pl_count; i++) {
      if(minor == di.pl[i].minor && strcmp(oldname, di.pl[i].name)) {
         break;
      }
   }
   /* If the minor number is zero, we can ignore any matches */
   if(i < di.pl_count && minor) {
      ss_sprintf(emsg, "Minor number %u already in use.  Please use a"
                 " different number.", minor);
      error = 1;
      goto end;
   }

   /* Write new name to disklabel */
   if((error = reminor_pool(oldname, minor, emsg)) !=0) {
      goto end;
   }
   else {
      log_warn("Minor number for %s changed to %u.", oldname, minor);
   }

end:
   destroy_di(&di);
   return error;

}

/* _scan_and_print_labels - finds the disklabels and prints the info found
 * @emsg:
 *
 */
static int _scan_and_print_labels(struct sstring *emsg)
{
   int        error = 0;
   int        i;
   diskinfo_t di ={0};
   char       type[STRING_SIZE];

   log_notice("Scanning disks");

   if((error = scan_disks(&di, emsg)) != 0) {
      log_notice("failed");

   }
   log_notice("ok");

   /* pool_tool -s */
   if(di.dl_count) {
      log_warn("  %-25s%35s", "Device", "Pool Label");
      log_warn("  %-25s%35s", "======", "==========");
   }
   for(i = 0; i < di.dl_count; i++) {
      if(di.dl[i].lt == pool_label) {
      log_warn("  %-25s%35s", di.dl[i].dev_path, di.dl[i].name);
      }
      else {
         strcpy(type, PRE_TAG);
         strncat(type, di.dl[i].name, STRING_SIZE-(sizeof(PRE_TAG POST_TAG)));
         strcat(type, POST_TAG);
         log_warn("  %-25s%35s", di.dl[i].dev_path, type);
      }
   }

   return error;
}

static int _erase_label(user_pool_t *pool, struct sstring *emsg)
{
   int error = 0;
   int len;
   char msg[STRING_SIZE];

   /* Check with user about erasure */
   if(check_flag(POOL_FORCE) == 0) {
      printf("Are you sure you want to erase the labels for %s? (y/n)",
             pool->name);
      if(tolower(_get_a_char(stdin)) != 'y') {
         error = 1;
         log_err("aborting");
         goto end;
      }
   }

   log_info("Erasing labels for %s...", pool->name);
   if((error = erase_label(pool, emsg)) != 0) {
      log_info("failed.");
      /* FIXME: not sure that this is really what i want to do... */
/*      log_notice(emsg);*/
      ss_len(emsg, &len);
      if(len > 0) {
         ss2cstring(msg, STRING_SIZE, emsg);
         log_err(msg);
      }
      ss_sprintf(emsg, "");
      log_err("Unable to erase label for %s", pool->name);
      goto end;
   }
   else {
      log_info("ok");
      log_warn("Erased %s", pool->name);
   }

  end:
   return error;
}
/* _erase_pool
 * @poolname: name of pool to erase
 * @emsg: error string
 */
static int _erase_pool(char *poolname, struct sstring *emsg)
{
   int        i,
              index,
              error = 0;
   diskinfo_t di = {0};

   log_info("Scanning disks");
   if((error = scan_disks(&di, emsg))) {
      log_info("failed");
      goto end;
   }
   log_info("ok");

   if(poolname) {
      log_info("looking for %s", poolname);
      if((index = _find_pool(&di, poolname)) < 0) {
         log_info("failed");
         error = 1;
         ss_sprintf(emsg, "No pool named %s detected", poolname);
         goto end;
      }
      log_info("ok");

      error = _erase_label(&di.pl[index], emsg);
   }
   else {
      for(i = 0; i < di.pl_count; i++) {
         error += _erase_label(&(di.pl[i]), emsg);
      }
   }

end:
   destroy_di(&di);
   return error;
}

/*
 * vim: et ts=3 sw=3 noai cin :
 */

