/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**  
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <fcntl.h>
#include <dirent.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
#include <errno.h>
#include <linux/raw.h>
#include <linux/kdev_t.h>

#include "ccs.h"
#include "ccs_ft.h"

#include "copyright.cf"
#include "ccs_release.cf"
#include "ccs_ondisk.h"
#include "ccs_hash.h"
#include "test_ccs.h"
#include "iddev.h"

#define DEFAULT_BBS 16384 /* default basic block size */
#define VERBOSE 0x1
#define LIST    0x2
#define DIFF    0x4
#define TEST    0x8
#define BUILD   0x10
#define RESTORE 0x20
#define STD_OUT 0x40
#define OVERRIDE 0x80
#define NO_CHECK 0x100

#define PAGE_SIZE (65536)

int flags=0;

static void print_usage(FILE *stream);
static int ok_to_overwrite(char *dev);
static int bind_to_raw_device(char *dev, char *rdev);
static int extract_archive(char *source, char *dest);
static int extract_file_archive(int arch_fd, char *dest);
static int extract_device_archive(int dev_fd, char *dest);
static int create_archive(char *source, char *dest);
static int create_tmp_archive(char *dir_name, char *tmp_file);
static int create_file_archive(char *tmp_file, char *arch_file);
static int create_device_archive(char *tmp_file, char *arch_device);

int main(int argc, char *argv[]){
  int c;
  char *source = NULL;
  char *dest = NULL;

  while((c = getopt(argc, argv, "hOoVvX")) != -1){
    switch(c){
    case 'h':
      print_usage(stdout);
      exit(EXIT_SUCCESS);
      break;
    case 'O':
      flags |= OVERRIDE;
      break;
    case 'o':
      flags |= STD_OUT;
      break;
    case 'V':
      printf("%s %s (built %s %s)\n", argv[0], CCS_RELEASE_NAME,
	     __DATE__, __TIME__);
      printf("%s\n", SISTINA_COPYRIGHT);
      exit(EXIT_SUCCESS);
      break;      
    case 'v':
      flags |= VERBOSE;
      break;
    case 'X':
      flags |= NO_CHECK;
      break;
    default:
      print_usage(stderr);
      exit(EXIT_FAILURE);
    }
  }

  if(optind < argc){
    if(!strcmp(argv[optind], "create")){
      flags |= BUILD;
    }
    else if(!strcmp(argv[optind], "extract")){
      flags |= RESTORE;
    }
    else if(!strcmp(argv[optind], "diff")){
      flags |= DIFF;
    }
    else if(!strcmp(argv[optind], "list")){
      flags |= LIST;
    }
    else if(!strcmp(argv[optind], "test")){
      flags |= TEST;
      flags |= BUILD;
    }
    else {
      fprintf(stderr, "Unknown command, %s.\n"
	      "Try 'ccs_tool -h' for help.\n", argv[optind]);
      exit(EXIT_FAILURE);
    }
  } else {
    fprintf(stderr, "Too few arguments.\n"
	    "Try 'ccs_tool -h' for help.\n");
    exit(EXIT_FAILURE);
  }

  if((flags & STD_OUT) && !(flags & RESTORE)){
    fprintf(stderr, "'-o' is only valid with 'extract'.\n"
	    "Try 'ccs_tool -h' for help.\n");
    exit(EXIT_FAILURE);
  }

  optind++;
  if(optind < argc){
    source = argv[optind];
  } else {
    fprintf(stderr, "Too few arguments.\n"
	    "Try 'ccs_tool -h' for help.\n");
    exit(EXIT_FAILURE);
  }

  optind++;
  if(flags & BUILD){  /* requires dest arg */
    if(!(flags & TEST)){
      if(optind < argc){
	dest = argv[optind++];
      } else {
	fprintf(stderr, "Too few arguments.\n"
		"Try 'ccs_tool -h' for help.\n");
	exit(EXIT_FAILURE);
      }
    } else {
      dest = NULL;
    }
  } else if(flags & (RESTORE|DIFF)){ /* optional dest arg */
    if(optind < argc){
      dest = argv[optind++];
    }
  } else { /* no dest arg */
    if(optind < argc){
      fprintf(stderr, "Too many arguments.\n"
	      "Try 'ccs_tool -h' for help.\n");
      exit(EXIT_FAILURE);
    }
  }

  if(optind != argc){
    fprintf(stderr, "Too many arguments.\n"
	    "Try 'ccs_tool -h' for help.\n");
    exit(EXIT_FAILURE);
  }
    
  if(flags & BUILD){
    if(create_archive(source, dest)){
      if(flags & TEST){
	fprintf(stderr, "Test failed.\n");
      } else {
	fprintf(stderr, "Failed to create archive.\n");
      }
      exit(EXIT_FAILURE);
    } else {
      if(flags & TEST){
	fprintf(stderr, "Test successful.\n");
      }
    }      
  } else { /* extract, list, diff */
    if(extract_archive(source, dest)){
      fprintf(stderr, "Unable to extract archive, %s\n", source);
      exit(EXIT_FAILURE);
    }
  }
  exit(EXIT_SUCCESS);
}


/* print_usage
 * @stream: where the usage will be printed (stdout, stderr)
 *
 * Returns: nothing
 */
static void print_usage(FILE *stream){
  fprintf(stream,
	  "Usage:\n\n"
	  "ccs_tool [options] <command>\n"
	  "\nOptions:\n\n"
	  " -h         Help.\n"
	  " -O         Override.  Force action without question.\n"
	  " -o         Print archive to stdout.  (Used only with extract.)\n"
	  " -V         Print version information.\n"
	  " -v         Verbose.  List files processed.\n"	  
	  "\nCommands:\n"
	  "\n"
	  " create     Creates a cluster configuration archive.\n"
	  "   ccs_tool create <directory> <device archive>\n"
	  "   or\n"
	  "   ccs_tool create <directory> <file archive>\n"
	  "\n"
	  " diff       Check differences between archive and CCS files.\n"
	  "   ccs_tool diff <device archive> [directory]\n"
	  "   or\n"
	  "   ccs_tool diff <file archive> [directory]\n"
	  "\n"
	  " extract    Extracts files from a cluster configuration archive.\n"
	  "   ccs_tool extract <device archive> [directory]\n"
	  "   or\n"
	  "   ccs_tool extract <file archive> [directory]\n"
	  "\n"
	  " list       List the files in a cluster configuration archive.\n"
	  "   ccs_tool list <device archive>\n"
	  "   or\n"
	  "   ccs_tool list <file archive>\n"
	  "\n"
	  " test       Test the CCS files for format correctness.\n"
	  "   ccs_tool test <directory>\n"
	  );
}


/* ok_to_overwrite
 * @dev: device to check contents of
 *
 * This function checks to see if there are any existing FSs on
 * the device before use.  If there is, the user will be prompted
 * if they wish to continue to overwrite the FS.
 *
 * Returns: 1 if ok to overwrite, 0 otherwise
 */
static int ok_to_overwrite(char *dev){
  int fd;
  char buf[1024];
  char input[32];
  int unknown;

  if((fd = open(dev, O_RDONLY)) < 0){
    fprintf(stderr, "Unable to open %s: %s\n", dev, strerror(errno));
    return 0;
  }

  unknown = identify_device(fd, buf, 1024);
  if (!unknown && strcmp(buf, "CCA device")){
    printf("\n%s appears to contain a %s.\n", dev, buf);

    printf("Are you sure you want to proceed? [y/n] ");
    fgets(input, 32, stdin);

    if (input[0] != 'y')
      return 0;
  }
  return 1;
}


/* unbind_raw_device
 * @dev: the device to unbind
 *
 * This function takes the name of the device that is bound to a
 * raw device (e.g. /dev/pool/brassow) and releases the raw
 * device from use.
 */
static void unbind_raw_device(char *dev){
  int raw_fd, i;
  int new_raw_method=1;
  int major, minor;
  struct raw_config_request raw_req;
  struct stat stat_buf;

  if(stat(dev, &stat_buf)){
    fprintf(stderr, "Unable to stat %s: %s\n",
	    dev, strerror(errno));
    return;
  }
  if(!S_ISBLK(stat_buf.st_mode)){
    fprintf(stderr, "%s is not a block device.\n", dev);
    return;
  } else {
    major = major(stat_buf.st_rdev);
    minor = minor(stat_buf.st_rdev);
  }

  /* Are we using the old way, or the new way? */
  if(stat("/dev/raw", &stat_buf)){
    fprintf(stderr, "Unable to stat /dev/raw: %s\n",
	    strerror(errno));
    return;
  }
  if(S_ISDIR(stat_buf.st_mode)){  /* The new way */
    new_raw_method = 1;
    if((raw_fd = open("/dev/rawctl", O_RDWR)) < 0){
      fprintf(stderr, "Unable to open raw control device: %s\n",
	      strerror(errno));
      return;
    }
  } else {  /* The old way */
    new_raw_method = 0;
    if((raw_fd = open("/dev/raw", O_RDWR)) < 0){
      fprintf(stderr, "Unable to open raw control device: %s\n",
	      strerror(errno));
      return;
    }
  }

  for(i=64; i<256; i++){
    raw_req.raw_minor = i;
    if(ioctl(raw_fd, RAW_GETBIND, &raw_req)< 0){
      fprintf(stderr, "Unable to perform ioctl on raw control device: %s\n",
	      strerror(errno));
      close(raw_fd);
      return;
    }
    if(raw_req.block_major == major && raw_req.block_minor == minor){
      raw_req.block_major = raw_req.block_minor = 0;
      if(ioctl(raw_fd, RAW_SETBIND, &raw_req)< 0){
	/*  It is possible for this to fail with EBUSY if the raw device **
	**  is bound and open.  This is not an error.................... */
	if(errno != EBUSY){
	  fprintf(stderr,
		  "Unable to perform ioctl on raw control device: %s\n",
		  strerror(errno));
	}
	close(raw_fd);
	return;
      }
    }
  }
  close(raw_fd);
}


static void _unbind_raw_device(int unused, void *arg){
  unbind_raw_device((char *)arg);
}


/* bind_to_raw_device
 * @dev: the block device to bind to raw device
 * @rdev: the raw device assigned - space must be allocated
 *
 * This function binds a block device to a specific raw device.  The
 * name of the raw device is returned in the space provided.
 *
 * Returns: 0 on success, -EXXX on error
 */
static int bind_to_raw_device(char *dev, char *rdev){
  int raw_fd, i, found=0;
  int new_raw_method=1;
  int major, minor;
  struct raw_config_request raw_req;
  struct stat stat_buf;

  if(stat(dev, &stat_buf)){
    fprintf(stderr, "Unable to stat %s: %s\n",
	    dev, strerror(errno));
    return -errno;
  }
  if(!S_ISBLK(stat_buf.st_mode)){
    fprintf(stderr, "%s is not a block device.\n", dev);
    return -ENOTBLK;
  } else {
    major = major(stat_buf.st_rdev);
    minor = minor(stat_buf.st_rdev);
  }

  /* Are we using the old way, or the new way? */
  if(stat("/dev/raw", &stat_buf)){
    fprintf(stderr, "Unable to stat /dev/raw: %s\n",
	    strerror(errno));
    return -errno;
  }
  if(S_ISDIR(stat_buf.st_mode)){  /* The new way */
    new_raw_method = 1;
    if((raw_fd = open("/dev/rawctl", O_RDWR)) < 0){
      fprintf(stderr, "Unable to open raw control device: %s\n",
	      strerror(errno));
      return -errno;
    }
  } else {  /* The old way */
    new_raw_method = 0;
    if((raw_fd = open("/dev/raw", O_RDWR)) < 0){
      fprintf(stderr, "Unable to open raw control device: %s\n",
	      strerror(errno));
      return -errno;
    }
  }

  for(i=64; i<256; i++){
    raw_req.raw_minor = i;
    if(ioctl(raw_fd, RAW_GETBIND, &raw_req)< 0){
      fprintf(stderr, "Unable to perform ioctl on raw control device: %s\n",
	      strerror(errno));
      close(raw_fd);
      return -errno;
    }
    if(raw_req.block_major == major && raw_req.block_minor == minor){
      goto already_bound;
    } else if(!raw_req.block_major && !raw_req.block_minor){
      if(!found)
	found = i;
    }
  }
  if(!found){
    fprintf(stderr, "All raw devices are in use.\n");
    close(raw_fd);
    return -EBUSY;
  }
  raw_req.raw_minor = found;
  raw_req.block_major = major;
  raw_req.block_minor = minor;
  if(ioctl(raw_fd, RAW_SETBIND, &raw_req)< 0){
    fprintf(stderr, "Unable to bind %s to a raw device: %s\n",
	    dev, strerror(errno));
    close(raw_fd);
    return -errno;
  }
 already_bound:
  close(raw_fd);
  if(new_raw_method){
    if(snprintf(rdev, 32, "/dev/raw/raw%d", raw_req.raw_minor) < 0){
      fprintf(stderr, "Programmer error, not enough space for return value.\n");
      return -1;
    }
  } else {
    if(snprintf(rdev, 32, "/dev/raw%d", raw_req.raw_minor) < 0){
      fprintf(stderr, "Programmer error, not enough space for return value.\n");
      return -1;
    }
  }
  
  if(stat(rdev, &stat_buf)){
    if(errno == ENOENT){
      /* Create device node */
      printf("The raw device (162,%d) does not exist.  Creating...\n",
	     raw_req.raw_minor);
      if(mknod(rdev,  S_IFCHR | S_IREAD | S_IWRITE,
	       osi_make_dev(162, raw_req.raw_minor))){
	fprintf(stderr, "Unable to make raw device node: %s\n",
		strerror(errno));
	return -errno;
      }
    } else {
      fprintf(stderr, "Unable to stat %s: %s\n",
	      rdev, strerror(errno));
      return -errno;
    }
  }
  
  on_exit(_unbind_raw_device, (void *)dev);
  return 0;
}


/* extract_archive
 * @source: the archive file/device to extract
 * @dest: NULL or the place where the archive will be extracted to
 *
 * This function takes an archive and separates all the files inside.
 * It will create a directory given by <dest> and place the files
 * inside.  If <dest> is NULL, it will use the name of the directory
 * from which the archive was created, which is contained in the
 * archive.  Note: the directory name stored in the archive is
 * only the last directory in the original path.
 *
 * Returns: 0 on success, -1 on error
 */
static int extract_archive(char *source, char *dest){
  int fd=-1;
  struct stat stat_buf;

  if(stat(source, &stat_buf)){
    fprintf(stderr, "The archive, %s, does not exist or is not accessible.\n",
	    source);
    return -1;
  } else if(S_ISBLK(stat_buf.st_mode)){ /* kernel must support direct I/O */
    char raw_source[32];
    if(bind_to_raw_device(source, raw_source)){
      return -1;
    }
    fd = open(raw_source, O_RDONLY/* | O_DIRECT*/);
    if(fd < 0){
      fprintf(stderr, "Unable to open %s (%s): %s\n",
	      source, raw_source, strerror(errno));
      return -1;
    }
    if(extract_device_archive(fd, dest)){
      close(fd);
      return -1;
    }
  } else if(S_ISREG(stat_buf.st_mode)){
    fd = open(source, O_RDONLY);
    if(fd < 0){
      fprintf(stderr, "Unable to open %s.\n", source);
      return -1;
    }
    if(extract_file_archive(fd, dest)){
      close(fd);
      return -1;
    }
  } else {
    fprintf(stderr, "%s is not a valid archive file.\n", source);
    return -1;
  }

  return 0;
}


/* extract_file_archive
 * @arch_fd: the archive file from which to extract
 * @dest: NULL or the destination directory
 *
 * Operates the same as described in extract_archive, except
 * that this function is specific to archive files.
 *
 * Returns: 0 on success, -1 on error
 */
static int extract_file_archive(int arch_fd, char *dest){
  FILE *arch_fp=NULL;
  int fd = -1;
  char cwd[256];
  DIR *dir=NULL;
  char buffer[4096];
  char line[256];
  char final_dest[256];
  struct stat stat_buf;

  arch_fp = fdopen(arch_fd, "r");
  if(arch_fp == NULL){
    fprintf(stderr, "Unable to open archive file.\n");
    return -1;
  }
  if(fgets(line, sizeof(line), arch_fp) == NULL){
    perror("Unable to read archive file");
    goto fail;
  }

  if(strncmp(line, "#CCA:", 5)){
    fprintf(stderr, "The specified file is not a CCS archive.\n");
    goto fail;
  }

  if(!getcwd(cwd, sizeof(cwd))){
    fprintf(stderr, 
	    "The path to the archive is too long "
	    "(greater than %u characters)\n", (unsigned int)sizeof(cwd)-1);
    goto fail;
  }
	      
  if(!dest){
    char tmp[256];
    sscanf(line, "#CCA:%s\n", tmp);

    sprintf(final_dest, "%s/%s", cwd, tmp);
  } else {
    if(dest[0] == '/'){ /* absolute path */
      strcpy(final_dest, dest);
    } else {
      sprintf(final_dest, "%s/%s", cwd, dest);
    }
  }

  if(!(flags & (LIST | DIFF | STD_OUT))){
    if(stat(final_dest, &stat_buf)){
      if(mkdir(final_dest, S_IRWXU)){
	fprintf(stderr, "Unable to make directory, %s\n", final_dest);
	goto fail;
      }
    } else { /* stat succeeded */
      if(flags & OVERRIDE){
	if(!S_ISDIR(stat_buf.st_mode)){
	  fprintf(stderr, "%s not a directory.\n", final_dest);
	  goto fail;
	}
      } else {
	fprintf(stderr, "%s already exists.\n", final_dest);
	goto fail;
      }
    }
  }
  while(fgets(line, sizeof(line), arch_fp)){
    uint mtime=0;
    uint size=0;
    char filename[256];
    char filename_cmp[256];

    if(sscanf(line, "#nigeb=%s mtime=%u size=%u\n",
	      filename, &mtime, &size) == 3){
      int diff, amount;
      char full_filename[256];
      sprintf(full_filename, "%s/%s", final_dest, filename);

      if(flags & (LIST | DIFF)){
	diff=size;
	amount=(diff < sizeof(buffer))?diff:sizeof(buffer);

	while(diff){
	  if(fread(buffer, sizeof(char), amount, arch_fp) != amount){
	    fprintf(stderr, "Unable to complete read of the archive.\n");
	    goto fail;
	  }
	  diff -= amount;
	  amount=(diff < sizeof(buffer))?diff:sizeof(buffer);
	}
      } else {
	if(flags & STD_OUT){
	  fd = STDOUT_FILENO;
	} else {
	  if((fd = open(full_filename, O_RDWR | O_CREAT | O_TRUNC,
			S_IWUSR|S_IRUSR)) < 0){
	    fprintf(stderr, "Unable to create file, %s\n", full_filename);
	    goto fail;
	  }
	}
	diff=size;
	amount=(diff < sizeof(buffer))?diff:sizeof(buffer);

	while(1){
	  if(fread(buffer, sizeof(char), amount, arch_fp) != amount){
	    fprintf(stderr, "Unable to complete read of the archive.\n");
	    goto fail;
	  }
	  diff -= amount;
	  if(write(fd, buffer, amount) < amount){
	    fprintf(stderr, "Unable to complete write to file, %s\n",
		    full_filename);
	    goto fail;
	  }
	  if(!diff){
	    uint32 tmp_hash;
	    if(!fgets(line, sizeof(line), arch_fp) ||
	       !sscanf(line, "#dne=%s hash=%X\n", filename_cmp, &tmp_hash) ||
	       strcmp(filename, filename_cmp)){
	      fprintf(stderr, "Archive does not contain end of file %s\n",
		      full_filename);
	      if(!(flags & STD_OUT)){
		close(fd); fd = -1;
	      }
	      unlink(full_filename);
	      goto fail;
	    } else {
	      break;
	    }
	  }
	  amount=(diff < sizeof(buffer))?diff:sizeof(buffer);
	}
      }
      if(flags & LIST){
	char *a,*b;
	a = final_dest;
	for(b=strstr(a,"/"); b; a=b+1,b=strstr(a,"/"));
	printf("%s/%s\n", a, filename);
      } else if(flags & DIFF){
	if(stat(full_filename, &stat_buf)){
	  fprintf(stderr, "Warning: Cannot stat %s\n", full_filename);
	} else {
	  if(stat_buf.st_mtime != mtime) printf("%s: Mod time differs\n",
						filename);
	  if(stat_buf.st_size != size) printf("%s: Size differs\n",
					      filename);
	}
      } else if(flags & VERBOSE){
	char *a,*b;
	a = final_dest;
	for(b=strstr(a,"/"); b; a=b+1,b=strstr(a,"/"));
	printf("%s/%s\n", a, filename);
      }	
      if(!(flags & STD_OUT)){
	close(fd); fd = -1;
      }
    }
  }

  return 0;
      
 fail:
  if((fd>0) && !(flags & STD_OUT))
    close(fd);
  if(arch_fp)
    fclose(arch_fp);
  if(dir)
    closedir(dir);
  return -1;
}


/* extract_file_archive
 * @dev_fd: the archive file from which to extract
 * @dest: NULL or the destination directory
 *
 * Operates the same as described in extract_archive, except
 * that this function is specific to archive devices.
 *
 * Returns: 0 on success, -1 on error
 */
static int extract_device_archive(int dev_fd, char *dest){
  int tmp_fd=-1;
  unsigned long remainder;
  char tmp_file[100];
  char *sector_orig = NULL;
  char *sector = NULL;
  ccs_dh_t dev_header;

  sector_orig = (char *)malloc(DEFAULT_BBS + PAGE_SIZE);
  if(!sector_orig){
    fprintf(stderr, "Not enough memory.\n");
    return -1;
  }
  remainder = ((unsigned long)sector_orig) & (PAGE_SIZE-1);
  sector = sector_orig + (PAGE_SIZE - remainder);

  if(read(dev_fd, sector, DEFAULT_BBS) < DEFAULT_BBS){
    fprintf(stderr, "Bad read on the archive device.\n");
    goto fail;
  }

  ccs_dh_in(&dev_header, sector);
  if(dev_header.dh_magic == CCS_DH_MAGIC){
    if(!dev_header.dh_active){
      lseek(dev_fd, dev_header.dh_maxsize, SEEK_SET);
    }
  } else {
    fprintf(stderr, "The specified device does not contain a CCS archive.\n");
    goto fail;
  }

  sprintf(tmp_file, "/tmp/tmp_%d", getpid());

  tmp_fd = open(tmp_file, O_RDWR | O_CREAT | O_TRUNC, S_IRUSR|S_IWUSR);
  if(tmp_fd < 0){
    fprintf(stderr, "Unable to create temporary archive.\n");
    goto fail;
  }    
  unlink(tmp_file);

  while(dev_header.dh_size){
    int write_size;
    if(read(dev_fd, sector, DEFAULT_BBS) < DEFAULT_BBS){
      fprintf(stderr, "Bad read on the archive device.\n");
      goto fail;
    }
    write_size = (dev_header.dh_size < DEFAULT_BBS) ? 
      dev_header.dh_size : DEFAULT_BBS;
    if(write(tmp_fd, sector, write_size) < write_size){
      fprintf(stderr, "Unable to write to temporary archive.\n");
      goto fail;
    }
    dev_header.dh_size -= write_size;
  }
  lseek(tmp_fd, 0, SEEK_SET);
  
  if(extract_file_archive(tmp_fd, dest)){
    goto fail;
  }

  free(sector_orig);
  close(dev_fd);
  return 0;

 fail:
  if(tmp_fd>0)
    close(tmp_fd);
  free(sector_orig);
  return -1;
}


/*
 * validate_ccs_files: ensure files in source are valid CCS files
 * @source: directory containing CCS files to be checked
 *
 * This function ensures that files with the .ccs extension, that
 * are contained in the source directory, have valid CCS format.
 *
 * Returns: -1 on error, 0 otherwise
 */
static int validate_ccs_files(char *source){
  int fd=-1;
  DIR *dir = NULL;
  struct dirent *dirent;
  char filename[256];
  char *filedata = NULL;
  struct stat statbuf;
  ccs_node_t *cn;

  dir = opendir(source);
  if(!dir)
    return -1;

  while((dirent = readdir(dir))){
    if((dirent->d_name[strlen(dirent->d_name)-4] == '.') &&
       (dirent->d_name[strlen(dirent->d_name)-3] == 'c') &&
       (dirent->d_name[strlen(dirent->d_name)-2] == 'c') &&
       (dirent->d_name[strlen(dirent->d_name)-1] == 's')
       ){
      sprintf(filename, "%s/%s", source, dirent->d_name);
      if(stat(filename, &statbuf)){
	goto fail;
      }
      filedata = malloc(statbuf.st_size + 1);
      if(!filedata){
	goto fail;
      }
      fd = open(filename, O_RDONLY);
      if(fd < 0){
	goto fail;
      }
      if(read(fd, filedata, statbuf.st_size) != statbuf.st_size){
	goto fail;
      }
      close(fd); fd = -1;
      filedata[statbuf.st_size] = '\0';
      if(parse_filedata(&cn, filedata)){
	fprintf(stderr, "Error while parsing \"%s\"\n", filename);
	goto fail;
      } else {
	free_ccs_node(cn);
	free(filedata);
	filedata = NULL;
      }
    }
  }

  closedir(dir);
  return 0;

 fail:
  closedir(dir);
  if(filedata)
    free(filedata);
  if(fd >= 0)
    close(fd);
  return -1;
}


/*
 * create_archive
 * @source: the directory containing the files to be archived
 * @dest: the final resting place of the archive
 *
 * This function creates a Cluster Configuration Archive, which
 * is (basically) all the files in the source directory appended
 * together.
 *
 * Returns: 0 on success, -1 on error
 */
static int create_archive(char *source, char *dest){
  struct stat stat_buf;
  char tmp_file[256];
  char *ptr;

  /* check source validity */
  if(stat(source, &stat_buf)){
    if(errno == ENOENT){
      fprintf(stderr, "The directory, %s does not exist.\n", source);
    } else {
      fprintf(stderr, "Unable to stat \"%s\": ", source);
      perror(NULL);
    }
    return -1;
  } else if(!S_ISDIR(stat_buf.st_mode)){
    fprintf(stderr, "%s is not a directory.\n", source);
    return -1;
  }

  if(flags & NO_CHECK){
    /* If no check is specified, we still make sure format is correct */
    if(validate_ccs_files(source)){
      return -1;
    }
  } else {
    /* more extensive checking of CCS files through library */
    if(validate_ccs_files(source) || test_ccs(source, flags & VERBOSE)){
      return -1;
    }
  }

  if(!dest)
    return 0;

  /* Check destination validity */
  for(ptr=dest; (ptr = strstr(ptr, "/")); ptr++){
    if(ptr == dest) continue; /* absolute path */
    memset(tmp_file, 0, sizeof(tmp_file));
    strncpy(tmp_file,dest, (ptr-dest));
    if(stat(tmp_file, &stat_buf)){
      fprintf(stderr, "The directory, %s does not exist.\n", tmp_file);
      return -1;
    } else if(!S_ISDIR(stat_buf.st_mode)){
      fprintf(stderr, "Bad path specified.  %s is not a directory.\n",
	      tmp_file);
      return -1;
    }
  }

  if(stat(dest, &stat_buf)){
    if(errno != ENOENT){
      perror(NULL);
      return -1;
    }
    sprintf(tmp_file, "%s_%d", dest, getpid());
    if(create_tmp_archive(source, tmp_file)){
      return -1;
    }

    if(create_file_archive(tmp_file, dest)){
      fprintf(stderr, "Unable to create file \"%s\"\n", dest);
      goto fail;
    }
    unlink(tmp_file);

    if(!strncmp(dest, "/dev/", 5)){
      fprintf(stderr,
	      "WARNING::\n"
	      "You appear to be attempting to write a CCS archive to a device.\n"
	      "However, the device node does not exist.  The file, %s,\n"
	      "has been created.  Be sure this is what you want.\n\n", dest);
    }
  } else if(S_ISBLK(stat_buf.st_mode)){
    if((flags & OVERRIDE) || ok_to_overwrite(dest)){
      sprintf(tmp_file, "/tmp/tmp_%d", getpid());
      if(create_tmp_archive(source, tmp_file)){
	return -1;
      }

      if(create_device_archive(tmp_file, dest)){
	fprintf(stderr, "Unable to write archive to \"%s\"\n", dest);
	goto fail;
      }
      unlink(tmp_file);
    } else {
      return -1;
    }
  } else {
    sprintf(tmp_file, "%s_%d", dest, getpid());
    if(flags & OVERRIDE){
      if(create_tmp_archive(source, tmp_file)){
	return -1;
      }

      if(create_file_archive(tmp_file, dest)){
	fprintf(stderr, "Unable to create file \"%s\"\n", dest);
	goto fail;
      }
      unlink(tmp_file);
    } else {
      char choice[10];
      printf("The file, %s, already exists.  Overwrite (y/n)? ", dest);
      fgets(choice, sizeof(choice), stdin);
      if((choice[0] == 'y') || (choice[0] == 'Y')){
	if(create_tmp_archive(source, tmp_file)){
	  return -1;
	}

	if(create_file_archive(tmp_file, dest)){
	  fprintf(stderr, "Unable to create file \"%s\"\n", dest);
	  goto fail;
	}
	unlink(tmp_file);
      } else {
	return -1;
      }
    }
  }

  return 0;

 fail:
  unlink(tmp_file);
  return -1;
}

/*
 * create_tmp_archive
 * @dir_name: the source directory containing the files
 * @tmp_file: the place where the temporary archive will be written
 *
 * This function is called before create_file_archive/create_device_archive.
 * Its purpose is to layout the archive so that it can be more easily written
 * in sector sizes when writing to a device, or atomically renamed when 
 * writing to a file.
 *
 * Returns: 0 on success, -1 on error
 */ 
static int create_tmp_archive(char *dir_name, char *tmp_file){
  int fd=-1;
  int tmp_fd=-1;
  int amount;
  int files_exist=0;
  int file_index=0;
  int change=1;
  char *entire_file = NULL;
  DIR *dir = NULL;
  char *a,*b;
  char dir_name2[256];
  struct dirent *dirent;  
  char tag[256];

  dir = opendir(dir_name);

  if(!dir){
    fprintf(stderr, "Couldn't open directory \"%s\"\n", dir_name);
    goto fail;
  }

  tmp_fd = open(tmp_file, O_RDWR | O_CREAT | O_TRUNC, S_IRUSR|S_IWUSR);
  if(tmp_fd < 0){
    printf("%s\n", tmp_file);
    perror("Unable to create temporary archive");
    goto fail;
  }    

  strncpy(dir_name2, dir_name, sizeof(dir_name2));
  if(dir_name2[strlen(dir_name2)-1] == '/')
    dir_name2[strlen(dir_name2)-1] = '\0';
  while(change){
    change=0;
    /* if path ends in "/." */
    if((strlen(dir_name2) >= 2) &&
       (dir_name2[strlen(dir_name2)-1] == '.') &&
       (dir_name2[strlen(dir_name2)-2] == '/')){
      for(a=b=dir_name2; (b=strstr(a, "/")); a=b+1);
      if(a>dir_name2){
	(a-1)[0] = '\0';
	change=1;
      }
      if(!strlen(dir_name2)){
	strcpy(dir_name2, "/");
      }
    }

    /* if path ends in "/.." */
    if((strlen(dir_name2) > 3) &&
       (dir_name2[strlen(dir_name2)-1] == '.') &&
       (dir_name2[strlen(dir_name2)-2] == '.') &&
       (dir_name2[strlen(dir_name2)-3] == '/')){
      for(a=b=dir_name2; (b=strstr(a, "/")); a=b+1);
      if(a>dir_name2){
	(a-1)[0] = '\0';
	change=1;
      }
      if(!strlen(dir_name2)){
	strcpy(dir_name2, "/");
      }
      /* can't handle this case */
      if(((strlen(dir_name2)>1) && 
	  (dir_name2[strlen(dir_name2)-1] == '.') &&
	  (dir_name2[strlen(dir_name2)-2] == '/')) ||
	 (!strcmp(dir_name2, "."))){
	fprintf(stderr, "Unable to resolve path \"%s\"\n", dir_name);
	goto fail;
      }
    }
  }

  if(!strcmp(dir_name2, ".")){
    if(!getcwd(dir_name2, sizeof(dir_name2))){
      perror("Unable to get current working directory");
      goto fail;
    }
  }
  else if(!strcmp(dir_name2, "..")){
    if(!getcwd(dir_name2, sizeof(dir_name2))){
      perror("Unable to get current working directory");
      goto fail;
    }
    if(dir_name2[strlen(dir_name2)-1] == '/')
      dir_name2[strlen(dir_name2)-1] = '\0';
    for(a=b=dir_name2; (b=strstr(a, "/")); a=b+1);
    if(a>dir_name2){
      (a-1)[0] = '\0';
    }
    if(!strlen(dir_name2)){
      strcpy(dir_name2, "cluster_configuration_files");
    }
  }
  if(!strcmp(dir_name2, "/")){
    strcpy(dir_name2, "cluster_configuration_files");
  }


  if(dir_name2[strlen(dir_name2)-1] == '/')
    dir_name2[strlen(dir_name2)-1] = '\0';
  for(a=b=dir_name2; (b=strstr(a, "/")); a=b+1);
  memset(tag, 0, sizeof(tag));
  sprintf(tag, "#CCA:%s\n", a);
  amount = write(tmp_fd, tag, strlen(tag));
  if(amount <= 0){
    fprintf(stderr, "Unable to write to temporary archive \"%s\"\n",
	    tmp_file);
    goto fail;
  }

  while((dirent = readdir(dir))){
    if((dirent->d_name[strlen(dirent->d_name)-4] == '.') &&
       (dirent->d_name[strlen(dirent->d_name)-3] == 'c') &&
       (dirent->d_name[strlen(dirent->d_name)-2] == 'c') &&
       (dirent->d_name[strlen(dirent->d_name)-1] == 's')
       ){
      struct stat stat_buf;
      char filename[256];
      char buffer[4096];

      sprintf(filename, "%s/%s", dir_name, dirent->d_name);

      if(stat(filename, &stat_buf)){
	fprintf(stderr, "Unable to stat \"%s\"\n", filename);
	goto fail;
      }

      entire_file = malloc(stat_buf.st_size);
      if(!entire_file){
	fprintf(stderr, "Unable to allocate memory for file, %s\n", filename);
	goto fail;
      }

      if((fd = open(filename, O_RDONLY)) < 0){
	fprintf(stderr, "Unable to open \"%s\"\n", filename);
	goto fail;
      } else {
	char *c;
	a = filename;
	b = strstr(a+1, "/");
	for(c=strstr(b+1,"/"); c; a=b,b=c,c=strstr(b+1,"/"));

	if(flags & VERBOSE) printf("%s\n", (a[0] == '/')?a+1:a);
	strcpy(filename, b+1);
      }
      memset(tag, 0, sizeof(tag));
      sprintf(tag, "#nigeb=%s mtime=%lu size=%u\n",
	      filename, stat_buf.st_mtime, (uint)stat_buf.st_size);
      amount = write(tmp_fd, tag, strlen(tag));
      if(amount <= 0){
	fprintf(stderr, "Unable to write to temporary archive \"%s\"\n",
		tmp_file);
	goto fail;
      }
      while((amount = read(fd, buffer, sizeof(buffer)))){
	if(amount < 0){
	  fprintf(stderr, "Unable to read \"%s\"\n",
		  tmp_file);
	  goto fail;
	}
	if((file_index+amount) > stat_buf.st_size){
	  fprintf(stderr,
		  "The file, %s, has changed during ccs_tool operation.\n",
		  filename);
	  goto fail;
	}
	memcpy(entire_file+file_index, buffer, amount);
	file_index += amount;
	if(write(tmp_fd, buffer, amount) < 0){
	  fprintf(stderr, "Unable to write to temporary archive \"%s\"\n",
		  tmp_file);
	  goto fail;
	}
      }
      close(fd); fd=-1; /* set fd for fail code */
      memset(tag, 0, sizeof(tag));
      /* ATTENTION -- need to be certain that capitols are used with %X ? */
      if(file_index != stat_buf.st_size){
	fprintf(stderr, "Amount read does not equal size of file, %s.\n",
		filename);
	goto fail;
      }
      sprintf(tag, "#dne=%s hash=%.8X\n", filename, 
	      ccs_hash(entire_file, file_index));
      amount = write(tmp_fd, tag, strlen(tag));
      if(amount <= 0){
	fprintf(stderr, "Unable to write to temporary archive \"%s\"\n",
		tmp_file);
	goto fail;
      }
      free(entire_file); entire_file = NULL;
      file_index = 0;
      files_exist++;
    }
  }
  closedir(dir);

  if(lseek(tmp_fd, 0, SEEK_SET) < 0){
    fprintf(stderr, "Unable to reset offset in temporary archive.\n");
    close(tmp_fd); tmp_fd = -1;
  }
  close(tmp_fd);
  if(!files_exist){
    fprintf(stderr, "Refusing to create empty archive.\n");
    unlink(tmp_file);
    return -1;
  }
  return 0;
  
 fail:
  if(entire_file)
    free(entire_file);
  unlink(tmp_file);
  if(fd>0)
    close(fd);
  if(dir)
    closedir(dir);
  if(tmp_fd>0)
    close(tmp_fd);
  return -1;
}


/* create_file_archive
 * @tmp_fd: the temporary archive file descriptor
 * @arch_file: the permanent archive file
 *
 * This function exists to provide an atomic way of committing an
 * archive to a file.
 *
 * Returns: 0 on success, -1 on error
 */
static int create_file_archive(char *tmp_file, char *arch_file){
  int error = 0;
  error = rename(tmp_file, arch_file);
  return error;
}


/* create_device_archive
 * @tmp_fd: the temporary archive file descriptor
 * @arch_device: the permanent archive device
 *
 * This function exits to provide an atomic way of committing an
 * archive to a device.
 *
 * Returns: 0 on success, -1 on failure
 */
static int create_device_archive(char *tmp_file, char *arch_device){
  int tmp_fd;
  int dev_fd=-1;
  int ioctl_fd=-1;
  int amount=0;
  unsigned long remainder;
  char *sector = NULL, *sector_orig= NULL;
  ccs_dh_t dev_header;
  char raw_arch_device[32];

  tmp_fd = open(tmp_file, O_RDONLY);
  if(tmp_fd < 0){
    fprintf(stderr, "Unable to reopen temporary archive.\n");
    return -1;
  }

  if(bind_to_raw_device(arch_device, raw_arch_device)){
    return -1;
  }
  dev_fd = open(raw_arch_device, O_RDWR/* | O_DIRECT*/);
  if(dev_fd < 0){
    fprintf(stderr, "Unable to open device %s (%s): %s\n",
	    arch_device, raw_arch_device, strerror(errno));
    close(tmp_fd);
    return -1;
  }

  ioctl_fd = open(arch_device, O_RDWR);
  if(ioctl_fd < 0){
    fprintf(stderr, "Unable to open device %s: %s\n",
	    arch_device, strerror(errno));
    goto fail;
  }

  sector_orig = (char *)malloc(DEFAULT_BBS + PAGE_SIZE);
  if(!sector_orig){
    fprintf(stderr, "Not enough memory.\n");
    goto fail;
  }
  remainder = ((unsigned long)sector_orig) & (PAGE_SIZE-1);
  sector = sector_orig + (PAGE_SIZE - remainder);

  if(read(dev_fd, sector, DEFAULT_BBS) < DEFAULT_BBS){
    fprintf(stderr, "Bad read on device, \"%s\"\n", arch_device);
    goto fail;
  }

  ccs_dh_in(&dev_header, sector);
  if(dev_header.dh_magic == CCS_DH_MAGIC){
    int error=0;
    uint64 get_size64=0;
    char choice[10];

    if(!(flags & OVERRIDE)){
      printf("A cluster archive already exists on %s.  Overwrite? (y/n) ",
	     arch_device);
      fgets(choice, sizeof(choice), stdin);
      if((choice[0] != 'y') && (choice[0] != 'Y')){
	goto fail;
      }
    }
    if(dev_header.dh_active){
      lseek(dev_fd, dev_header.dh_maxsize, SEEK_SET);
    }

    error = device_size(ioctl_fd, &get_size64);
    if (error < 0){
      fprintf(stderr, "Unable to determine size of device, \"%s\"\n",
	      arch_device);
      goto fail;
    }
    if(get_size64 < dev_header.dh_maxsize*2){
      fprintf(stderr, 
	      "\nThe device size of %s has been altered since the\n"
	      "previous archive was written.\n"
	      "\n"
	      "Reinitialize the device? (y/n) ", arch_device);
      fgets(choice, sizeof(choice), stdin);
      if((choice[0] != 'y') && (choice[0] != 'Y')){
	goto fail;
      } else {
	goto reinit;
      }
    }
  } else { /* Initialize device */
    int tmp;
    int error=0;
    int dev_size=0;
    uint64 get_size64=0;

  reinit:
    error = device_size(ioctl_fd, &get_size64);
    if (error < 0){
      fprintf(stderr, "Unable to determine size of device, \"%s\"\n",
	      arch_device);
      goto fail;
    }

    dev_size = (get_size64 > (100 << 20)) ? (100 << 20) : get_size64;
    dev_size -= dev_size%(DEFAULT_BBS*2);

    if(!dev_size){
      fprintf(stderr, "Not enough room for header.\n");
      fprintf(stderr, "Archive device, %s, is not large enough.\n",
	      arch_device);
      goto fail;
    }
    memset(sector, 0, DEFAULT_BBS);
    lseek(dev_fd, 0, SEEK_SET);
    printf("\nInitializing device for first time use...  ");
    fflush(NULL);
    for(tmp = dev_size; tmp > 0;){
      error = write(dev_fd, sector, (tmp>DEFAULT_BBS)? DEFAULT_BBS: tmp);
      if(error <= 0){
	fprintf(stderr, "Bad write on device, \"%s\"\n", arch_device);
	goto fail;
      }
      
      tmp -= error;
    }
    printf("done.\n");
    lseek(dev_fd, DEFAULT_BBS, SEEK_SET);

    memset(&dev_header, 0, sizeof(dev_header));
    dev_header.dh_magic = CCS_DH_MAGIC;
    dev_header.dh_active = 0;
    dev_header.dh_maxsize = dev_size/2;
    }

  close(ioctl_fd); ioctl_fd=-1;

  memset(sector, 0, DEFAULT_BBS);
  dev_header.dh_size = 0;
  while((amount = read(tmp_fd, sector, DEFAULT_BBS))){
    if(amount < 0){
      fprintf(stderr, "Error while reading temporary archive.\n");
      goto fail;
    }
    if((amount+dev_header.dh_size+DEFAULT_BBS) > dev_header.dh_maxsize){
      fprintf(stderr, "Archive device, %s, is not large enough.\n",
	      arch_device);
      goto fail;
    }
    if(write(dev_fd, sector, DEFAULT_BBS) != DEFAULT_BBS){
      fprintf(stderr, "Error writing to device \"%s\"\n", arch_device);
      goto fail;
    }
    dev_header.dh_size += amount;
    memset(sector, 0, DEFAULT_BBS);
  }

  if(dev_header.dh_size){
    if(dev_header.dh_active){
      dev_header.dh_active = 0;
      dev_header.dh_generation += 1;
      lseek(dev_fd, 0, SEEK_SET);
      memset(sector, 0, DEFAULT_BBS);

      ccs_dh_out(&dev_header, sector);

      if(write(dev_fd, sector, DEFAULT_BBS) < DEFAULT_BBS){
	fprintf(stderr, "Unable to finalize archive on device, \"%s\"\n",
		arch_device);
	goto fail;
      }
    } else {
      dev_header.dh_active = 1;
      dev_header.dh_generation += 1;
      lseek(dev_fd, 0, SEEK_SET);
      memset(sector, 0, DEFAULT_BBS);

      ccs_dh_out(&dev_header, sector);

      if(write(dev_fd, sector, DEFAULT_BBS) < DEFAULT_BBS){
	fprintf(stderr, "Unable to finalize archive on device, \"%s\"\n",
		arch_device);
	goto fail;
      }
    }
  }
  close(dev_fd);
  free(sector_orig);
  return 0;

 fail:
  if(sector_orig) free(sector_orig);
  close(tmp_fd);
  if(ioctl_fd >=0)
    close(ioctl_fd);
  if(dev_fd>0)
    close(dev_fd);
  return -1;
}
