/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2001  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#define GNBD_RELEASE_NAME "v6.0.2"

#define MAJOR_NR gnbd_major	/* force definitions on in blk.h */
int gnbd_major;			/* must be declared before including blk.h */

#define DEVICE_NR(device) MINOR(device)	/* gnbd has no partition bits */
#define DEVICE_NAME "gnbd"	/* name for messaging */
#define DEVICE_NO_RANDOM	/* no entropy to contribute */

#define DEVICE_OFF(d)		/* do-nothing */
#define DEVICE_ON(d)		/* do-nothing */
#define DEVICE_REQUEST gnbd_request

#define LOCAL_END_REQUEST

#include "gnbd.h"
#include "trans.h"
#include "receiver.h"
#include "monitor.h"
#include "fixed_div64.h"

#include <linux/sched.h>
#include <linux/kernel.h>	/* printk() */
#include <linux/fs.h>		/* everything... */
#include <linux/errno.h>	/* error codes */
#include <linux/timer.h>
#include <linux/types.h>	/* size_t */
#include <linux/fcntl.h>	/* O_ACCMODE */
#include <linux/hdreg.h>	/* HDIO_GETGEO */
#include <linux/ioctl.h>
#include <linux/genhd.h>	/* so we can get on /proc/partitions for pool */
#include <linux/devfs_fs_kernel.h>
#include <linux/blkpg.h>
#include <asm/system.h>		/* cli(), *_flags */

#include "macros.h"

static devfs_handle_t gnbd_dir;
static devfs_handle_t gnbd_devfs_entries[GNBD_MAX];

struct rw_semaphore gnbd_devices_lk;
gnbd_t *gnbd_devices[GNBD_MAX];
int gnbd_blksizes[GNBD_MAX];
int gnbd_sizes[GNBD_MAX];
char *gnbd_device_names[GNBD_MAX];
struct hd_struct gnbd_partitions[GNBD_MAX];
static int init_device(char *name, int length, u8 * minor_nr);
int cleanup_device(int minor_num);
uint64_t insmod_time;

int gnbd_open(struct inode *inode, struct file *filp);
int gnbd_release(struct inode *inode, struct file *filp);
int gnbd_ioctl(struct inode *inode, struct file *filp,
	       unsigned int cmd, unsigned long arg);

struct block_device_operations gnbd_device_ops = {
	open:gnbd_open,
	release:gnbd_release,
	ioctl:gnbd_ioctl,
};

struct gendisk gnbd_gendisk = {
	0,			/* Major Number, set by init_module */
	DEVICE_NAME,		/* defined in main.c */
	0,			/* minor_shift. device number is the minor number */
	0,			/* there are NO partitions for this device */
	gnbd_partitions,	/* Partition array */
	gnbd_sizes,		/* Block sizes */
	GNBD_MAX,		/* actual number of devices */
	NULL,			/* real devices pointer. unused */
	NULL,			/* Next */
	&gnbd_device_ops,	/* standard device ops */
	gnbd_devfs_entries,	/* one per physical disc */
	NULL,			/* *flags                   one per physical disc */
#if 0
	gnbd_device_names	/* names of the devices */
#endif
};

static int major = 0;
MODULE_PARM(major, "i");
MODULE_PARM_DESC(major, "Used to change the major number that is used.");

/* 
 * Few random bits of information.
 */

int
gnbd_get_minor(gnbd_t * dev)
{
	int i;
	for (i = 0; i < GNBD_MAX && gnbd_devices[i] != dev; i++) ;
	if (i >= GNBD_MAX)
		return -1;
	return i;
}

void
gnbd_timer_func(unsigned long arg)
{
	gnbd_t *device = (gnbd_t *) arg;
	if (device->sock)
		send_sig(SIGINT, device->receiver_tsk, 0);
}

void
notify_devices(uint32_t ip)
{
	int i;
	for (i = 0; i < GNBD_MAX; i++) {
		if (gnbd_devices[i] && gnbd_devices[i]->servip == ip &&
		    gnbd_devices[i]->timeout) {
			down(&gnbd_devices[i]->notify_lk);
			if (gnbd_devices[i]->server_alive) {
				gnbd_devices[i]->server_alive = FALSE;
				send_sig(SIGTERM, gnbd_devices[i]->receiver_tsk,
					 0);
				complete(&gnbd_devices[i]->wait);
			}
			up(&gnbd_devices[i]->notify_lk);
		}
	}
}

/******************************************************************************
 * empty_trans_queue()
 * Clears out and pending requests by setting the err, ending the request,
 * and then freeing the memory.
 */
void
empty_trans_queue(gnbd_t * dev)
{
	struct list_head *tmp, *nxt;
	struct request *req;

	if (dev == NULL)
		goto exit;

	down(&dev->trans_queue_lk);

	if (list_empty(&dev->queue))
		goto exit_locked;

	for (tmp = dev->queue.next, nxt = tmp->next;
	     tmp != &dev->queue; tmp = nxt, nxt = nxt->next) {
		req = list_entry(tmp, struct request, queue);
		list_del(tmp);
		req->errors = -ESHUTDOWN;
		gnbd_end_request(req);
		yield();
	}
	if (dev->timeout)
		del_timer(&dev->timer);

      exit_locked:
	up(&dev->trans_queue_lk);
      exit:
	return;
}

/******************************************************************************
 * gndb_connect()
 * what needs to be done on a connect action.  Got pulled out here since
 * its called in two places. code reuse, modularity and all that funky
 * jazz.
 */
int
gnbd_connect(int num, gnbd_t * dev)
{
	int err = 0;

	err = start_receiver(num, dev);
	if (!err) {
		uint64_t n = dev->size;
		do_div(n, 1024);
		gnbd_sizes[num] = (int) n;
		gnbd_partitions[num].nr_sects = (unsigned long)2 * gnbd_sizes[num];
	}

	return err;
}

int
gnbd_send_ping(gnbd_t * device)
{
	signal_info_t sig_info;
	trans_hdr_t ping;
	int err;

	ping.type = gnbd_read_req;
	ping.err = 0;
	ping.key = GNBD_PING_KEY;
	ping.offset = cpu_to_be64(0);
	ping.len = cpu_to_be32(0);

	down(&device->trans_queue_lk);
	if (device->timeout && !timer_pending(&device->timer) &&
	    gnbd_connected(device) == INFO_CONNECTED)
		mod_timer(&device->timer,
			  jiffies + ((unsigned long) device->timeout * HZ));
	up(&device->trans_queue_lk);

	block_signals(&sig_info);

	down(&device->sender_lk);

	if (gnbd_connected(device) != INFO_CONNECTED)
		err = -1;
	else {
		err =
		    transfer_info(SEND, device->sock, &ping,
				  sizeof (trans_hdr_t));
		if (err)
			send_sig(SIGINT, device->receiver_tsk, 0);
	}

	up(&device->sender_lk);

	unblock_signals(&sig_info);

	if (err)
		printe("error pinging the server for device %d\n",
		       gnbd_get_minor(device));

	return err;
}

int
gnbd_open(struct inode *inode, struct file *filp)
{
	gnbd_t *dev;		/* device information */
	int num;
	int err = 0;

	if (!inode) {
		err = -EINVAL;
		goto exit;
	}
	num = MINOR(inode->i_rdev);
	printd("opening gnbd minor number %d\n", num);
	down_read(&gnbd_devices_lk);
	if (num >= GNBD_MAX || gnbd_devices[num] == NULL) {
		err = -ENODEV;
		goto exit_dev_lock;
	}
	dev = gnbd_devices[num];
	printd("usage=%d connected=%u pending=%u\n",
	       dev->usage, gnbd_connected(dev), gnbd_pending(dev));
	down(&dev->opener_lk);
	if (num != 0) {		/* don't connect minor 0 */
		if ((filp->f_mode & FMODE_WRITE) && dev->readonly) {
			printe
			    ("cannot open read only gnbd device read/write\n");
			err = -EROFS;
			goto exit_opener_lock;
		}
		err = start_monitor(dev);
		if (err)
			goto exit_opener_lock;
		err = gnbd_connect(num, dev);
		if (err) {
			stop_monitor();
			goto exit_opener_lock;
		}
		dev->usage++;
	}
	MOD_INC_USE_COUNT;

      exit_opener_lock:
	up(&dev->opener_lk);
      exit_dev_lock:
	up_read(&gnbd_devices_lk);
      exit:
	return err;
}

int
gnbd_release(struct inode *inode, struct file *filp)
{
	int num;
	int err = 0;
	gnbd_t *dev;

	if (!inode) {
		err = -EINVAL;
		goto exit;
	}
	num = MINOR(inode->i_rdev);
	down_read(&gnbd_devices_lk);
	if (num >= GNBD_MAX || gnbd_devices[num] == NULL) {
		err = -ENODEV;
		goto exit_dev_lock;
	}
	dev = gnbd_devices[num];
	down(&dev->opener_lk);
	dev->usage--;

	if (num != 0 && !dev->usage) {
		fsync_dev(inode->i_rdev);
		invalidate_buffers(inode->i_rdev);
		stop_receiver(dev);
		empty_trans_queue(dev);
		stop_monitor();
	}
	MOD_DEC_USE_COUNT;

	up(&dev->opener_lk);
      exit_dev_lock:
	up_read(&gnbd_devices_lk);
      exit:
	return 0;
}

static int
init_device(char *name, int length, u8 * minor_nr)
{
	int err = 0;
	int num;
	gnbd_t *temp;

	length++;
	*minor_nr = 0;

	for (num = 0; num < GNBD_MAX; num++) {
		if (gnbd_devices[num] == NULL) {
			if (*minor_nr == 0)
				*minor_nr = num;
		} else if (strcmp(gnbd_device_names[num], name) == 0) {
			printe("gnbd_device name '%s' already in use\n", name);
			err = -EEXIST;
			goto fail;
		}
	}
	if (*minor_nr == 0) {
		printe("could not find a free minor number\n");
		err = -ENODEV;
		goto fail;
	}
	num = (int) *minor_nr;
	temp = kmalloc(sizeof (gnbd_t), GFP_KERNEL);
	if (!temp) {
		printe("could not allocate memory for device %s\n", name);
		err = -ENOMEM;
		goto fail;
	}
	gnbd_device_names[num] = kmalloc(length, GFP_KERNEL);
	if (gnbd_device_names[num] == NULL) {
		printe("couldn't allocate memory for device name: %s\n", name);
		err = -ENOMEM;
		goto fail_malloc;
	}
	if (gnbd_dir != NULL)
		gnbd_devfs_entries[num] =
		    devfs_register(gnbd_dir, name, DEVFS_FL_NONE, gnbd_major,
				   num, S_IFBLK | S_IRWXU, &gnbd_device_ops,
				   NULL);

	strncpy(gnbd_device_names[num], name, length);
	memset(temp, 0, sizeof (gnbd_t));
	temp->size = 0;
	temp->server_alive = FALSE;
	init_MUTEX(&temp->sender_lk);
	init_MUTEX(&temp->trans_queue_lk);
	init_MUTEX(&temp->notify_lk);
	init_MUTEX(&temp->receiver_lk);
	init_MUTEX(&temp->opener_lk);
	INIT_LIST_HEAD(&temp->queue);
	init_timer(&temp->timer);
	temp->timer.function = gnbd_timer_func;
	temp->timer.data = (unsigned long) temp;
	temp->server_name = NULL;
	spin_lock_init(&temp->stat_lock);
	gnbd_devices[num] = temp;

	return 0;

      fail_malloc:
	kfree(temp);
      fail:
	return err;
}

int
cleanup_device(int minor_num)
{
	int err = 0;
	if (gnbd_devices[minor_num]->usage > 0) {
		err = -EBUSY;
		goto out;
	}
	gnbd_sizes[minor_num] = 0;
	if (gnbd_devices[minor_num]->server_name != NULL)
		kfree(gnbd_devices[minor_num]->server_name);
	devfs_unregister(gnbd_devfs_entries[minor_num]);
	gnbd_devfs_entries[minor_num] = NULL;
	gnbd_partitions[minor_num].nr_sects = 0;
	if (gnbd_device_names[minor_num] != NULL) {
		kfree(gnbd_device_names[minor_num]);
		gnbd_device_names[minor_num] = NULL;
	}
	kfree(gnbd_devices[minor_num]);
	gnbd_devices[minor_num] = NULL;

      out:
	return err;
}

int
gnbd_list(char *buf, int *size)
{
	int max = *size;
	int i;
	int count = 0;
	char *ptr = buf;
	gnbd_info_t gnbd_info;

	if (max)
		memset(buf, 0, max);

	for (i = 1; i < GNBD_MAX; i++) {
		if (gnbd_devices[i] == NULL)
			continue;
		gnbd_info.ip = gnbd_devices[i]->servip;
		gnbd_info.timeout = gnbd_devices[i]->timeout;
		gnbd_info.port = gnbd_devices[i]->servport;
		gnbd_info.readonly = gnbd_devices[i]->readonly;
		gnbd_info.minor_nr = i;
		gnbd_info.opened = gnbd_opened(gnbd_devices[i]);
		gnbd_info.connected = gnbd_connected(gnbd_devices[i]);
		gnbd_info.pending = gnbd_pending(gnbd_devices[i]);
		strncpy(gnbd_info.name, gnbd_device_names[i], 256);
		gnbd_info.name[255] = 0;

		count += sizeof (gnbd_info);
		if (max && count <= max) {
			char *tmp = ptr;
			memcpy(tmp, (void *) &gnbd_info, sizeof (gnbd_info));
			ptr += sizeof (gnbd_info);
		}
	}

	*size = count;
	if (!max || count <= max)
		return 0;
	return -1;
}

int
gnbdcomm_ioctl(unsigned int cmd, unsigned long arg)
{
	int err = 0;
	gnbd_t *dev = NULL;
	gnbd_list_arg_t list;
	gnbd_info_t info;

	if (cmd == GNBD_DO_CREATE_DEV || cmd == GNBD_DO_REMOVE_DEV) {
		if (!down_write_trylock(&gnbd_devices_lk)) {
			printe("resources temporarily locked\n");
			err = -ERESTARTSYS;
			goto fail_lock;
		}
	} else
		down_read(&gnbd_devices_lk);

	if (GNBD_ACTION(cmd)) {
		err =
		    (copy_from_user(&info, (void *) arg, sizeof (info))) ?
		    -EFAULT : 0;
		if (err) {
			printe("Couldn't copy action arguments from user\n");
			goto exit;
		}
		if (GNBD_NEEDS_MINOR(cmd)) {
			if (VALID_MINOR_NR(info.minor_nr) == FALSE) {
				printe("invalid minor number %d\n",
				       info.minor_nr);
				err = -EINVAL;
				goto exit;
			}
			dev = gnbd_devices[info.minor_nr];
		}
	}

	switch (cmd) {

	case GNBD_DO_GET_TIME:
		err =
		    (copy_to_user
		     ((void *) arg, &insmod_time,
		      sizeof (uint64_t))) ? -EFAULT : 0;
		if (err)
			printe("couldn't copy time arguments to user\n");
		break;

	case GNBD_DO_LIST:
		{
			gnbd_list_arg_t actual_list;
			err =
			    (copy_from_user(&list, (void *) arg, sizeof (list)))
			    ? -EFAULT : 0;
			if (err) {
				printe
				    ("couldn't copy list arguments from user\n");
				break;
			}
			if (list.size > 0) {
				actual_list.size = list.size;
				actual_list.buf =
				    (char *) kmalloc(list.size, GFP_KERNEL);
				if (!actual_list.buf) {
					printe
					    ("couldn't allocate memory for list data\n");
					err = -ENOMEM;
					break;
				}
			} else
				actual_list.buf = NULL;
			err = gnbd_list(actual_list.buf, &list.size);
			if (err)
				goto list_free;
			if (actual_list.buf) {
				err =
				    (copy_to_user
				     (list.buf, actual_list.buf,
				      actual_list.size)) ? -EFAULT : 0;
				if (err) {
					printe
					    ("couldn't copy list data to user\n");
					goto list_free;
				}
			}
			err =
			    (copy_to_user((void *) arg, &list, sizeof (list))) ?
			    -EFAULT : 0;
			if (err)
				printe
				    ("couldn't copy list arguments to user\n");

		      list_free:
			if (actual_list.buf)
				kfree(actual_list.buf);
		}
		break;

	case GNBD_DO_CREATE_DEV:
		printd("creating device %s\n", info.name);
		err = init_device(info.name, strlen(info.name), &info.minor_nr);
		if (err)
			break;
		err =
		    (copy_to_user((void *) arg, &info, sizeof (info))) ? -EFAULT
		    : 0;
		if (err)
			printe("couldn't copy device information to user\n");
		break;

	case GNBD_DO_REMOVE_DEV:
		printd("removing device %d\n", info.minor_nr);
		err = cleanup_device(info.minor_nr);
		break;

	case GNBD_DO_SET_IP_PORT:
		{
			printd("Setting ip/port on dev %d: ip %x port %d\n",
			       info.minor_nr, info.ip, info.port);
			dev->servip = info.ip;
			dev->servport = info.port;
		}
		break;

	case GNBD_DO_PING:
		printd("Pinging dev %d\n", info.minor_nr);
		err = gnbd_send_ping(dev);
		break;

	case GNBD_DO_CONNECT:
		printd("Connecting dev %d to server\n", info.minor_nr);
		err = gnbd_connect(info.minor_nr, dev);
		if (err)
			break;
		info.readonly = (u8) dev->readonly;
		err =
		    (copy_to_user((void *) arg, &info, sizeof (info))) ? -EFAULT
		    : 0;
		if (err)
			printe("couldn't copy device information to user\n");
		if (gnbd_opened(dev) || gnbd_pending(dev)) {
			printd("exitting with open device\n");
			break;
		}
		printd("rolling through to disconnect\n");
	case GNBD_DO_DISCONNECT:
		if (dev->server_alive) {
			printd("disconnecting %d\n", info.minor_nr);
			stop_receiver(dev);
		}
		break;

	case GNBD_DO_FAIL_SERV:
		{
			u32 server_ip;
			server_ip = (u32) arg;
			notify_devices(server_ip);
		}
		break;

	case GNBD_DO_FORCE_CLOSE:
		err = -ENOTTY;
		break;

	default:
		printe("unknown command");
		err = -ENOTTY;
	}

      exit:
	if (cmd == GNBD_DO_CREATE_DEV || cmd == GNBD_DO_REMOVE_DEV)
		up_write(&gnbd_devices_lk);
	else
		up_read(&gnbd_devices_lk);
      fail_lock:
	return err;
}

/******************************************************************************
 * The ioctl() implementation
 * I am the io control master!
 */

int
gnbd_ioctl(struct inode *inode, struct file *filp,
	   unsigned int cmd, unsigned long arg)
{
	gnbd_t *dev;
	int err = 0;
	kdev_t kdev;
	int num;

	if (!inode) {
		err = -EINVAL;
		goto exit;
	}
	kdev = inode->i_rdev;
	num = MINOR(kdev);
	if (num == 0) {
		err = gnbdcomm_ioctl(cmd, arg);
		goto exit;
	}
	if (VALID_MINOR_NR(num) == FALSE) {
		err = -ENODEV;
		goto exit;
	}
	dev = gnbd_devices[num];

	switch (cmd) {
	case GNBD_STATS:
		{
			gnbd_stat_t stat;
			spin_lock(&dev->stat_lock);
			stat = dev->stat;
			spin_unlock(&dev->stat_lock);
			if (copy_to_user
			    ((void *) arg, &stat, sizeof (gnbd_stat_t)))
				err = -EFAULT;
		}
		break;

	case GNBD_STATS_CLEAR:
		spin_lock(&dev->stat_lock);
		memset(&dev->stat, 0, sizeof (gnbd_stat_t));
		spin_unlock(&dev->stat_lock);
		break;

	case BLKROSET:
	case BLKROGET:
	case BLKRAGET:
	case BLKFLSBUF:
	case BLKSSZGET:
	case BLKGETSIZE:
	case BLKGETSIZE64:
	case BLKELVGET:
	case BLKELVSET:
	case BLKBSZGET:
	case BLKBSZSET:
		err = blk_ioctl(kdev, cmd, arg);
		break;

	case BLKRASET:
		printe("BLKRASET ioctl support does not exist in GNBD\n");
		err = -ENOTTY;
		break;

	case BLKRRPART:
		printe("BLKRRPART ioctl support does not exist in GNBD\n");
		err = -ENOTTY;
		break;

	case HDIO_GETGEO:
	case HDIO_GETGEO_BIG:
		printe("HDIO_GETGEO ioctl support does not exist in GNBD\n");
		err = -ENOTTY;
		break;

	default:
		err = -ENOTTY;
		break;
	}

      exit:
	return err;
}

void
gnbd_end_request(struct request *req)
{
	struct buffer_head *old, *bh;
	unsigned long flags;
	bh = req->bh;
	spin_lock_irqsave(&io_request_lock, flags);
	while (bh != NULL) {
		old = bh;
		bh = bh->b_reqnext;
		old->b_reqnext = NULL;
		old->b_end_io(old, !req->errors);
	}
	blk_finished_io(req->hard_nr_sectors);
	req->errors = 0;
	req->bh = NULL;
	end_that_request_last(req);
	spin_unlock_irqrestore(&io_request_lock, flags);
}

void
gnbd_request(request_queue_t * q)
{
	gnbd_t *device;
	struct request *req;
	trans_hdr_t header;
	int num;
	u32 len;
	u64 offset;
	int result;
	signal_info_t sig_info;

	while (1) {
		INIT_REQUEST;
		if (q->plugged)
			break;
		num = DEVICE_NR(CURRENT_DEV);

		if (num >= GNBD_MAX || gnbd_devices[num] == NULL) {
			printe("request for unknown device\n");
			req = CURRENT;
			blkdev_dequeue_request(req);
			spin_unlock(&io_request_lock);
			req->errors = TRUE;
			gnbd_end_request(req);
			spin_lock(&io_request_lock);
			continue;
		} else
			device = gnbd_devices[num];

		if (!device->server_alive && !device->timeout) {
			printe("server disconnected, quitting request_fn\n");
			break;
		}

		req = CURRENT;
		blkdev_dequeue_request(req);
		spin_unlock_irq(&io_request_lock);

		offset = (u64) (req->sector) * GNBD_HARDSECT;
		len = req->nr_sectors * GNBD_HARDSECT;
		if (offset + len > device->size) {
			printe("request past end of device\n");
			req->errors = TRUE;
			gnbd_end_request(req);
			spin_lock_irq(&io_request_lock);
			continue;
		}
		header.type =
		    (req->cmd == READ) ? gnbd_read_req : gnbd_write_req;
		header.err = 0;
		header.key = (unsigned long) (&req->queue);	/* don't need byte order */
		header.offset = cpu_to_be64(offset);
		header.len = cpu_to_be32(len);

		down(&device->trans_queue_lk);
		if (!device->server_alive) {
			if (!device->timeout) {
				list_add(&req->queue, &device->queue);
				printe
				    ("server disconnected, quitting request_fn\n");
				up(&device->trans_queue_lk);
				spin_lock_irq(&io_request_lock);
				break;
			}
			printe("server disconnected, failing out\n");
			req->errors = TRUE;
			gnbd_end_request(req);
			up(&device->trans_queue_lk);
			spin_lock_irq(&io_request_lock);
			continue;
		}
		if (device->timeout && !timer_pending(&device->timer) &&
		    device->sock)
			mod_timer(&device->timer,
				  jiffies +
				  ((unsigned long) device->timeout * HZ));
		list_add(&req->queue, &device->queue);
		up(&device->trans_queue_lk);

		block_signals(&sig_info);

		result = 0;

		down(&device->sender_lk);

		if (gnbd_connected(device) != INFO_CONNECTED)
			result = -1;
		else {
			if (req->cmd == READ)
				result =
				    transfer_info(SEND, device->sock, &header,
						  sizeof (trans_hdr_t));
			if (req->cmd == WRITE)
				result =
				    send_request_data(device->sock, req,
						      &header, device->iov);
			if (result < 0)
				send_sig(SIGINT, device->receiver_tsk, 0);
		}

		up(&device->sender_lk);

		spin_lock(&device->stat_lock);
		if (req->cmd == READ) {
			device->stat.read_count++;
			device->stat.read_block += req->nr_sectors;
		} else {
			device->stat.write_count++;
			device->stat.write_block += req->nr_sectors;
		}
		spin_unlock(&device->stat_lock);

		unblock_signals(&sig_info);

		if (result != 0 && !device->timeout) {
			printe("error sending request %" PRIx64
			       " on device %d\n", header.key, num);
			spin_lock_irq(&io_request_lock);
			break;
		}

		spin_lock_irq(&io_request_lock);
	}
	return;
}

void
remove_request(gnbd_t * device, struct request *req)
{
	down(&device->trans_queue_lk);
	list_del(&req->queue);
	if (device->timeout > 0) {
		if (gnbd_pending(device))
			mod_timer(&device->timer,
				  jiffies + (device->timeout * HZ));
		else
			del_timer(&device->timer);
	}
	up(&device->trans_queue_lk);
}

int
gnbd_back_merge(request_queue_t * q, struct request *req,
		struct buffer_head *bh, int max_segments)
{
	if (bh->b_size != req->bh->b_size)
		return 0;
	if (req->nr_segments >= MAX_BH_LEN)
		return 0;
	req->nr_segments++;
	return 1;
}

int
gnbd_front_merge(request_queue_t * q, struct request *req,
		 struct buffer_head *bh, int max_segments)
{
	if (bh->b_size != req->bh->b_size)
		return 0;
	if (req->nr_segments >= MAX_BH_LEN)
		return 0;
	req->nr_segments++;
	return 1;
}

int
gnbd_merge_requests(request_queue_t * q, struct request *req,
		    struct request *next, int max_segments)
{
	if (req->bh->b_size != next->bh->b_size)
		return 0;
	if (req->nr_segments + next->nr_segments > MAX_BH_LEN)
		return 0;
	req->nr_segments = req->nr_segments + next->nr_segments;
	return 1;
}

/******************************************************************************
 * Finally, the module stuff
 */
int __init
init_gnbd(void)
{
	int result, i;
	gnbd_t *temp;
	struct timeval tv;

	gnbd_major = major;
	init_rwsem(&gnbd_devices_lk);

	/*
	 * Register your major with a static number or get a dynamic number
	 */
	result = devfs_register_blkdev(gnbd_major, "gnbd", &gnbd_device_ops);
	if (result < 0) {
		printe("can't get major %d\n", gnbd_major);
		goto fail;
	}
	if (gnbd_major == 0)
		gnbd_major = result;	/* dynamic */
	gnbd_gendisk.major = gnbd_major;	/* was unknown at load time */
	/* overwrite the merge requests, to limit clustering to
	   128 requests */
	blk_init_queue(BLK_DEFAULT_QUEUE(gnbd_major), gnbd_request);
	(BLK_DEFAULT_QUEUE(gnbd_major))->back_merge_fn = gnbd_back_merge;
	(BLK_DEFAULT_QUEUE(gnbd_major))->front_merge_fn = gnbd_front_merge;
	(BLK_DEFAULT_QUEUE(gnbd_major))->merge_requests_fn =
	    gnbd_merge_requests;
	blk_queue_headactive(BLK_DEFAULT_QUEUE(gnbd_major), 0);

	for (i = 0; i < GNBD_MAX; i++)
		gnbd_blksizes[i] = BLOCK_SIZE;
	blksize_size[gnbd_major] = gnbd_blksizes;

	for (i = 0; i < GNBD_MAX; i++)
		gnbd_sizes[i] = 0;	/* This is zero so that unused gnbd devices don't
					   show up on /proc/partition */
	blk_size[gnbd_major] = gnbd_sizes;	/* in blocks.. */
	read_ahead[gnbd_major] = 0;
	hardsect_size[gnbd_major] = NULL;
	max_readahead[gnbd_major] = NULL;
	max_sectors[gnbd_major] = NULL;
	blkdev_varyio[gnbd_major] = NULL;
	/* 
	 * allocate the devices -- we can't have them static, as the number
	 * can be specified at load time
	 */

	for (i = 0; i < GNBD_MAX; i++) {
		gnbd_partitions[i].nr_sects = 0;
		gnbd_partitions[i].start_sect = 0;	/* disable partition checking */
	}
	add_gendisk(&gnbd_gendisk);

	gnbd_dir = devfs_mk_dir(NULL, "gnbd", NULL);

	/* initialize gnbdcomm (the node used to communicate with gnbd) */
	temp = kmalloc(sizeof (gnbd_t), GFP_KERNEL);
	if (!temp) {
		printe("could not allocate memory for gnbdcomm\n");
		result = -ENOMEM;
		goto fail_cleanup;
	}
	gnbd_device_names[0] = kmalloc(sizeof (char) * 9, GFP_KERNEL);
	if (!gnbd_device_names[0]) {
		kfree(temp);
		printe("couldn't allocate memory for device name gnbdcomm\n");
		result = -ENOMEM;
		goto fail_cleanup;
	}
	strcpy(gnbd_device_names[0], "gnbdcomm");
	memset(temp, 0, sizeof (gnbd_t));
	init_MUTEX(&temp->opener_lk);
	gnbd_devices[0] = temp;

	gnbd_devfs_entries[0] = devfs_register(NULL, "gnbdcomm", DEVFS_FL_NONE,
					       gnbd_major, 0, S_IFBLK | S_IRWXU,
					       &gnbd_device_ops, NULL);

	init_monitor();

	do_gettimeofday(&tv);
	insmod_time = (uint64_t) tv.tv_sec * 1000000 + tv.tv_usec;

	printk("GNBD %s (built %s %s) installed\n",
	       GNBD_RELEASE_NAME, __DATE__, __TIME__);

	return 0;

      fail_cleanup:
	if (gnbd_devfs_entries[0])
		devfs_unregister(gnbd_devfs_entries[0]);
	if (gnbd_devices[0])
		kfree(gnbd_devices[0]);
	if (gnbd_device_names[0])
		kfree(gnbd_device_names[0]);
	if (gnbd_dir)
		devfs_unregister(gnbd_dir);
	del_gendisk(&gnbd_gendisk);
	gnbd_gendisk.next = NULL;
	blk_size[gnbd_major] = NULL;
	blksize_size[gnbd_major] = NULL;
	blk_cleanup_queue(BLK_DEFAULT_QUEUE(gnbd_major));
	devfs_unregister_blkdev(gnbd_major, "gnbd");
      fail:
	return result;
}

void __exit
exit_gnbd(void)
{
	int i;
	/* first of all, flush it all and reset all the data structures */

	for (i = 1; i < GNBD_MAX; i++) {
		if (gnbd_devices[i]) {
			fsync_dev(MKDEV(gnbd_major, i));	/* flush the devices */
			cleanup_device(i);
		}
	}

	/* special cleanup for gnbdcomm */

	devfs_unregister(gnbd_devfs_entries[0]);
	kfree(gnbd_device_names[0]);
	kfree(gnbd_devices[0]);
	del_gendisk(&gnbd_gendisk);
	gnbd_gendisk.next = NULL;

	blk_cleanup_queue(BLK_DEFAULT_QUEUE(gnbd_major));
	blk_size[gnbd_major] = NULL;
	blksize_size[gnbd_major] = NULL;

	/* finally, the usual cleanup */

	devfs_unregister(gnbd_dir);

	if (devfs_unregister_blkdev(gnbd_major, "gnbd") != 0)
		printe("couldn't unregister gnbd block device\n");
	else
		printd("gnbd: module cleaned up.\n");
	printd("Module gnbd unloaded\n");

}

module_init(init_gnbd);
module_exit(exit_gnbd);

MODULE_AUTHOR("Red Hat, Inc.");
MODULE_DESCRIPTION("Global Network Block Device " GNBD_RELEASE_NAME);
MODULE_LICENSE("GPL");

EXPORT_NO_SYMBOLS;
