/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <netdb.h>
#include <errno.h>
#include <dirent.h>
#include <ctype.h>
#include <sys/utsname.h>
#include "global.h"
#include "osi_endian.h"
#include "gnbd_import.h"
#include "gnbd_release.cf"
#include "copyright.cf"
#include "fence_return.h"

int message_flag = 1;
int override = 0;

#define MAN_MSG   "Please see man page for details.\n"

int gnbd_major = -1;
devlist_t *gnbd_devs;
int gnbdcomm;
unsigned int server_port = 14243;
int maxminor = 0;


int do_device_connect(int minor_nr){
  int result;
  gnbd_info_t info;
  info.minor_nr = minor_nr;
  result = ioctl(gnbdcomm, GNBD_DO_CONNECT, &info);
  if (result < 0)
    return result;
  return (int)info.readonly;
}

int get_timestamp(uint64 *timestamp){
  int result = ioctl(gnbdcomm, GNBD_DO_GET_TIME, timestamp);
  if (result < 0)
    printe("ioctl to get timestamp failed: %s\n", strerror(errno));
  return result;
}

int do_device_ping(int minor_nr){
  gnbd_info_t info;
  info.minor_nr = minor_nr;
  return ioctl(gnbdcomm, GNBD_DO_PING, &info);
}

int do_device_disconnect(int minor_nr){
  gnbd_info_t info;
  info.minor_nr = minor_nr;
  return ioctl(gnbdcomm, GNBD_DO_DISCONNECT, &info);
}

int do_fail_server(u32 ip){
  return ioctl(gnbdcomm, GNBD_DO_FAIL_SERV, ip);
}

int do_set_ip_port(int minor_nr, int port, u32 ip){
  gnbd_info_t info;
  info.minor_nr = minor_nr;
  info.ip = ip;
  info.port = port;
  return ioctl(gnbdcomm, GNBD_DO_SET_IP_PORT, &info);
}

int do_device_create(char *dev_name){
  gnbd_info_t info;
  strcpy(info.name, dev_name);
  if (ioctl(gnbdcomm, GNBD_DO_CREATE_DEV, &info) < 0)
    return -1;
  return (int)info.minor_nr;
}

int do_device_remove(int minor_nr){
  gnbd_info_t info;
  info.minor_nr = minor_nr;
  return ioctl(gnbdcomm, GNBD_DO_REMOVE_DEV, &info);
}

int get_major_nr(void){
  char device[LINE_MAX];
  char buf[LINE_MAX];
  int major_nr;
  FILE *fp;
  fp = fopen("/proc/devices", "r");
  if (!fp){
    printe("could not open /proc/devices: %s\n", strerror(errno));
    return -1;
  }
  while(fgets(buf, LINE_MAX, fp) != NULL)
    if (sscanf(buf, "%d %s", &major_nr, device) == 2 &&
	strcmp(device, "gnbd") == 0)
      break;
  if(strcmp(device, "gnbd") == 0)
    return major_nr;
  printe(
"could not find gnbd registered in /proc/devices.  This\n"
"probably means that you have not loaded the gnbd module.\n"
);
  return -1;
}

int get_comm_fd(void){
  int fd;
  struct stat stats;
  if (stat("/dev/gnbdcomm", &stats) == 0){
    if (major(stats.st_rdev) == gnbd_major && minor(stats.st_rdev) == 0){
      if( (fd=open("/dev/gnbdcomm", O_WRONLY)) < 0){
        printerr("cannot open comm device");
        return -1;
      }
      return fd;
    }
    else{
      printv("removing old gnbdcomm\n");
      if(unlink("/dev/gnbdcomm") < 0){
        printerr("cannot remove old gnbdcomm");
        return -1;
      }
    }
  }
  else if (errno != ENOENT){
    printerr("could not stat /dev/gnbdcomm");
    return -1;
  }
  printv("creating /dev/gnbdcomm\n");
  if (mknod("/dev/gnbdcomm", S_IFBLK | S_IRUSR | S_IWUSR |
            S_IRGRP | S_IROTH, makedev(gnbd_major, 0)) < 0){
    printerr("cannot create gnbdcomm");
    return -1;
  }
  if( (fd = open("/dev/gnbdcomm", O_WRONLY)) < 0){
    perror("cannot open comm device?!?!");
    return -1;
  }
  return fd;
}

int create_gnbd_dir(void){
  if(mkdir("/dev/gnbd", S_IRWXU | S_IRGRP | S_IXGRP |
           S_IROTH | S_IXOTH) == -1){
    if (errno == EEXIST)
      return 0;
    else{
      printerr("unable to create directory /dev/gnbd/");
      return -1;
    }
  }
  printm("created directory /dev/gnbd\n");
  return 0;
}

int update_devs(char *name, int minor_num, int readonly){
  struct stat buf;
  char path[LINE_MAX] = "/dev/gnbd/";
  mode_t mode;


  if (readonly)
    mode = S_IFBLK | S_IRUSR | S_IRGRP | S_IROTH;
  else
    mode = S_IFBLK | S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
  strcat(path, name);
  if(stat(path, &buf) == 0){
    if (major(buf.st_rdev) == gnbd_major && minor(buf.st_rdev) == minor_num &&
        buf.st_mode == mode)
      return 0;
    else{
      printv("%s does not match device. deleting\n", path);
      if (unlink(path) < 0){
	printe("cannot delete %s : %s\n", path, strerror(errno));
	return -1;
      }
    }
  }
  else if (errno != ENOENT){
    printe("could not stat %s : %s\n", path, strerror(errno));
    return -1;
  }
  printv("Creating block device for server %s\n", name);
  if (mknod(path, mode, makedev(gnbd_major, minor_num)) < 0){
    printe("couldn't create block device file %s : %s\n", path,
	   strerror(errno));
    return -1;
  }
  return 0;
}

devlist_t *make_dev_list(void){
  devlist_t *entry, *list = NULL;
  gnbd_info_t *gnbd_info;
  gnbd_list_arg_t list_req; 

  list_req.buf = NULL;
  list_req.size = 0;
  
  if (ioctl(gnbdcomm, GNBD_DO_LIST, &list_req) < 0){
    printe("cannot get gnbd list size: %s\n", strerror(errno));
    goto fail;
  }
  while(1){
    int buf_size = list_req.size;
    list_req.buf = (void *)malloc(buf_size);
    if (list_req.buf == NULL){
      printe("couldn't allocate enough memory for the gnbd device list : %s\n",
             strerror(errno));
      goto fail;
    }
    if (ioctl(gnbdcomm, GNBD_DO_LIST, &list_req) < 0){
      free(list_req.buf);
      if (list_req.size > buf_size)
        continue;
      printe("couldn't get list : %s\n", strerror(errno));
      goto fail;
    }
    break;
  }
  for (gnbd_info = (gnbd_info_t *)list_req.buf;
       (void *)gnbd_info < list_req.buf + list_req.size;
       gnbd_info++){
    entry = (devlist_t *)malloc(sizeof(devlist_t));
    if (entry == NULL){
      printe("couldn't allocated space for %s device information : %s\n",
             gnbd_info->name, strerror(errno));
      goto list_fail;
    }
    strncpy(entry->name, gnbd_info->name, 256);
    entry->name[255] = 0;
    entry->minor_nr = gnbd_info->minor_nr;
    entry->ip = gnbd_info->ip;
    entry->port = gnbd_info->port;
    entry->opened = gnbd_info->opened;
    entry->pending = gnbd_info->pending;
    entry->readonly = gnbd_info->readonly;
    entry->connected = gnbd_info->connected;
    entry->connect_failed = FALSE;
    entry->next = NULL;
    
    if (entry->minor_nr > maxminor)
      maxminor = entry->minor_nr;
    if (update_devs(entry->name, entry->minor_nr, entry->readonly) == -1)
      goto entry_fail;
    list = list_add(entry, list);
  }
  
  free(list_req.buf);
  return list;

 entry_fail:
  free(entry);
 list_fail:
  clear_list(list);
  free(list_req.buf);
 fail:
  return NULL;
}

int create_generic_file(int i){
  char path[LINE_MAX];
  int err;
  if (i >= 26)
    sprintf(path, "/dev/gnbd%c%c", 'a' + (i / 26) - 1, (i % 26) + 'a');
  else
    sprintf(path, "/dev/gnbd%c", i + 'a');
retry:
  err = mknod(path, S_IFBLK | S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH,
              makedev(gnbd_major, i));
  if (err < 0 && errno != EEXIST){
    printe("could not create file %s : %s\n", path, strerror(errno));
    return -1;
  }
  if (err < 0){
    struct stat stats;
    if (stat(path, &stats) < 0){
      printe("couldn't stat file %s : %s\n", path, strerror(errno));
      return -1;
    }
    if (major(stats.st_rdev) != gnbd_major || minor(stats.st_rdev) != i){
      if (unlink(path) < 0){
        printe("couldn't unlink file %s : %s\n", path, strerror(errno));
        return -1;
      }
      goto retry;
    }
  }
  return 0;
}


devlist_t *check_files(void){
  DIR *dp;
  struct dirent *entry;
  devlist_t *dev_list;
  int i;
  char path[LINE_MAX] = "/dev/gnbd/";
  dev_list = make_dev_list();
  dp = opendir("/dev/gnbd");
  if (dp == NULL){
    printerr("unable to open directory /dev/gnbd");
    clear_list(dev_list);
    return NULL;
  }
  while( (entry = readdir(dp)) != NULL){
    path[10] = 0;
    if (strcmp(entry->d_name, ".") == 0 ||
	strcmp(entry->d_name, "..") == 0)
      continue;
    if (list_match(entry->d_name, dev_list) == -1){
      strcat(path, entry->d_name);
      printv("%s has no server. Deleting\n", path);
      if(unlink(path) < 0){
	printe("cannot delete %s : %s\n", path, strerror(errno));
	clear_list(dev_list);
	closedir(dp);
	return NULL;
      }
    }
  }
  closedir(dp);
  /* make sure that there are enough generic gnbd devices to handle
     all the largest minor number */
  for (i = 1; i <= maxminor; i++){
    if (create_generic_file(i) < 0)
      return NULL;
  }
  return dev_list;
}


int do_server_connect(struct in_addr ipaddr, u16 port){
  struct sockaddr_in server;
  int sock_fd;
  sock_fd = socket(AF_INET, SOCK_STREAM, 0);
  if (sock_fd < 0){
    printe("error creating socket: %s\n", strerror(errno));
    return -1;
  }
  server.sin_addr.s_addr = ipaddr.s_addr;
  server.sin_port = htons((u16)port);
  server.sin_family = AF_INET;
  if (connect(sock_fd, (struct sockaddr *)&server, sizeof(server)) < 0){
    close(sock_fd);
    printe("error connecting to server: %s\n", strerror(errno));
    return -1;
  }
  return sock_fd;
}

int connect_to_server(char *hostname, u16 port,
		      struct in_addr *ip_addr){
  struct in_addr ip;
  struct hostent *hp;
  hp = gethostbyname(hostname);
  if (hp == NULL){
    printe("cannot resolve host ip from name %s: %s\n", hostname, 
	   hstrerror(h_errno));
    return -1;
  }
  if (memcpy(&(ip.s_addr), hp->h_addr, 4) == NULL){
    printe("cannot copy host address: %s\n", strerror(errno));
    return -1;
  }
  if (ip_addr != NULL)
    ip_addr->s_addr = ip.s_addr;
  return do_server_connect(ip, port);
}

int fail_server(char *host){
  u32 serv_ip;
  struct hostent *hp;
  
  hp = gethostbyname(host);
  if (hp == NULL){
    printe("cannot resolve server ip from name %s: %s\n", host,
           hstrerror(h_errno));
    return 1;
  }
  memcpy(&serv_ip, hp->h_addr, 4);
  if (do_fail_server(serv_ip) < 0){
    printe("ioctling gnbdcomm failed : %s\n", strerror(errno));
    return 1;
  }
  return 0;
}
  

int fence(char *host, char *server, int is_fence, uint64 timestamp){
  u32 fenceip;
  int sock_fd;
  uint32 msg;
  struct hostent *hp;
  
  hp = gethostbyname(host);
  if (hp == NULL){
    printe("cannot resolve host ip from name %s: %s\n", host,
	   hstrerror(h_errno));
    return 1;
  }
  memcpy(&fenceip, hp->h_addr, 4);
  sock_fd = connect_to_server(server, (u16)server_port, NULL);
  if (sock_fd < 0)
    return SERVER_ERR;
  if (is_fence)
    msg = cpu_to_be32(GNBD_SERV_FENCE_REQ);
  else
    msg = cpu_to_be32(GNBD_SERV_UNFENCE_REQ);
  if (write(sock_fd, &msg, sizeof(msg)) != sizeof(msg)){
    printe("can't send request to server : %s\n", strerror(errno));
    return SERVER_ERR;
  }
  if (write(sock_fd, &fenceip, sizeof(fenceip)) != sizeof(fenceip)){
    printe("can't send ip to server : %s\n", strerror(errno));
    return SERVER_ERR;
  }
  if (!is_fence){
    timestamp = cpu_to_be64(timestamp);
    if (write(sock_fd, &timestamp, sizeof(timestamp)) != sizeof(timestamp)){
      printe("can't send timestamp to server : %s\n", strerror(errno));
      return SERVER_ERR;
    }
  }
  if (read(sock_fd, &msg, sizeof(msg)) != sizeof(msg)){
    printe("can't read reply from server : %s\n", strerror(errno));
    return SERVER_ERR;
  }
  close(sock_fd);
  msg = be32_to_cpu(msg);
  if (msg != GNBD_SERV_SUCCESS_REPLY){
    if (is_fence)
      printe("fence failed : %s\n", strerror(REPLY_ERR(msg)));
    else
      printe("unfence failed : %s\n", strerror(REPLY_ERR(msg)));
    return SERVER_ERR;
  }
  if (is_fence){
    printv("%s fenced\n", host);
  }
  else{
    printv("%s unfenced\n", host);
  }
  return 0;
}

int remove_client(char *name){
  char path[300] = "/dev/gnbd/";
  devlist_t *ptr = gnbd_devs;
  while (ptr != NULL){
    if(strcmp(ptr->name, name) == 0){
      if (do_device_remove(ptr->minor_nr) < 0){
	printerr("cannot remove device");
	clear_list(gnbd_devs);
	return 1;
      }
      if (unlink(strcat(path,name)) < 0 && errno != ENOENT){
	printe("cannot remove file %s : %s\n", path, strerror(errno));
	clear_list(gnbd_devs);
	return 1;
      }
      printm("removed GNBD device \"%s\"\n", name);
      return 0;
    }
    ptr = ptr->next;
  }
  printe("couldn't find a device named %s\n", name);
  clear_list(gnbd_devs);
  return 1;
}

int remove_all(void){
  char path[300] = "/dev/gnbd/";
  devlist_t *ptr = gnbd_devs;
  while(ptr != NULL){
    path[10] = 0;
    if(do_device_remove(ptr->minor_nr) < 0){
      printe("cannot remove device %s : %s\n", ptr->name, strerror(errno));
      clear_list(gnbd_devs);
      return 1;
    }
    if (unlink(strcat(path,ptr->name)) < 0 && errno != ENOENT){
      printe("cannot remove file %s : %s\n", path, strerror(errno));
      clear_list(gnbd_devs);
      return 1;
    }
    printm("removed GNBD device \"%s\"\n", ptr->name);
    ptr = ptr->next;
  }
  clear_list(gnbd_devs);
  return 0;
}

/*
 * Makes sure that all gnbd_devs entries are either currently connected,
 * have information pending, or can connect to the server.  If the
 * entry is not and can not connect to the server and it has no pending
 * information, then remove it.
 *
 * Returns : -1 on error, Otherwise the number of invalid entries found.
 */
int validate_entries(void){
  int invalid = 0;
  char path[300] = "/dev/gnbd/";
  char *file = &path[10];
  devlist_t *prev, *ptr = gnbd_devs;

  prev = NULL;
  while (ptr != NULL){
    if (ptr->connected == INFO_CONNECTED && do_device_ping(ptr->minor_nr) < 0)
      ptr->connected = INFO_WAITING;
    if (ptr->connected == INFO_DISCONNECTED &&
	do_device_connect(ptr->minor_nr) < 0){
      invalid++;
      file[0] = 0;
      printm("GNBD '%s' unable to contact server.\n", ptr->name);
      if (ptr->pending == TRUE || ptr->opened == TRUE){
	ptr->connect_failed = TRUE;
	printv("GNBD '%s' in use. Ignoring\n",
               ptr->name);
	goto next_entry;
      }
      printm("Deleting entry\n");
      if (do_device_remove(ptr->minor_nr) < 0){
	printe("couldn't remove invalid server '%s' : %s\n", ptr->name,
	       strerror(errno));
	clear_list(gnbd_devs);
	return -1;
      }
      strcpy(file, ptr->name);
      if(unlink(path) < 0 && errno != ENOENT){
	printe("couldn't remove file /dev/gnbd/%s : %s\n", file,
	       strerror(errno));
	clear_list(gnbd_devs);
	return -1;
      }
      if(ptr == gnbd_devs){
	prev = ptr;
        ptr = ptr->next;
        gnbd_devs = ptr;
        free(prev);
      }
      else{
        devlist_t *tmp;
	prev->next = ptr->next;
	tmp = ptr;
	ptr = ptr->next;
        free(tmp);
      }
    }
    else{
    next_entry:
      prev = ptr;
      ptr = ptr->next;
    }
  }
  return invalid;
}

int read_from_server(char *host, u32 request, char **buf, struct in_addr *ip){
  struct in_addr ipaddr;
  int sock_fd;
  int n, total;
  uint32 msg;

  sock_fd = connect_to_server(host, (unsigned int)server_port, &ipaddr);
  if (sock_fd < 0)
    goto fail;
  msg = cpu_to_be32(request);
  if (write(sock_fd, &msg, sizeof(msg)) != sizeof(msg)){
    printe("error sending list request to %s : %s\n", host, strerror(errno));
    goto sock_fail;
  }
  if (read(sock_fd, &msg, sizeof(msg)) != sizeof(msg)){
    printe("error reading reply from %s : %s\n", host, strerror(errno));
    goto sock_fail;
  }
  msg = cpu_to_be32(msg);
  if (msg != GNBD_SERV_SUCCESS_REPLY){
    printe("list request to %s failed : %s\n", host, strerror(REPLY_ERR(msg)));
    goto sock_fail;
  }
  if (read(sock_fd, &msg, sizeof(msg)) != sizeof(msg)){
    printe("error reading size of list from %s : %s\n", host,
           strerror(errno));
    goto sock_fail;
  }
  msg = be32_to_cpu(msg);
  *buf = malloc(msg);
  if (*buf == NULL){
    printe("couldn't allocate memory for list : %s\n", strerror(errno));
    goto sock_fail;
  }
  memset(*buf, 0, msg);
  total = 0;
  while(total < msg){
    n = read(sock_fd, *buf + total, msg - total);
    if (n <= 0){
      printe("error reading list from server %s : %s\n", host,
             strerror(errno));
      goto malloc_fail;
    }
    total += n;
  }
  if (ip)
    *ip = ipaddr;
  close(sock_fd);
  return total;

  malloc_fail:
  free(*buf);
 sock_fail:
  close(sock_fd);
 fail:
  *buf = NULL;
  return -1;
}

int create_device(char *dev_name, int port, struct in_addr ipaddr){
  int result;
  devlist_t *ptr = gnbd_devs;
  int minor_nr;
  int readonly;
  mode_t mode;
  char path[300] = "/dev/gnbd/";
  while (ptr != NULL){
    if(strcmp(ptr->name, dev_name) == 0){
      if(ptr->connect_failed == FALSE){
        if (ptr->ip != ipaddr.s_addr){
          printm("There is already a working gnbd client with the name %s.\n" 
                 "Skipping.\n", dev_name);
        }
        else{
          printv("There is already a working gnbd client with the name %s.\n" 
                 "Skipping.\n", dev_name);
        }
	return 0;
      }
      if (!override){
        struct in_addr old_ip;
        old_ip.s_addr = ptr->ip;
        printf(
"GNBD '%s' unable to contact server at %s.\n"
"Alternate server found at %s. Warning! Connecting to\n"
"this server could result in data corruption if it does not export the\n"
"device previously exported by the server at %s\n"
"Are you sure you want to continue [yN]? ", dev_name, inet_ntoa(old_ip),
inet_ntoa(ipaddr), inet_ntoa(old_ip));
        while(1){
          int junk;
          result = tolower(getchar());
          if (result == '\n' || result == EOF)
            break;
          do
            junk = getchar();
          while (junk != '\n' && junk != EOF);
          if (result == 'y' || result == 'n')
            break;
          printf("Please type either 'y' or 'n' : ");
        }
        if (result != 'y'){
          printf("Alternate GNBD '%s' will not be imported\n", dev_name);
          return -1;
        }
      }
       ptr->ip = (u32)ipaddr.s_addr;
      ptr->port = port;
      if (do_set_ip_port(ptr->minor_nr, ptr->port, ptr->ip) < 0){
	printe("Could not set ip/port of %s : %s\n", dev_name,
	       strerror(errno));
	return -1;
      }
      if (do_device_connect(ptr->minor_nr) < 0){
	printe("Could not connect with the alternate server either : %s\n",
	       strerror(errno));
	return -1;
      }
      ptr->connect_failed = FALSE;
      printm("switched device %s to new server\n", dev_name);
      return 0; /* we return 0 so that the calling function doesn't
		   add a device entry to the device list */
    }
    ptr = ptr->next;
  }
  if( (minor_nr = do_device_create(dev_name)) < 0){
    printe("Could not create GNBD device %s : %s\n", dev_name,
           strerror(errno));
    return -1;
  }
  if(do_set_ip_port(minor_nr, port, (u32)ipaddr.s_addr) < 0){
    printe("Could not correctly set up GNBD device %s : %s\n", dev_name,
	   strerror(errno));
    do_device_remove(minor_nr);
    return -1;
  }
  readonly = do_device_connect(minor_nr);
  
  if (readonly < 0){
    printe("could not connect to %s : %s\n", dev_name, strerror(errno));
    return -1;
  }
  if (readonly)
    mode = S_IFBLK | S_IRUSR | S_IRGRP | S_IROTH;
  else
    mode = S_IFBLK | S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
  strcat(path,dev_name);
  if (mknod(path, mode, makedev(gnbd_major, minor_nr)) < 0 &&
      errno != EEXIST){
    printe("could not create block device file for %s : %s\n",
	   dev_name, strerror(errno));
    do_device_remove(minor_nr);
    return -1;
  }
  if (minor_nr > maxminor){
    if (create_generic_file(minor_nr) < 0){
      if (unlink(path) < 0)
	printe("CAN'T REMOVE THE FILE WE JUST MADE?!?!!? : %s\n",
	       strerror(errno));
      do_device_remove(minor_nr);
      return -1;
    }
  }
  printm("created GNBD device \"%s\"\n", dev_name);
  return minor_nr;
}


int setclients(char *serv){
  struct in_addr ipaddr;
  char *buf;
  char *ptr;
  uint64 timestamp;
  int minor_nr;
  struct utsname nodeinfo;
  int size;

  if (validate_entries() < 0)
    goto fail;
  if (serv == NULL){
    clear_list(gnbd_devs);
    return 0;
  }
  if (uname(&nodeinfo) < 0){
    printe("couldn't get my hostname\n");
    printe("because uname failed : %s\n", strerror(errno));
    return 1;
  }
  if (get_timestamp(&timestamp) < 0)
    goto fail;
  if (fence(nodeinfo.nodename, serv, 0, timestamp) != 0){
    printe("If you have been I/O fenced from the server, you must reload\n"
           "the gnbd module before re-importing devices\n"); 
    goto fail;
  }
  size = read_from_server(serv, GNBD_SERV_NAMES_LIST_REQ, &buf, &ipaddr);
  if (size < 0)
    goto fail;
  ptr = buf;
  while(ptr < buf + size){
    minor_nr = create_device(ptr, server_port, ipaddr);
    if (minor_nr < 0)
      goto buf_fail;
    ptr += strlen(ptr) + 1;
  }
  
  free(buf);
  clear_list(gnbd_devs);
  return 0;

 buf_fail:
  free(buf);
 fail:
  clear_list(gnbd_devs);
  return 1;
}

int list(void){
  devlist_t *ptr = gnbd_devs;
  u32 ip;
  char procname[8];
  while (ptr != NULL){
    ip = ntohl(ptr->ip);
    if (ptr->minor_nr >= 26)
      sprintf(procname, "gnbd%c%c", 'a' + (ptr->minor_nr / 26) - 1,
	      (ptr->minor_nr % 26) + 'a');
    else
      sprintf(procname, "gnbd%c", ptr->minor_nr + 'a');
    printm("Device name : %s\n"
	   "----------------------\n"
	   "    Minor # : %d\n"
	   "  Proc name : %s\n"
	   "         IP : %u.%u.%u.%u\n"
	   "       Port : %d\n"
	   "      State : %s %s %s\n"
           "   Readonly : %s\n\n",
	   ptr->name,
	   ptr->minor_nr,
	   procname,
	   (unsigned char)(0xff & (ip >> 24)),
	   (unsigned char)(0xff & (ip >> 16)),
	   (unsigned char)(0xff & (ip >> 8)),
	   (unsigned char)(0xff & ip),
	   ptr->port,
	   (ptr->opened)?"Open":"Close",
	   (ptr->connected == INFO_CONNECTED)?"Connected": ((ptr->connected == INFO_DISCONNECTED)? "Disconnected" : "Waiting"),
	   (ptr->pending)?"Pending":"Clear",
           (ptr->readonly)? "Yes" : "No");
    ptr = ptr->next;
  }
  return 0;
}

int get_serv_list(char *host){
  char *buf;
  char *ptr;
  int size;
  size = read_from_server(host, GNBD_SERV_NAMES_LIST_REQ, &buf, NULL);
  if (size < 0)
    return 1;
  ptr = buf;
  while (ptr < buf + size){
    printf("%s\n", ptr);
    ptr += strlen(ptr) + 1;
  }
  printf("\n");
  free(buf);
  return 0;
}

int get_banned_list(char *host){
  char *buf, *ptr;
  int size;
  struct in_addr ipaddr;
  size = read_from_server(host, GNBD_SERV_LIST_BANNED_REQ, &buf, NULL);
  if (size < 0)
    return 1;
  ptr = buf;
  while (ptr < buf + size){
    memcpy(&ipaddr.s_addr, ptr, sizeof(uint32));
    printm("%s\n", inet_ntoa(ipaddr));
    ptr += sizeof(uint32);
  }
  printm("\n");
  return 0;
}

int stats(int cmd, char *name) {
  devlist_t *ptr;
  char pathname[128];
  gnbd_stat_t stat;
  int fd;

  for (ptr = gnbd_devs; ptr != NULL; ptr = ptr->next) {

    if (name && strcmp(name, ptr->name))
        continue;

    memset(pathname, 0, 128);
    strcpy(pathname, "/dev/gnbd/");
    strcat(pathname, ptr->name);

    fd = open(pathname, O_RDONLY);
    if (fd < 0)
        continue;

    if (ioctl(fd, cmd, &stat))
        continue;

    close(fd);

    if (cmd == GNBD_STATS)
        printm("[%10s]  rd: %"PRIu64" rdblk: %"PRIu64" wr: %"PRIu64" wrblk: %"PRIu64"\n",
               ptr->name, stat.read_count, stat.read_block, 
               stat.write_count, stat.write_block);
  }

  return 0;
}

int usage(void){
  printf(
"Usage:\n"
"\n"
"gnbd_import [options]\n"
"\n"
"Options:\n"
"  -a               validate all imported GNBDs, and remove the invalid ones\n"
"  -C [GNBD | list] clear all gnbd statistics\n"
"  -c <server>      list all IPs currently IO fenced from the server\n"
"  -e <server>      list all GNBDs exported by the server\n"
"  -h               print this help message\n"
"  -i <server>      import all GNBDs from the server\n"
"  -k <server>      notify gnbd that the server has been fenced\n"
"  -l               list all imported GNBDs [default action]\n"
"  -O               Override prompts\n"
"  -p <port>        change the port to check on the GNBD server [default 14243]\n"
"  -q               quiet mode.\n"
"  -R               remove all imported GNBDs\n"
"  -r [GNBD | list] remove specified GNBD(s)\n"
"  -S [GNBD | list] print gnbd statistics\n"
"  -s <host>        IO fence the host (from known servers by default)\n"
"  -t <server>      specify a server for the IO fence (with -s or -u)\n"
"  -u <host>        unfence the host (from known servers by default)\n" 
"  -V               version information\n"
"  -v               verbose output\n"
"\n");
  return 0;
}
  
#define ACTION_FENCE         1
#define ACTION_REMOVE        2
#define ACTION_IMPORT        3
#define ACTION_LIST          4
#define ACTION_VALIDATE      5
#define ACTION_LIST_EXPORTED 6
#define ACTION_REMOVE_ALL    7
#define ACTION_UNFENCE       8
#define ACTION_LIST_BANNED   9
#define ACTION_STATS        10
#define ACTION_STATS_CLEAR  11
#define ACTION_FAIL_SERVER  12

char action_to_flag(int action){
  switch(action){
  case ACTION_FENCE:
    return 's';
  case ACTION_REMOVE:
    return 'r';
  case ACTION_IMPORT:
    return 'i';
  case ACTION_LIST:
    return 'l';
  case ACTION_VALIDATE:
    return 'a';
  case ACTION_LIST_EXPORTED:
    return 'e';
  case ACTION_REMOVE_ALL:
    return 'R';
  case ACTION_UNFENCE:
    return 'u';
  case ACTION_LIST_BANNED:
    return 'c';
  case ACTION_STATS:
    return 'S';
  case ACTION_STATS_CLEAR:
    return 'c';
  case ACTION_FAIL_SERVER:
    return 'k';
  default:
    printe("invalid action value\n");
    return 0;
  }
}

#define set_action(x) \
do{ \
  if (action){ \
    printe("flags -%c and -%c are not compatible\n", action_to_flag(action), \
           action_to_flag(x)); \
    fprintf(stderr, "Please see man page for details.\n"); \
    return 1; \
  } \
  action = (x); \
} while(0)

int main(int argc, char **argv){
  int c, i;
  int action = 0;
  char *host = NULL;
  char *fence_server = NULL;
  struct utsname nodeinfo;
  while( (c = getopt(argc, argv, "aCc:e:hi:k:lOp:qRrSs:t:u:Vv")) != -1){
    switch(c){
    case 'a':
      set_action(ACTION_VALIDATE);
      continue;
    case 'C':
      set_action(ACTION_STATS_CLEAR);
      continue;
    case 'c':
      set_action(ACTION_LIST_BANNED);
      host = optarg;
      continue;
    case 'e':
      set_action(ACTION_LIST_EXPORTED);
      host = optarg;
      continue;
    case 'h':
      return usage();
    case 'i':
      set_action(ACTION_IMPORT);
      host = optarg;
      continue;
    case 'k':
      set_action(ACTION_FAIL_SERVER);
      host = optarg;
      continue;
    case 'l':
      set_action(ACTION_LIST);
      continue;
    case 'O':
      override = 1;
      continue;
    case 'p':
      if (sscanf(optarg, "%u", &server_port) != 1){
        printe("invalid port number, %s\n" MAN_MSG, optarg);
        return 1;
      }
      continue;
    case 'q':
      message_flag = 0;
      continue;
    case 'R':
      set_action(ACTION_REMOVE_ALL);
      continue;
    case 'r':
      set_action(ACTION_REMOVE);
      continue;
    case 'S':
      set_action(ACTION_STATS);
      continue;
    case 's':
      set_action(ACTION_FENCE);
      host = optarg;
      continue;
    case 't':
      fence_server = optarg;
      continue;
    case 'u':
      set_action(ACTION_UNFENCE);
      host = optarg;
      continue;
    case 'V':
      printf("%s %s (built %s %s)\n", argv[0],
             GNBD_RELEASE_NAME, __DATE__, __TIME__);
      printf("%s\n", SISTINA_COPYRIGHT);
      return 0;
    case 'v':
      message_flag = 2;
      continue;
    default:
      printe("invalid option -- %c\n", c);
      printf("Please use '-h' for usage.\n");
      return 1;
    }
  }
  if (!action)
    action = ACTION_LIST;
  if (fence_server && action != ACTION_FENCE && action != ACTION_UNFENCE){
    printe("-t can only be used with -s or -u\n" MAN_MSG);
    return 1;
  }
  if ((action == ACTION_FENCE || action == ACTION_UNFENCE) &&
      !strcmp(host, "localhost")){
    if (uname(&nodeinfo) < 0){
      printe("couldn't get change localhost to actual hostname\n");
      printe("because uname failed : %s\n", strerror(errno));
      return 1;
    }
    host = nodeinfo.nodename;
  }
  if (argc != optind && action != ACTION_REMOVE && action != ACTION_STATS &&
      action != ACTION_STATS_CLEAR){
    printe("extra operand for action: %s\n", argv[optind]);
    fprintf(stderr, "please use '-h' for usage.\n");
    return 1;
  }
  if (action != ACTION_LIST_EXPORTED && action != ACTION_LIST_BANNED &&
      action != ACTION_FENCE && action != ACTION_UNFENCE){
    if (gnbd_major == -1)
      gnbd_major = get_major_nr();
    if (gnbd_major == -1)
      return 1;
    gnbdcomm = get_comm_fd();
    if (gnbdcomm == -1)
      return 1;
    if (create_gnbd_dir() == -1)
      return 1;
    gnbd_devs = check_files();
  }
  switch(action){
  case ACTION_FENCE:
  case ACTION_UNFENCE:
    if (!fence_server){
      printe("you must use -t with -e\n" MAN_MSG);
      return 1;
    }
    return fence(host, fence_server, (action == ACTION_FENCE)? 1 : 0, 0);
  case ACTION_REMOVE:
    if (optind == argc){
      printe("-r requires an argument or list of arguments\n");
      return 1;
    }
    for (i = optind; i < argc; i++){
      if (!argv[i]){
        printe("argv[%d] == NULL\n", i);
        return 1;
      }
      if (remove_client(argv[i]) != 0)
        return 1;
    }
    clear_list(gnbd_devs);
    return 0;
  case ACTION_REMOVE_ALL:
    return remove_all();
  case ACTION_VALIDATE:
    return setclients(NULL);
  case ACTION_IMPORT:
    return setclients(host);
  case ACTION_LIST:
    return list();
  case ACTION_LIST_EXPORTED:
    return get_serv_list(host);
  case  ACTION_LIST_BANNED:
    return get_banned_list(host);
  case ACTION_STATS:
  case ACTION_STATS_CLEAR:
    if (optind == argc)
      stats((action == ACTION_STATS) ? GNBD_STATS : GNBD_STATS_CLEAR, NULL);
    else
      for (i = optind; i < argc; i++)
        stats((action == ACTION_STATS) ? GNBD_STATS : GNBD_STATS_CLEAR,
              argv[i]);
    return 0;
  case ACTION_FAIL_SERVER:
    return fail_server(host);
  default:
    printe("unknown action\n");
    return 1;
  }
}
